;****************************************************************************
;****************************************************************************
;*
;* c64_memory.nasm
;*
;****************************************************************************
;****************************************************************************


;TODO:
;
;* move the 6510 I/O-Registers into link.nasm




;%define C64_MEMORY_TEST
;%define C64_MEMORY_DEBUG
;%define C64_MEMORY_TEST_VIA





%ifndef C64_MEMORY_TEST
%include "vic6569.inc"
%include "sid6581.inc"
%include "cia6526.inc"
%include "main.inc"

%ifdef C64_MEMORY_TEST_VIA
%include "via6522.inc"
%endif
%endif

%define C64_MEMORY_NASM
%include "c64_memory.inc"





;****************************************************************************
;****************************************************************************
;*
;* PART 1: Tables of 64KB RAM, Basic-, Character-, Kernel-ROM, Color-RAM
;*         6510 IO Ports
;*
;****************************************************************************
;****************************************************************************





section .data
%ifdef C64_MEMORY_DEBUG
ab__c64_test:		incbin "../../../../ec64-tests/images/vic2.prg"
;ab__c64_test:		incbin "../../../../ec64-tests/images/via6522_test.prg"
;ab__c64_test:		incbin "../../../../ec64-tests/images/CIA1TAB.P00"
ab__c64_monitor:	incbin "../../../../ec64-tests/images/c64_monitor.prg"
%endif

ab__c64_basic_rom:	incbin "../roms/c64_basic.rom"
ab__c64_character_rom:	incbin "../roms/c64_character.rom"
ab__c64_kernel_rom:	incbin "../roms/c64_kernel.rom"

section .bss
b__cpu6510_ddr:		resb	1
b__cpu6510_io:		resb	1
b__cpu6510_data:	resb	1
b__cpu6510_mask:	resb	1
ab__c64_ram:		resb	010000h
ab__c64_color_ram:	resb	000400h





;****************************************************************************
;****************************************************************************
;*
;* PART 2: The Memory Access Routines
;*
;****************************************************************************
;****************************************************************************





;****************************************************************************
;* c64_memory_initialize ****************************************************
;****************************************************************************
section .text
c64_memory_initialize:
			call	c64_memory_reset
			ret



;****************************************************************************
;* c64_memory_exit **********************************************************
;****************************************************************************
section .text
c64_memory_exit:
			ret



;****************************************************************************
;* c64_memory_save_snapshot *************************************************
;****************************************************************************
section .text
c64_memory_save_snapshot:
			ret



;****************************************************************************
;* c64_memory_load_snapshot *************************************************
;****************************************************************************
section .text
c64_memory_load_snapshot:
			ret



;****************************************************************************
;* c64_memory_reset *********************************************************
;****************************************************************************
c64_memory_reset:
			push	dword eax
			push	dword ecx
			push	dword edi
			xor	dword ecx, ecx
			mov	dword edi, ab__c64_ram
			cld
.fill_memory:		xor	dword eax, eax
			mov	byte  cl, 010h
			rep stosd
			dec	dword eax
			mov	byte  cl, 010h
			rep stosd
			cmp	dword edi, (ab__c64_ram + 010000h)
			jb	.fill_memory

%ifdef C64_MEMORY_DEBUG
			push	dword esi
			mov	dword esi, ab__c64_monitor
			mov	dword edi, ab__c64_ram + 0c000h
			mov	dword ecx, 0400h
			rep movsd
			mov	dword esi, ab__c64_test
			mov	dword edi, ab__c64_ram + 01000h
			mov	dword ecx, 0800h
			rep movsd
			pop	dword esi
%endif

			push	dword ebx
			xor	word  bx, bx		;set the correct 
			mov	byte  al, 02fh		;data direction
			call	c64_memory_write	;on the 6510
			inc	word  bx
			mov	byte  al, 027h		;Basic, I/O, Kernel
			call	c64_memory_write
			pop	dword ebx
			pop	dword edi
			pop	dword ecx
			pop	dword eax
			ret



;****************************************************************************
;* c64_memory_read **********************************************************
;****************************************************************************
section .text
c64_memory_read:
			and	dword ebx, 0ffffh
			mov	dword eax, ebx
			shr	dword eax, 12
			add	byte  al, [.b__functions_offset]
			jmp	dword [.ap__read_functions + 4 * eax]
.read_cpu6510:		cmp	word  bx, 00002h
			jb	.read_cpu6510_1
.read_ram:		mov	byte  al, [ab__c64_ram + ebx]
			ret
.read_cpu6510_1:	mov	byte  al, [b__cpu6510_ddr + ebx]
			ret
.read_basic_rom:	mov	byte  al, [ab__c64_basic_rom - 0a000h + ebx]
			ret
.read_io:		cmp	word  bx, 0d800h
			jae	.read_io_1
			cmp	word  bx, 0d400h
			jae	.read_sid
.read_vic:		call	vic6569_read
			ret
.read_sid:		call	sid6581_read
			ret
.read_io_1:		cmp	word  bx, 0dc00h
			jae	.read_io_2
.read_color_ram:	mov	byte  al, [ab__c64_color_ram - 0d800h + ebx]
			or	byte  al, 0f0h
			ret
.read_io_2:		cmp	word  bx, 0dd00h
			jb	.read_cia1
			cmp	word  bx, 0de00h
			jb	.read_cia2
%ifdef C64_MEMORY_TEST_VIA
			cmp	word  bx, 0df00h
			jb	.read_via1
%endif
.read_rubbish:		mov	byte  al, 05ah
			ret
.read_cia1:		mov	dword edi, cia6526__c64_1
			call	cia6526_read
			ret
.read_cia2:		mov	dword edi, cia6526__c64_2
			call	cia6526_read
			ret
%ifdef C64_MEMORY_TEST_VIA
.read_via1:		mov	dword edi, via6522__c64_1
			call	via6522_read
			ret
%endif
.read_character_rom:	mov	byte  al, [ab__c64_character_rom - 0d000h + ebx]
			ret
.read_kernel_rom:	mov	byte  al, [ab__c64_kernel_rom - 0e000h + ebx]
			ret

section .data
.b__functions_offset:	db	0
.ap__read_functions:	
			;BASIC ROM - I/O - KERNEL ROM

			dd	.read_cpu6510		;$0000 - $0fff
			dd	.read_ram		;$1000 - $1fff
			dd	.read_ram		;$2000 - $2fff
			dd	.read_ram		;$3000 - $3fff
			dd	.read_ram		;$4000 - $4fff
			dd	.read_ram		;$5000 - $5fff
			dd	.read_ram		;$6000 - $6fff
			dd	.read_ram		;$7000 - $7fff
			dd	.read_ram		;$8000 - $8fff
			dd	.read_ram		;$9000 - $9fff
			dd	.read_basic_rom		;$a000 - $afff
			dd	.read_basic_rom		;$b000 - $bfff
			dd	.read_ram		;$c000 - $cfff
			dd	.read_io		;$d000 - $dfff
			dd	.read_kernel_rom	;$e000 - $efff
			dd	.read_kernel_rom	;$f000 - $ffff

			;RAM - I/O - KERNEL ROM

			dd	.read_cpu6510		;$0000 - $0fff
			dd	.read_ram		;$1000 - $1fff
			dd	.read_ram		;$2000 - $2fff
			dd	.read_ram		;$3000 - $3fff
			dd	.read_ram		;$4000 - $4fff
			dd	.read_ram		;$5000 - $5fff
			dd	.read_ram		;$6000 - $6fff
			dd	.read_ram		;$7000 - $7fff
			dd	.read_ram		;$8000 - $8fff
			dd	.read_ram		;$9000 - $9fff
			dd	.read_ram		;$a000 - $afff
			dd	.read_ram		;$b000 - $bfff
			dd	.read_ram		;$c000 - $cfff
			dd	.read_io		;$d000 - $dfff
			dd	.read_kernel_rom	;$e000 - $efff
			dd	.read_kernel_rom	;$f000 - $ffff

			;RAM - I/O - RAM

			dd	.read_cpu6510		;$0000 - $0fff
			dd	.read_ram		;$1000 - $1fff
			dd	.read_ram		;$2000 - $2fff
			dd	.read_ram		;$3000 - $3fff
			dd	.read_ram		;$4000 - $4fff
			dd	.read_ram		;$5000 - $5fff
			dd	.read_ram		;$6000 - $6fff
			dd	.read_ram		;$7000 - $7fff
			dd	.read_ram		;$8000 - $8fff
			dd	.read_ram		;$9000 - $9fff
			dd	.read_ram		;$a000 - $afff
			dd	.read_ram		;$b000 - $bfff
			dd	.read_ram		;$c000 - $cfff
			dd	.read_io		;$d000 - $dfff
			dd	.read_ram		;$e000 - $efff
			dd	.read_ram		;$f000 - $ffff

			;RAM - RAM - RAM

			dd	.read_cpu6510		;$0000 - $0fff
			dd	.read_ram		;$1000 - $1fff
			dd	.read_ram		;$2000 - $2fff
			dd	.read_ram		;$3000 - $3fff
			dd	.read_ram		;$4000 - $4fff
			dd	.read_ram		;$5000 - $5fff
			dd	.read_ram		;$6000 - $6fff
			dd	.read_ram		;$7000 - $7fff
			dd	.read_ram		;$8000 - $8fff
			dd	.read_ram		;$9000 - $9fff
			dd	.read_ram		;$a000 - $afff
			dd	.read_ram		;$b000 - $bfff
			dd	.read_ram		;$c000 - $cfff
			dd	.read_ram		;$d000 - $dfff
			dd	.read_ram		;$e000 - $efff
			dd	.read_ram		;$f000 - $ffff

			;BASIC ROM - CHARACTER ROM - KERNEL ROM

			dd	.read_cpu6510		;$0000 - $0fff
			dd	.read_ram		;$1000 - $1fff
			dd	.read_ram		;$2000 - $2fff
			dd	.read_ram		;$3000 - $3fff
			dd	.read_ram		;$4000 - $4fff
			dd	.read_ram		;$5000 - $5fff
			dd	.read_ram		;$6000 - $6fff
			dd	.read_ram		;$7000 - $7fff
			dd	.read_ram		;$8000 - $8fff
			dd	.read_ram		;$9000 - $9fff
			dd	.read_basic_rom		;$a000 - $afff
			dd	.read_basic_rom		;$b000 - $bfff
			dd	.read_ram		;$c000 - $cfff
			dd	.read_character_rom	;$d000 - $dfff
			dd	.read_kernel_rom	;$e000 - $efff
			dd	.read_kernel_rom	;$f000 - $ffff

			;RAM - CHARACTER ROM - KERNEL ROM

			dd	.read_cpu6510		;$0000 - $0fff
			dd	.read_ram		;$1000 - $1fff
			dd	.read_ram		;$2000 - $2fff
			dd	.read_ram		;$3000 - $3fff
			dd	.read_ram		;$4000 - $4fff
			dd	.read_ram		;$5000 - $5fff
			dd	.read_ram		;$6000 - $6fff
			dd	.read_ram		;$7000 - $7fff
			dd	.read_ram		;$8000 - $8fff
			dd	.read_ram		;$9000 - $9fff
			dd	.read_ram		;$a000 - $afff
			dd	.read_ram		;$b000 - $bfff
			dd	.read_ram		;$c000 - $cfff
			dd	.read_character_rom	;$d000 - $dfff
			dd	.read_kernel_rom	;$e000 - $efff
			dd	.read_kernel_rom	;$f000 - $ffff

			;RAM - CHARACTER ROM - RAM

			dd	.read_cpu6510		;$0000 - $0fff
			dd	.read_ram		;$1000 - $1fff
			dd	.read_ram		;$2000 - $2fff
			dd	.read_ram		;$3000 - $3fff
			dd	.read_ram		;$4000 - $4fff
			dd	.read_ram		;$5000 - $5fff
			dd	.read_ram		;$6000 - $6fff
			dd	.read_ram		;$7000 - $7fff
			dd	.read_ram		;$8000 - $8fff
			dd	.read_ram		;$9000 - $9fff
			dd	.read_ram		;$a000 - $afff
			dd	.read_ram		;$b000 - $bfff
			dd	.read_ram		;$c000 - $cfff
			dd	.read_character_rom	;$d000 - $dfff
			dd	.read_ram		;$e000 - $efff
			dd	.read_ram		;$f000 - $ffff

;****************************************************************************
;* c64_memory_write *********************************************************
;****************************************************************************
section .text
c64_memory_write:
			and	dword ebx, 0ffffh
			cmp	word  bx, 00002h
			jb	.write_cpu6510
			jmp	dword [.p__write_function]
.write_cpu6510:		cmp	byte  bl, 001h
			je	.write_cpu6510_io
.write_cpu6510_ddr:	mov	byte  [b__cpu6510_ddr], al
			not	byte  al
			and	byte  al, 0dfh
			test	byte  al, 008h
			jz	.write_cpu6510_ddr_1
			mov	byte  ah, [b__cpu6510_io]
			and	byte  al, 0f7h
			and	byte  ah, 008h
			or	byte  al, ah
.write_cpu6510_ddr_1:	and	byte  al, 03fh
			mov	byte  [b__cpu6510_mask], al
			mov	byte  al, [b__cpu6510_data]
			jmp	.calculate_offset
.write_cpu6510_io:	mov	byte  [b__cpu6510_data], al
.calculate_offset:	and	byte  al, [b__cpu6510_ddr]
			or	byte  al, [b__cpu6510_mask]
			mov	byte  [b__cpu6510_io], al
			and	byte  al, 007h
			jnz	.calculate_offset_1
			mov	byte  al, 004h
.calculate_offset_1:	xor	byte  al, 007h
			shl	byte  al, 4
			mov	byte  [c64_memory_read.b__functions_offset], al
			cmp	byte  al, 030h
			jb	.calculate_offset_2
			mov	dword [.p__write_function], .write_ram
			ret
.calculate_offset_2:	mov	dword [.p__write_function], .write_io
			ret
.write_io:		cmp	word  bx, 0d000h
			jb	.write_ram
			cmp	word  bx, 0e000h
			jb	.write_io_1
.write_ram:		mov	byte  [ab__c64_ram + ebx], al
			ret
.write_io_1:		cmp	word  bx, 0d800h
			jae	.write_io_2
			cmp	word  bx, 0d400h
			jae	.write_sid
.write_vic:		call	vic6569_write
			ret
.write_sid:		call	sid6581_write
			ret
.write_io_2:		cmp	word  bx, 0dc00h
			jae	.write_io_3
.write_color_ram:	mov	byte  [ab__c64_color_ram - 0d800h + ebx], al
			ret
.write_io_3:		cmp	word  bx, 0dd00h
			jb	.write_cia1
%ifdef C64_MEMORY_TEST_VIA
			cmp	word  bx, 0de00h
			jb	.write_cia2
			cmp	word  bx, 0df00h
			jae	.write_io_4
.write_via1:		mov	dword edi, via6522__c64_1
			call	via6522_write
			ret
%else
			cmp	word  bx, 0de00h
			jae	.write_io_4
%endif
.write_cia2:		mov	dword edi, cia6526__c64_2
			call	cia6526_write
.write_io_4:		ret
.write_cia1:		mov	dword edi, cia6526__c64_1
			call	cia6526_write
			ret


section .data
.p__write_function:	dd	.write_io





;****************************************************************************
;****************************************************************************
;*
;* PART 3: Code for Selftest
;*
;****************************************************************************
;****************************************************************************





%ifdef C64_MEMORY_TEST

%include "sys_calls.inc"
%include "write.inc"

global _start

;****************************************************************************
;* Output Routines **********************************************************
;****************************************************************************
section .text
write_pass:
			inc	word  [.w__pass_counter]
			mov	dword eax, .cb__pass
			mov	dword ebx, 5
			call	write_text
			mov	word  ax, [.w__pass_counter]
			mov	byte  bl, 3
			call	write_hex
			mov	dword eax, .cb__points
			mov	dword ebx, 5
			call	write_text
			ret
section .data
.cb__pass:		db	"Pass "
.cb__points:		db	" ... "
.w__pass_counter:	dw	0

section .text
write_passed:
			mov	dword eax, .cb__passed
			mov	dword ebx, 7
			call	write_text
			ret

section .data
.cb__passed:		db	"passed", 00ah

section .text
write_failed:
			mov	dword eax, .cb__failed
			mov	dword ebx, 7
			call	write_text
			ret

section .data
.cb__failed:		db	"failed", 00ah



;****************************************************************************
;* Simple IO Device Function ************************************************
;****************************************************************************
section .text
vic6569_read:
vic6569_write:
sid6581_read:
sid6581_write:
cia6526_read:
cia6526_write:
			mov	byte  al, 081h
			ret

section .data
cia6526__c64_1:		dd	0
cia6526__c64_2:		dd	0



;****************************************************************************
;* The Test Program *********************************************************
;****************************************************************************
section .text
_start:
			mov	dword eax, .cb__message
			mov	dword ebx, .message_length
			call	write_text

			call	c64_memory_initialize
			call	c64_memory_reset

			xor	word  bx, bx
			mov	byte  al, 02fh
			call	c64_memory_write
			inc	word  bx
			mov	byte  al, 023h
			call	c64_memory_write

.test_pass_1:		call	write_pass
			mov	dword ebx, 0a000h
.pass_1_run:		call	c64_memory_read
			cmp	byte  al, [ab__c64_basic_rom - 0a000h + ebx]
			jne	.pass_1_failed
			inc	dword ebx
			cmp	dword ebx, 0c000h
			jne	.pass_1_run
			call	write_passed
			jmp	.test_pass_2
.pass_1_failed:		call	write_failed

.test_pass_2:		call	write_pass
			mov	dword ebx, 0d000h
.pass_2_run:		call	c64_memory_read
			cmp	byte  al, [ab__c64_character_rom - 0d000h + ebx]
			jne	.pass_2_failed
			inc	dword ebx
			cmp	dword ebx, 0e000h
			jne	.pass_2_run
			call	write_passed
			jmp	.test_pass_3
.pass_2_failed:		call	write_failed

.test_pass_3:		call	write_pass
			mov	dword ebx, 0e000h
.pass_3_run:		call	c64_memory_read
			cmp	byte  al, [ab__c64_kernel_rom - 0e000h + ebx]
			jne	.pass_3_failed
			inc	dword ebx
			cmp	dword ebx, 010000h
			jne	.pass_3_run
			call	write_passed
			jmp	.test_pass_4
.pass_3_failed:		call	write_failed

.test_pass_4:		call	write_pass
			mov	dword ebx, 0a000h
.pass_4_run:		call	c64_memory_read
			call	c64_memory_write
			inc	dword ebx
			cmp	dword ebx, 0c000h
			jne	.pass_4_run
			mov	dword ebx, 0a000h
.pass_4_control:	mov	byte  al, [ab__c64_ram + ebx]
			cmp	byte  al, [ab__c64_basic_rom - 0a000h + ebx]
			jne	.pass_4_failed
			inc	dword ebx
			cmp	dword ebx, 0c000h
			jne	.pass_4_control
			call	write_passed
			jmp	.test_pass_5
.pass_4_failed:		call	write_failed

.test_pass_5:		call	write_pass
			mov	dword ebx, 0d000h
.pass_5_run:		call	c64_memory_read
			call	c64_memory_write
			inc	dword ebx
			cmp	dword ebx, 0e000h
			jne	.pass_5_run
			mov	dword ebx, 0d000h
.pass_5_control:	mov	byte  al, [ab__c64_ram + ebx]
			cmp	byte  al, [ab__c64_character_rom - 0d000h + ebx]
			jne	.pass_5_failed
			inc	dword ebx
			cmp	dword ebx, 0e000h
			jne	.pass_5_control
			call	write_passed
			jmp	.test_pass_6
.pass_5_failed:		call	write_failed

.test_pass_6:		call	write_pass
			mov	dword ebx, 0e000h
.pass_6_run:		call	c64_memory_read
			call	c64_memory_write
			inc	dword ebx
			cmp	dword ebx, 010000h
			jne	.pass_6_run
			mov	dword ebx, 0e000h
.pass_6_control:	mov	byte  al, [ab__c64_ram + ebx]
			cmp	byte  al, [ab__c64_kernel_rom - 0e000h + ebx]
			jne	.pass_6_failed
			inc	dword ebx
			cmp	dword ebx, 010000h
			jne	.pass_6_control
			call	write_passed
			jmp	.test_pass_7
.pass_6_failed:		call	write_failed

.test_pass_7:		call	write_pass
			xor	word  bx, bx
			mov	byte  al, 02fh
			call	c64_memory_write
			inc	word  bx
			mov	byte  al, 027h
			call	c64_memory_write
			call	c64_memory_read
			cmp	byte  al, 037h
			jne	.pass_7_failed
			xor	word  bx, bx
			mov	byte  al, 000h
			call	c64_memory_write
			inc	word  bx
			call	c64_memory_read
			cmp	byte  al, 017h
			jne	.pass_7_failed
			mov	word  bx, 00001h
			mov	byte  al, 0ffh
			call	c64_memory_write
			call	c64_memory_read
			cmp	byte  al, 017h
			jne	.pass_7_failed
			xor	word  bx, bx
			mov	byte  al, 0ffh
			call	c64_memory_write
			inc	word  bx
			call	c64_memory_read
			cmp	byte  al, 0ffh
			jne	.pass_7_failed
.pass_7_passed:		call	write_passed
			jmp	.end
.pass_7_failed:		call	write_failed

.end:
			SYS_EXIT
			
section .data
.cb__message:		db	"C64 Memory Emulation Selftest", 00ah
.message_length:	equ	$ - .cb__message

%endif
;********************************************* karsten.scheibler@bigfoot.de *

