;****************************************************************************
;****************************************************************************
;*
;* cia6526.nasm
;*
;****************************************************************************
;****************************************************************************


;TODO:
;
;* TOD
;* Timeroutput on PB6, PB7
;* serial Register
;* register handshakes





;%define CIA6526_DEBUG





%include "defines.inc"
%include "main.inc"
%include "link.inc"

%define CIA6526_NASM
%include "cia6526.inc"



;****************************************************************************
;* assign's & defines's *****************************************************
;****************************************************************************
%assign PRA		CIA6526_PRA
%assign PRB		CIA6526_PRB
%assign DDRA		CIA6526_DDRA
%assign DDRB		CIA6526_DDRB
%assign TA_LO		CIA6526_TA_LO
%assign TA_HI		CIA6526_TA_HI
%assign TB_LO		CIA6526_TB_LO
%assign TB_HI		CIA6526_TB_HI
%assign TOD_10THS	CIA6526_TOD_10THS
%assign TOD_SEC		CIA6526_TOD_SEC
%assign TOD_MIN		CIA6526_TOD_MIN
%assign TOD_HR		CIA6526_TOD_HR
%assign SDR		CIA6526_SDR
%assign ICR		CIA6526_ICR
%assign CRA		CIA6526_CRA
%assign CRB		CIA6526_CRB

%assign PIPE_COUNT_1	000000001h
%assign PIPE_COUNT_2	000000002h
%assign PIPE_COUNT_3	000000004h
%assign PIPE_COUNT_4	000000008h
%assign PIPE_MASK	0ffffff00h
%assign PIPE_LOAD_1	000000100h
%assign PIPE_LOAD_2	000000200h
%assign PIPE_ONE_SHOT_1	000010000h
%assign PIPE_ONE_SHOT_2	000020000h



%assign LOCK_TIMER_A	001h
%assign LOCK_TIMER_B	002h
%assign GENERATE_IRQ	004h
%assign TRIGGER_TIMER_B	008h



%define MODUL_ID		CIA6526_MODUL_ID
%define EXECUTE_CYCLE		CIA6526_EXECUTE_CYCLE
%define EXECUTE_CYCLE_TA	CIA6526_EXECUTE_CYCLE_TA
%define EXECUTE_CYCLE_TB	CIA6526_EXECUTE_CYCLE_TB
%define LINK_IRQ		CIA6526_LINK_IRQ
%define LINK_PRA		CIA6526_LINK_PRA
%define LINK_PRB		CIA6526_LINK_PRB
%define STATE			CIA6526_STATE
%define CYCLE_PIPE_TA		CIA6526_CYCLE_PIPE_TA
%define LATCH_TA		CIA6526_LATCH_TA
%define COUNTER_TA		CIA6526_COUNTER_TA
%define CYCLE_PIPE_TB		CIA6526_CYCLE_PIPE_TB
%define LATCH_TB		CIA6526_LATCH_TB
%define COUNTER_TB		CIA6526_COUNTER_TB
%define LATCH_PRA		CIA6526_LATCH_PRA
%define SIGNAL_PRA		CIA6526_SIGNAL_PRA
%define LATCH_PRB		CIA6526_LATCH_PRB
%define SIGNAL_PRB		CIA6526_SIGNAL_PRB
%define REGISTERS		CIA6526_REGISTERS
%define IRQ_MASK		CIA6526_IRQ_MASK


;****************************************************************************
;* cia6526_initialize *******************************************************
;****************************************************************************
section .text
cia6526_initialize:
			ret



;****************************************************************************
;* cia6526_exit *************************************************************
;****************************************************************************
section .text
cia6526_exit:
			ret



;****************************************************************************
;* cia6526_save_snapshot ****************************************************
;****************************************************************************
section .text
cia6526_save_snapshot:
			ret



;****************************************************************************
;* cia6526_load_snapshost ***************************************************
;****************************************************************************
section .text
cia6526_load_snapshot:
			ret



%ifdef CIA6526_DEBUG
%include "sys_calls.inc"
;****************************************************************************
;* cia6526_debug ************************************************************
;****************************************************************************
section .text
cia6526_debug:
.read:			push	dword eax
			mov	byte  ah, 0ffh
			mov	byte  al, ah
			jmp	.common
.write:			push	dword eax
			xor	byte  ah, ah
.common:		mov	byte  [.ab__status], ah
			mov	byte  [.ab__status + 1], al
			mov	word  [.ab__status + 2], bx
			push	dword ebx
			push	dword ecx
			push	dword edx
			SYS_WRITE  2, .ab__status, 4
			pop	dword edx
			pop	dword ecx
			pop	dword ebx
			pop	dword eax
			ret

section .data
			align 4
.ab__status:		times 4 db 0
%endif


;****************************************************************************
;* cia6526_reset ************************************************************
;****************************************************************************
section .text
cia6526_reset:
			xor	dword eax, eax
			xor	dword ecx, ecx
			push	dword edi
			mov	byte  cl, 010h
			lea	dword edi, [REGISTERS]
			cld
			rep stosb
			pop	dword edi
			mov	dword [CYCLE_PIPE_TA], eax
			mov	dword [LATCH_TA], eax
			mov	dword [COUNTER_TA], eax
			mov	dword [CYCLE_PIPE_TB], eax
			mov	dword [LATCH_TB], eax
			mov	dword [COUNTER_TB], eax
			mov	dword [STATE], eax
			mov	byte  [LATCH_PRA], al
			mov	byte  [SIGNAL_PRA], al
			mov	byte  [LATCH_PRB], al
			mov	byte  [SIGNAL_PRB], al
			mov	byte  [IRQ_MASK], al
			mov	dword [EXECUTE_CYCLE], cia6526_cycle
			mov	dword [EXECUTE_CYCLE_TA], cia6526_cycle.return
			mov	dword [EXECUTE_CYCLE_TB], cia6526_cycle.return
			ret



;****************************************************************************
;* cia6526_cycle ************************************************************
;****************************************************************************
section .text
cia6526_cycle:
			call	dword [EXECUTE_CYCLE_TA]
			call	dword [EXECUTE_CYCLE_TB]
			mov	dword eax, [STATE]
			test	dword eax, GENERATE_IRQ
			jz	.no_irq
			mov	byte  bl, [REGISTERS + ICR]
			and	dword eax, (NOT32 - GENERATE_IRQ)
			or	byte  bl, 080h
			mov	dword [STATE], eax
			mov	byte  al, 1
			mov	byte  [REGISTERS + ICR], bl
			call	dword [LINK_IRQ]
.no_irq:
.return:		ret



;****************************************************************************
;* cia6526_dump *************************************************************
;****************************************************************************
section .text
cia6526_dump:
			ret



;****************************************************************************
;* cia6526_read *************************************************************
;****************************************************************************
section .text
cia6526_read:
			;-----------------------
			;get the handler address
			;-----------------------

%ifdef CIA6526_DEBUG
			call	cia6526_debug.read
%endif
			mov	dword ebp, ebx
			push	dword ecx
			and	dword ebp, 00000000fh
			mov	dword ecx, [STATE]
			jmp	dword [.ap__registers + 4 * ebp]
.ap__registers:		dd	.read_register
			dd	.read_register
			dd	.read_register
			dd	.read_register
			dd	.read_ta_lo
			dd	.read_ta_hi
			dd	.read_tb_lo
			dd	.read_tb_hi
			dd	.read_register
			dd	.read_register
			dd	.read_register
			dd	.read_tod_hr
			dd	.read_register
			dd	.read_icr
			dd	.read_register
			dd	.read_register

			;----------------------
			;ordinary register read
			;----------------------

.read_register:		mov	byte  al, [REGISTERS + ebp]
			pop	dword ecx
			ret

			;--------------------------
			;read low byte from timer a
			;--------------------------

.read_ta_lo:		or	dword ecx, LOCK_TIMER_A
			mov	byte  al, [REGISTERS + TA_LO]
			mov	dword [STATE], ecx
			pop	dword ecx
			ret

			;---------------------------
			;read high byte from timer a
			;---------------------------

.read_ta_hi:		and	dword ecx, (NOT32 - LOCK_TIMER_A)
			mov	byte  al, [REGISTERS + TA_HI]
			mov	dword [STATE], ecx
			pop	dword ecx
			ret

			;--------------------------
			;read low byte from timer b
			;--------------------------

.read_tb_lo:		or	dword ecx, LOCK_TIMER_B
			mov	byte  al, [REGISTERS + TB_LO]
			mov	dword [STATE], ecx
			pop	dword ecx
			ret

			;---------------------------
			;read high byte from timer b
			;---------------------------

.read_tb_hi:		and	dword ecx, (NOT32 - LOCK_TIMER_B)
			mov	byte  al, [REGISTERS + TB_HI]
			mov	dword [STATE], ecx
			pop	dword ecx
			ret

			;-----------
			;read tod hr
			;-----------

.read_tod_hr:		mov	byte  al, 080h
			pop	dword ecx
			ret

			;-----------------------------------------------
			;read the icr
			;Note: every read clears all irq's from this CIA
			;-----------------------------------------------

.read_icr:		and	dword ecx, (NOT32 - GENERATE_IRQ)
			xor	byte  al, al
			mov	dword [STATE], ecx
			call	dword [LINK_IRQ]
			xor	byte  cl, cl
			mov	byte  al, [REGISTERS + ICR]
			mov	byte  [REGISTERS + ICR], cl
			pop	dword ecx
			ret



;****************************************************************************
;* cia6526_write ************************************************************
;****************************************************************************
section .text
cia6526_write:
			;-----------------------
			;get the handler address
			;-----------------------

%ifdef CIA6526_DEBUG
			call	cia6526_debug.write
%endif
			mov	dword ebp, ebx
			and	dword ebp, 00000000fh
			jmp	dword [.ap__registers + 4 * ebp]
.ap__registers:		dd	.write_pra
			dd	.write_prb
			dd	.write_ddra
			dd	.write_ddrb
			dd	.write_ta_lo
			dd	.write_ta_hi
			dd	.write_tb_lo
			dd	.write_tb_hi
			dd	.write_ignore
			dd	.write_ignore
			dd	.write_ignore
			dd	.write_ignore
			dd	.write_ignore
			dd	.write_icr
			dd	.write_cra
			dd	.write_crb

			;-------------
			;ignore writes
			;-------------

.write_ignore:		ret

			;-------
			;set PRA
			;-------

.write_pra:		mov	byte  [LATCH_PRA], al
			mov	byte  ah, [REGISTERS + DDRA]
			and	byte  al, ah
			not	byte  ah
			and	byte  ah, [SIGNAL_PRA]
			or	byte  al, ah
			mov	byte  [REGISTERS + PRA], al
			call	dword [LINK_PRA]
			ret

			;-------
			;set PRB
			;-------

.write_prb:		mov	byte  [LATCH_PRB], al
			mov	byte  ah, [REGISTERS + DDRB]
			and	byte  al, ah
			not	byte  ah
			and	byte  ah, [SIGNAL_PRB]
			or	byte  al, ah
			mov	byte  [REGISTERS + PRB], al
			call	dword [LINK_PRB]
			ret

			;--------
			;set DDRA
			;--------

.write_ddra:		mov	byte  [REGISTERS + DDRA], al
			mov	byte  ah, [LATCH_PRA]
			and	byte  ah, al
			not	byte  al
			and	byte  al, [SIGNAL_PRA]
			or	byte  ah, al
			mov	byte  [REGISTERS + PRA], ah
			call	dword [LINK_PRA]
			ret

			;--------
			;set DDRB
			;--------

.write_ddrb:		mov	byte  [REGISTERS + DDRB], al
			mov	byte  ah, [LATCH_PRB]
			and	byte  ah, al
			not	byte  al
			and	byte  al, [SIGNAL_PRB]
			or	byte  ah, al
			mov	byte  [REGISTERS + PRB], ah
			call	dword [LINK_PRB]
			ret

			;---------
			;set TA_LO
			;---------

.write_ta_lo:		mov	byte  [LATCH_TA], al
			ret

			;---------
			;set TA_HI
			;---------

.write_ta_hi:		mov	byte  [LATCH_TA + 1], al
			test	byte  [REGISTERS + CRA], 001h
			jnz	.write_ta_hi_1
			mov	dword ebp, [LATCH_TA]
			and	dword [CYCLE_PIPE_TA], PIPE_MASK
			mov	dword [COUNTER_TA], ebp
.write_ta_hi_1:		ret

			;---------
			;set TB_LO
			;---------

.write_tb_lo:		mov	byte  [LATCH_TB], al
			ret

			;---------
			;set TB_HI
			;---------

.write_tb_hi:		mov	byte  [LATCH_TB + 1], al
			test	byte  [REGISTERS + CRB], 001h
			jnz	.write_tb_hi_1
			mov	dword ebp, [LATCH_TB]
			and	dword [CYCLE_PIPE_TB], PIPE_MASK
			mov	dword [COUNTER_TB], ebp
.write_tb_hi_1:		ret

			;--------------
			;write into ICR
			;--------------

.write_icr:		test	byte  al, 080h
			jz	.write_icr_1
			and	byte  al, 07fh
			or	byte  [IRQ_MASK], al
			ret
.write_icr_1:		xor	byte  al, 07fh
			and	byte  [IRQ_MASK], al
			ret

			;--------------
			;write into cra
			;--------------

.write_cra:		test	byte  al, 010h
			jz	.write_cra_1
			or	dword [CYCLE_PIPE_TA], PIPE_LOAD_1
			and	byte  al, 0efh
.write_cra_1:		mov	byte  [REGISTERS + CRA], al
			test	byte  al, 001h
			jz	.write_cra_3
			and	dword [STATE], (NOT32 - LOCK_TIMER_A)
			test	byte  al, 020h
			jz	.write_cra_2
			mov	dword [EXECUTE_CYCLE_TA], timer_a.no_phi2
			ret
.write_cra_2:		mov	dword [EXECUTE_CYCLE_TA], timer_a
			ret
.write_cra_3:		mov	dword [EXECUTE_CYCLE_TA], cia6526_cycle.return
			ret

			;--------------
			;write into crb
			;--------------

.write_crb:		test	byte  al, 010h
			jz	.write_crb_1
			or	dword [CYCLE_PIPE_TB], PIPE_LOAD_1
			and	byte  al, 0efh
.write_crb_1:		mov	byte  [REGISTERS + CRB], al
			mov	dword ebp, [STATE]
			test	byte  al, 001h
			jz	.write_crb_4
			and	dword ebp, (NOT32 - LOCK_TIMER_B)
			and	byte  al, 060h
			jz	.write_crb_3
			and	dword ebp, (NOT32 - TRIGGER_TIMER_B)
			cmp	byte  al, 020h
			je	.write_crb_2
			or	dword ebp, TRIGGER_TIMER_B
.write_crb_2:		mov	dword [EXECUTE_CYCLE_TB], timer_b.no_phi2
			mov	dword [STATE], ebp
			ret
.write_crb_3:		mov	dword [EXECUTE_CYCLE_TB], timer_b
			mov	dword [STATE], ebp
			ret
.write_crb_4:		mov	dword [EXECUTE_CYCLE_TB], cia6526_cycle.return
			ret



;****************************************************************************
;* cia6526_signal_pra *******************************************************
;****************************************************************************
section .text
cia6526_signal_pra:
			push	dword eax
			mov	byte  [SIGNAL_PRA], al
			mov	byte  ah, [REGISTERS + DDRA]
			xor	byte  ah, NOT8
			and	byte  al, ah
			mov	byte  ah, [LATCH_PRA]
			and	byte  ah, [REGISTERS + DDRA]
			or	byte  ah, al
			mov	byte  [REGISTERS + PRA], ah
			pop	dword eax
			ret



;****************************************************************************
;* cia6526_signal_prb *******************************************************
;****************************************************************************
section .text
cia6526_signal_prb:
			push	dword eax
			mov	byte  [SIGNAL_PRB], al
			mov	byte  ah, [REGISTERS + DDRB]
			xor	byte  ah, NOT8
			and	byte  al, ah
			mov	byte  ah, [LATCH_PRB]
			and	byte  ah, [REGISTERS + DDRB]
			or	byte  ah, al
			mov	byte  [REGISTERS + PRB], ah
			pop	dword eax
			ret



;****************************************************************************
;* Timer A Emulation ********************************************************
;****************************************************************************
timer_a:
			or	dword [CYCLE_PIPE_TA], PIPE_COUNT_2
.no_phi2:		mov	dword eax, [CYCLE_PIPE_TA]
			mov	dword ebx, [COUNTER_TA]
			test	dword eax, PIPE_COUNT_4
			jz	.count_1
			dec	dword ebx
.count_1:		or	dword ebx, ebx
			jnz	.count_2
			test	dword eax, PIPE_COUNT_3
			jz	.count_2
.load:			mov	dword ebx, [LATCH_TA]
			and	dword eax, (NOT32 - PIPE_COUNT_3)

			;!!!!!!!!!!!!
			;TODO: PB6Out
			;!!!!!!!!!!!!

			test	dword [STATE], TRIGGER_TIMER_B
			jz	.one_shot_1
			or	dword [CYCLE_PIPE_TB], PIPE_COUNT_2
.one_shot_1:		test	byte  [REGISTERS + CRA], 008h
			jnz	.one_shot_2
			test	dword eax, PIPE_ONE_SHOT_2
			jz	.set_icr
.one_shot_2:		and	byte  [REGISTERS + CRA], (NOT8 - 001h)
			mov	dword [EXECUTE_CYCLE_TA], cia6526_cycle.return
.set_icr:		mov	byte  cl, 001h
			test	byte  [IRQ_MASK], cl
			jz	.no_irq
			or	dword [STATE], GENERATE_IRQ
.no_irq:		or	byte  cl, [REGISTERS + ICR]
			mov	byte  [REGISTERS + ICR], cl
			jmp	.exit
.count_2:		test	dword eax, PIPE_LOAD_2
			jz	.exit
			mov	dword ebx, [LATCH_TA]
			and	dword eax, (NOT32 - PIPE_COUNT_3)
.exit:			shl	dword eax, 1
			mov	dword [COUNTER_TA], ebx
			and	dword eax, 07f7f7f7fh
			mov	dword [CYCLE_PIPE_TA], eax
			mov	byte  [REGISTERS + TA_LO], bl
			test	dword [STATE], LOCK_TIMER_A
			jnz	.exit_1
			mov	byte  [REGISTERS + TA_HI], bh
.exit_1:		ret



;****************************************************************************
;* Timer B Emulation ********************************************************
;****************************************************************************
timer_b:
			or	dword [CYCLE_PIPE_TB], PIPE_COUNT_2
.no_phi2:		mov	dword eax, [CYCLE_PIPE_TB]
			mov	dword ebx, [COUNTER_TB]
			test	dword eax, PIPE_COUNT_4
			jz	.count_1
			dec	dword ebx
.count_1:		or	dword ebx, ebx
			jnz	.count_2
			test	dword eax, PIPE_COUNT_3
			jz	.count_2
.load:			mov	dword ebx, [LATCH_TB]
			and	dword eax, (NOT32 - PIPE_COUNT_3)

			;!!!!!!!!!!!!
			;TODO: PB7Out
			;!!!!!!!!!!!!

			test	byte  [REGISTERS + CRB], 008h
			jnz	.one_shot
			test	dword eax, PIPE_ONE_SHOT_2
			jz	.set_icr
.one_shot:		and	byte  [REGISTERS + CRB], (NOT8 - 001h)
			mov	dword [EXECUTE_CYCLE_TB], cia6526_cycle.return
.set_icr:		mov	byte  cl, 002h
			test	byte  [IRQ_MASK], cl
			jz	.no_irq
			or	dword [STATE], GENERATE_IRQ
.no_irq:		or	byte  cl, [REGISTERS + ICR]
			mov	byte  [REGISTERS + ICR], cl
			jmp	.exit
.count_2:		test	dword eax, PIPE_LOAD_2
			jz	.exit
			mov	dword ebx, [LATCH_TB]
			and	dword eax, (NOT32 - PIPE_COUNT_3)
.exit:			shl	dword eax, 1
			mov	dword [COUNTER_TB], ebx
			and	dword eax, 07f7f7f7fh
			mov	dword [CYCLE_PIPE_TB], eax
			mov	byte  [REGISTERS + TB_LO], bl
			test	dword [STATE], LOCK_TIMER_B
			jnz	.exit_1
			mov	byte  [REGISTERS + TB_HI], bh
.exit_1:		ret
;********************************************* karsten.scheibler@bigfoot.de *

