;****************************************************************************
;****************************************************************************
;*
;* cpu65xx.inc
;*
;****************************************************************************
;****************************************************************************





%ifndef CPU65XX_INC
%define CPU65XX_INC

;****************************************************************************
;* Exported Symbols *********************************************************
;****************************************************************************
%ifdef CPU65XX_NASM
global cpu65xx_initialize
global cpu65xx_exit
global cpu65xx_save_snapshot
global cpu65xx_load_snapshot
global cpu65xx_reset
global cpu65xx_cycle
global cpu65xx_dump
global cpu65xx_signal_set_irq
global cpu65xx_signal_clear_irq
global cpu65xx_signal_set_nmi
global cpu65xx_signal_clear_nmi
global cpu65xx_signal_set_dma
global cpu65xx_signal_clear_dma
global cpu65xx_signal_set_overflow
%else
extern cpu65xx_initialize
extern cpu65xx_exit
extern cpu65xx_save_snapshot
extern cpu65xx_load_snapshot
extern cpu65xx_reset
extern cpu65xx_cycle
extern cpu65xx_reset
extern cpu65xx_dump
extern cpu65xx_signal_set_irq
extern cpu65xx_signal_clear_irq
extern cpu65xx_signal_set_nmi
extern cpu65xx_signal_clear_nmi
extern cpu65xx_signal_set_dma
extern cpu65xx_signal_clear_dma
extern cpu65xx_signal_set_overflow
%endif



;****************************************************************************
;* Module Data **************************************************************
;****************************************************************************
				struc cpu65xx_data
				alignb 4
d__cpu_modul_id:		resd	1
p__cpu_execute_cycle:		resd	1
p__cpu_delay_address:		resd	1
p__cpu_rw_address:		resd	1
p__cpu_opcode_address:		resd	1
p__cpu_rmw_address:		resd	1
p__cpu_memory_read:		resd	1
p__cpu_memory_read_temp:	resd	1
p__cpu_memory_write:		resd	1
d__cpu_state:			resd	1
d__cpu_program_counter:		resd	1
d__cpu_stack_pointer:		resd	1
d__cpu_absolute_address:	resd	1
d__cpu_zeropage_address:	resd	1
b__cpu_cycle:			resb	1
b__cpu_nmi_cycle:		resb	1
b__cpu_irq_cycle:		resb	1
b__cpu_opcode:			resb	1
b__cpu_accumulator_temp:	resb	1
b__cpu_accumulator:		resb	1
b__cpu_index_x:			resb	1
b__cpu_index_y:			resb	1
b__cpu_flag_register:		resb	1
				endstruc



;****************************************************************************
;* assign's & defines's *****************************************************
;****************************************************************************
%define CPU65XX_MODUL_ID		byte esi + d__cpu_modul_id
%define CPU65XX_EXECUTE_CYCLE		byte esi + p__cpu_execute_cycle
%define CPU65XX_DELAY_ADDRESS		byte esi + p__cpu_delay_address
%define CPU65XX_RW_ADDRESS		byte esi + p__cpu_rw_address
%define CPU65XX_OPCODE_ADDRESS		byte esi + p__cpu_opcode_address
%define CPU65XX_RMW_ADDRESS		byte esi + p__cpu_rmw_address
%define CPU65XX_MEMORY_READ		byte esi + p__cpu_memory_read
%define CPU65XX_MEMORY_READ_TEMP	byte esi + p__cpu_memory_read_temp
%define CPU65XX_MEMORY_WRITE		byte esi + p__cpu_memory_write
%define CPU65XX_STATE			byte esi + d__cpu_state
%define CPU65XX_PROGRAM_COUNTER		byte esi + d__cpu_program_counter
%define CPU65XX_STACK_POINTER   	byte esi + d__cpu_stack_pointer
%define CPU65XX_ABSOLUTE_ADDRESS	byte esi + d__cpu_absolute_address
%define CPU65XX_ZEROPAGE_ADDRESS	byte esi + d__cpu_zeropage_address
%define CPU65XX_CYCLE			byte esi + b__cpu_cycle
%define CPU65XX_IRQ_CYCLE		byte esi + b__cpu_irq_cycle
%define CPU65XX_NMI_CYCLE		byte esi + b__cpu_nmi_cycle
%define CPU65XX_OPCODE			byte esi + b__cpu_opcode
%define CPU65XX_ACCUMULATOR_TEMP	byte esi + b__cpu_accumulator_temp
%define CPU65XX_ACCUMULATOR		byte esi + b__cpu_accumulator
%define CPU65XX_INDEX_X			byte esi + b__cpu_index_x
%define CPU65XX_INDEX_Y			byte esi + b__cpu_index_y
%define CPU65XX_FLAG_REGISTER		byte esi + b__cpu_flag_register



%assign CPU65XX_ILLEGAL_OPCODE		000000001h
%assign CPU65XX_ILLEGAL_ADDRESS_MODE	000000002h
%assign CPU65XX_SIGNAL_IRQ1		000000100h
%assign CPU65XX_SIGNAL_IRQ2		000000200h
%assign CPU65XX_SIGNAL_IRQ3		000000400h
%assign CPU65XX_SIGNAL_IRQ4		000000800h
%assign CPU65XX_SIGNAL_IRQ		000002000h
%assign CPU65XX_SIGNAL_NMI		000004000h
%assign CPU65XX_SIGNAL_DMA		000008000h
%assign CPU65XX_IRQ_DELAY_OK		000080000h
%assign CPU65XX_NMI_DELAY_OK		000100000h
%assign CPU65XX_PROCESSING_IRQ  	000200000h
%assign CPU65XX_PROCESSING_NMI  	000400000h
%assign CPU65XX_PROCESSING_RESET	000800000h



;****************************************************************************
;* Macro's ******************************************************************
;****************************************************************************
%macro CPU65XX_CALL_CYCLE 0-1
%if %0 > 0
			mov	dword esi, %1
%endif
			inc	byte  [CPU65XX_CYCLE]
			call	dword [CPU65XX_EXECUTE_CYCLE]
%endmacro



%macro CPU65XX_CALL_RESET 4
			mov	dword esi, %1
			mov	dword eax, %2
			mov	dword ebx, %3
			mov	dword ecx, %4
			mov	dword [CPU65XX_MODUL_ID], eax
			mov	dword [CPU65XX_MEMORY_READ], ebx
			mov	dword [CPU65XX_MEMORY_READ_TEMP], ebx
			mov	dword [CPU65XX_MEMORY_WRITE], ecx
			call	cpu65xx_reset
%endmacro



%endif
;********************************************* karsten.scheibler@bigfoot.de *

