;****************************************************************************
;****************************************************************************
;*
;* cpu65xx.nasm
;*
;****************************************************************************


;TODO:
;
;* fix decimal mode
;* possibility to make snapshots
;* include illegal opcodes





;%define CPU65XX_DEBUG





%ifdef CPU65XX_DEBUG
%include "sys_calls.inc"
%endif

%include "defines.inc"
%include "write.inc"

%ifndef CPU65XX_TEST
%include "main.inc"
%endif

%define CPU65XX_NASM
%include "cpu65xx.inc"



;****************************************************************************
;* assign's & defines's *****************************************************
;****************************************************************************
%define MODUL_ID		CPU65XX_MODUL_ID
%define EXECUTE_CYCLE		CPU65XX_EXECUTE_CYCLE
%define DELAY_ADDRESS		CPU65XX_DELAY_ADDRESS
%define RW_ADDRESS		CPU65XX_RW_ADDRESS
%define OPCODE_ADDRESS		CPU65XX_OPCODE_ADDRESS
%define RMW_ADDRESS		CPU65XX_RMW_ADDRESS
%define MEMORY_READ		CPU65XX_MEMORY_READ
%define MEMORY_READ_TEMP	CPU65XX_MEMORY_READ_TEMP
%define MEMORY_WRITE		CPU65XX_MEMORY_WRITE
%define STATE			CPU65XX_STATE
%define PROGRAM_COUNTER		CPU65XX_PROGRAM_COUNTER
%define STACK_POINTER   	CPU65XX_STACK_POINTER
%define ABSOLUTE_ADDRESS	CPU65XX_ABSOLUTE_ADDRESS
%define ZEROPAGE_ADDRESS	CPU65XX_ZEROPAGE_ADDRESS
%define CYCLE			CPU65XX_CYCLE
%define IRQ_CYCLE		CPU65XX_IRQ_CYCLE
%define NMI_CYCLE		CPU65XX_NMI_CYCLE
%define OPCODE			CPU65XX_OPCODE
%define ACCUMULATOR_TEMP	CPU65XX_ACCUMULATOR_TEMP
%define ACCUMULATOR		CPU65XX_ACCUMULATOR
%define INDEX_X			CPU65XX_INDEX_X
%define INDEX_Y			CPU65XX_INDEX_Y
%define FLAG_REGISTER		CPU65XX_FLAG_REGISTER



;****************************************************************************
;* The 6502/6510 Flags ******************************************************
;****************************************************************************
;* the reserved and the break flag seem to be always set.
;*
%assign CPU65XX_FLAG_N		080h	;negative, 1 = true
%assign CPU65XX_FLAG_N_BIT	7
%assign CPU65XX_FLAG_V		040h	;overflow, 1 = true
%assign CPU65XX_FLAG_V_BIT	6
%assign CPU65XX_FLAG_R		020h	;reserved flag
%assign CPU65XX_FLAG_R_BIT	5
%assign CPU65XX_FLAG_B		010h	;break
%assign CPU65XX_FLAG_B_BIT	4
%assign CPU65XX_FLAG_D		008h	;decimal mode, 1 = true
%assign CPU65XX_FLAG_D_BIT	3
%assign CPU65XX_FLAG_I		004h	;irq disable, 1 = disable
%assign CPU65XX_FLAG_I_BIT	2
%assign CPU65XX_FLAG_Z		002h	;zero, 1 = result is zero
%assign CPU65XX_FLAG_Z_BIT	1
%assign CPU65XX_FLAG_C		001h	;carry, 1 = true
%assign CPU65XX_FLAG_C_BIT	0



;****************************************************************************
;* Constants for ap__opcode_details *****************************************
;****************************************************************************
%assign ADDRESS_ILLEGAL			0x00
%assign ADDRESS_IMPLICIT_READ		0x01
%assign ADDRESS_ACCUMULATOR_RMW		0x02
%assign ADDRESS_IMMEDIATE_READ		0x03
%assign ADDRESS_ZEROPAGE_READ		0x04
%assign ADDRESS_ZEROPAGE_WRITE		0x05
%assign ADDRESS_ZEROPAGE_RMW		0x04
%assign ADDRESS_ZEROPAGE_X_READ		0x06
%assign ADDRESS_ZEROPAGE_X_WRITE	0x07
%assign ADDRESS_ZEROPAGE_X_RMW		0x06
%assign ADDRESS_ZEROPAGE_Y_READ		0x08
%assign ADDRESS_ZEROPAGE_Y_WRITE	0x09
%assign ADDRESS_ABSOLUTE_READ		0x0a
%assign ADDRESS_ABSOLUTE_WRITE		0x0b
%assign ADDRESS_ABSOLUTE_RMW		0x0a
%assign ADDRESS_ABSOLUTE_X_READ		0x0c
%assign ADDRESS_ABSOLUTE_X_WRITE	0x0d
%assign ADDRESS_ABSOLUTE_X_RMW		0x0e
%assign ADDRESS_ABSOLUTE_Y_READ		0x0f
%assign ADDRESS_ABSOLUTE_Y_WRITE	0x10
%assign ADDRESS_INDIRECT_READ		0x11
%assign ADDRESS_X_INDIRECT_READ		0x12
%assign ADDRESS_X_INDIRECT_WRITE	0x13
%assign	ADDRESS_INDIRECT_Y_READ		0x14
%assign ADDRESS_INDIRECT_Y_WRITE	0x15
%assign ADDRESS_RELATIVE_READ		0x16

%assign OPCODE_ILLEGAL		0x00
%assign OPCODE_ADC		0x01
%assign OPCODE_AND		0x02
%assign OPCODE_ASL		0x03
%assign OPCODE_BCC		0x04
%assign OPCODE_BCS		0x05
%assign OPCODE_BEQ		0x06
%assign OPCODE_BIT		0x07
%assign OPCODE_BMI		0x08
%assign OPCODE_BNE		0x09
%assign OPCODE_BPL		0x0a
%assign OPCODE_BRK		0x0b
%assign OPCODE_BVC		0x0c
%assign OPCODE_BVS		0x0d
%assign OPCODE_CLC		0x0e
%assign OPCODE_CLD		0x0f
%assign OPCODE_CLI		0x10
%assign OPCODE_CLV		0x11
%assign OPCODE_CMP		0x12
%assign OPCODE_CPX		0x13
%assign OPCODE_CPY		0x14
%assign OPCODE_DEC		0x15
%assign OPCODE_DEX		0x16
%assign OPCODE_DEY		0x17
%assign OPCODE_EOR		0x18
%assign OPCODE_INC		0x19
%assign OPCODE_INX		0x1a
%assign OPCODE_INY		0x1b
%assign OPCODE_JMP		0x1c
%assign OPCODE_JSR		0x1d
%assign OPCODE_LDA		0x1e
%assign OPCODE_LDX		0x1f
%assign OPCODE_LDY		0x20
%assign OPCODE_LSR		0x21
%assign OPCODE_NOP		0x22
%assign OPCODE_ORA		0x23
%assign OPCODE_PHA		0x24
%assign OPCODE_PHP		0x25
%assign OPCODE_PLA		0x26
%assign OPCODE_PLP		0x27
%assign OPCODE_ROL		0x28
%assign OPCODE_ROR		0x29
%assign OPCODE_RTI		0x2a
%assign OPCODE_RTS		0x2b
%assign OPCODE_SBC		0x2c
%assign OPCODE_SEC		0x2d
%assign OPCODE_SED		0x2e
%assign OPCODE_SEI		0x2f
%assign OPCODE_STA		0x30
%assign OPCODE_STX		0x31
%assign OPCODE_STY		0x32
%assign OPCODE_TAX		0x33
%assign OPCODE_TAY		0x34
%assign OPCODE_TSX		0x35
%assign OPCODE_TXA		0x36
%assign OPCODE_TXS		0x37
%assign OPCODE_TYA		0x38



;****************************************************************************
;****************************************************************************
;*
;* PART 1: Opcode Decoding Tables
;*
;****************************************************************************
;****************************************************************************





;****************************************************************************
;* The Jumptable ************************************************************
;****************************************************************************
section .data
ab__opcode_details:
			;---------
			;$00 - $0f
			;---------

			db	ADDRESS_IMPLICIT_READ,		OPCODE_BRK
			db	ADDRESS_X_INDIRECT_READ,	OPCODE_ORA
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ZEROPAGE_READ,		OPCODE_ORA
			db	ADDRESS_ZEROPAGE_RMW,		OPCODE_ASL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_IMPLICIT_READ,		OPCODE_PHP
			db	ADDRESS_IMMEDIATE_READ,		OPCODE_ORA
			db	ADDRESS_ACCUMULATOR_RMW,	OPCODE_ASL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ABSOLUTE_READ,		OPCODE_ORA
			db	ADDRESS_ABSOLUTE_RMW,		OPCODE_ASL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL

			;---------
			;$10 - $1f
			;---------

			db	ADDRESS_RELATIVE_READ,		OPCODE_BPL
			db	ADDRESS_INDIRECT_Y_READ,	OPCODE_ORA
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ZEROPAGE_X_READ,	OPCODE_ORA
			db	ADDRESS_ZEROPAGE_X_RMW,		OPCODE_ASL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_IMPLICIT_READ,		OPCODE_CLC
			db	ADDRESS_ABSOLUTE_Y_READ,	OPCODE_ORA
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ABSOLUTE_X_READ,	OPCODE_ORA
			db	ADDRESS_ABSOLUTE_X_RMW,		OPCODE_ASL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL

			;---------
			;$20 - $2f
			;---------

			db	ADDRESS_IMPLICIT_READ,		OPCODE_JSR
			db	ADDRESS_X_INDIRECT_READ,	OPCODE_AND
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ZEROPAGE_READ,		OPCODE_BIT
			db	ADDRESS_ZEROPAGE_READ,		OPCODE_AND
			db	ADDRESS_ZEROPAGE_RMW,		OPCODE_ROL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_IMPLICIT_READ,		OPCODE_PLP
			db	ADDRESS_IMMEDIATE_READ,		OPCODE_AND
			db	ADDRESS_ACCUMULATOR_RMW,	OPCODE_ROL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ABSOLUTE_READ,		OPCODE_BIT
			db	ADDRESS_ABSOLUTE_READ,		OPCODE_AND
			db	ADDRESS_ABSOLUTE_RMW,		OPCODE_ROL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL

			;---------
			;$30 - $3f
			;---------

			db	ADDRESS_RELATIVE_READ,		OPCODE_BMI
			db	ADDRESS_INDIRECT_Y_READ,	OPCODE_AND
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ZEROPAGE_X_READ,	OPCODE_AND
			db	ADDRESS_ZEROPAGE_X_RMW,		OPCODE_ROL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_IMPLICIT_READ,		OPCODE_SEC
			db	ADDRESS_ABSOLUTE_Y_READ,	OPCODE_AND
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ABSOLUTE_X_READ,	OPCODE_AND
			db	ADDRESS_ABSOLUTE_X_RMW,		OPCODE_ROL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL

			;---------
			;$40 - $4f
			;---------

			db	ADDRESS_IMPLICIT_READ,		OPCODE_RTI
			db	ADDRESS_X_INDIRECT_READ,	OPCODE_EOR
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ZEROPAGE_READ,		OPCODE_EOR
			db	ADDRESS_ZEROPAGE_RMW,		OPCODE_LSR
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_IMPLICIT_READ,		OPCODE_PHA
			db	ADDRESS_IMMEDIATE_READ,		OPCODE_EOR
			db	ADDRESS_ACCUMULATOR_RMW,	OPCODE_LSR
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_IMPLICIT_READ,		OPCODE_JMP
			db	ADDRESS_ABSOLUTE_READ,		OPCODE_EOR
			db	ADDRESS_ABSOLUTE_RMW,		OPCODE_LSR
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL

			;---------
			;$50 - $5f
			;---------

			db	ADDRESS_RELATIVE_READ,		OPCODE_BVC
			db	ADDRESS_INDIRECT_Y_READ,	OPCODE_EOR
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ZEROPAGE_X_READ,	OPCODE_EOR
			db	ADDRESS_ZEROPAGE_X_RMW,		OPCODE_LSR
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_IMPLICIT_READ,		OPCODE_CLI
			db	ADDRESS_ABSOLUTE_Y_READ,	OPCODE_EOR
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ABSOLUTE_X_READ,	OPCODE_EOR
			db	ADDRESS_ABSOLUTE_X_RMW,		OPCODE_LSR
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL

			;---------
			;$60 - $6f
			;---------

			db	ADDRESS_IMPLICIT_READ,		OPCODE_RTS
			db	ADDRESS_X_INDIRECT_READ,	OPCODE_ADC
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ZEROPAGE_READ,		OPCODE_ADC
			db	ADDRESS_ZEROPAGE_RMW,		OPCODE_ROR
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_IMPLICIT_READ,		OPCODE_PLA
			db	ADDRESS_IMMEDIATE_READ,		OPCODE_ADC
			db	ADDRESS_ACCUMULATOR_RMW,	OPCODE_ROR
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_INDIRECT_READ,		OPCODE_JMP
			db	ADDRESS_ABSOLUTE_READ,		OPCODE_ADC
			db	ADDRESS_ABSOLUTE_RMW,		OPCODE_ROR
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL

			;---------
			;$70 - $7f
			;---------

			db	ADDRESS_RELATIVE_READ,		OPCODE_BVS
			db	ADDRESS_INDIRECT_Y_READ,	OPCODE_ADC
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ZEROPAGE_X_READ,	OPCODE_ADC
			db	ADDRESS_ZEROPAGE_X_RMW,		OPCODE_ROR
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_IMPLICIT_READ,		OPCODE_SEI
			db	ADDRESS_ABSOLUTE_Y_READ,	OPCODE_ADC
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ABSOLUTE_X_READ,	OPCODE_ADC
			db	ADDRESS_ABSOLUTE_X_RMW,		OPCODE_ROR
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL

			;---------
			;$80 - $8f
			;---------

			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_X_INDIRECT_WRITE,	OPCODE_STA
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ZEROPAGE_WRITE,		OPCODE_STY
			db	ADDRESS_ZEROPAGE_WRITE,		OPCODE_STA
			db	ADDRESS_ZEROPAGE_WRITE,		OPCODE_STX
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_IMPLICIT_READ,		OPCODE_DEY
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_IMPLICIT_READ,		OPCODE_TXA
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ABSOLUTE_WRITE,		OPCODE_STY
			db	ADDRESS_ABSOLUTE_WRITE,		OPCODE_STA
			db	ADDRESS_ABSOLUTE_WRITE,		OPCODE_STX
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL

			;---------
			;$90 - $9f
			;---------

			db	ADDRESS_RELATIVE_READ,		OPCODE_BCC
			db	ADDRESS_INDIRECT_Y_WRITE,	OPCODE_STA
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ZEROPAGE_X_WRITE,	OPCODE_STY
			db	ADDRESS_ZEROPAGE_X_WRITE,	OPCODE_STA
			db	ADDRESS_ZEROPAGE_Y_WRITE,	OPCODE_STX
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_IMPLICIT_READ,		OPCODE_TYA
			db	ADDRESS_ABSOLUTE_Y_WRITE,	OPCODE_STA
			db	ADDRESS_IMPLICIT_READ,		OPCODE_TXS
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ABSOLUTE_X_WRITE,	OPCODE_STA
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL

			;---------
			;$a0 - $af
			;---------

			db	ADDRESS_IMMEDIATE_READ,		OPCODE_LDY
			db	ADDRESS_X_INDIRECT_READ,	OPCODE_LDA
			db	ADDRESS_IMMEDIATE_READ,		OPCODE_LDX
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ZEROPAGE_READ,		OPCODE_LDY
			db	ADDRESS_ZEROPAGE_READ,		OPCODE_LDA
			db	ADDRESS_ZEROPAGE_READ,		OPCODE_LDX
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_IMPLICIT_READ,		OPCODE_TAY
			db	ADDRESS_IMMEDIATE_READ,		OPCODE_LDA
			db	ADDRESS_IMPLICIT_READ,		OPCODE_TAX
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ABSOLUTE_READ,		OPCODE_LDY
			db	ADDRESS_ABSOLUTE_READ,		OPCODE_LDA
			db	ADDRESS_ABSOLUTE_READ,		OPCODE_LDX
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL

			;---------
			;$b0 - $bf
			;---------

			db	ADDRESS_RELATIVE_READ,		OPCODE_BCS
			db	ADDRESS_INDIRECT_Y_READ,	OPCODE_LDA
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ZEROPAGE_X_READ,	OPCODE_LDY
			db	ADDRESS_ZEROPAGE_X_READ,	OPCODE_LDA
			db	ADDRESS_ZEROPAGE_Y_READ,	OPCODE_LDX
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_IMPLICIT_READ,		OPCODE_CLV
			db	ADDRESS_ABSOLUTE_Y_READ,	OPCODE_LDA
			db	ADDRESS_IMPLICIT_READ, 		OPCODE_TSX
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ABSOLUTE_X_READ,	OPCODE_LDY
			db	ADDRESS_ABSOLUTE_X_READ,	OPCODE_LDA
			db	ADDRESS_ABSOLUTE_Y_READ,	OPCODE_LDX
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL

			;---------
			;$c0 - $cf
			;---------

			db	ADDRESS_IMMEDIATE_READ,		OPCODE_CPY
			db	ADDRESS_X_INDIRECT_READ,	OPCODE_CMP
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ZEROPAGE_READ,		OPCODE_CPY
			db	ADDRESS_ZEROPAGE_READ,		OPCODE_CMP
			db	ADDRESS_ZEROPAGE_RMW,		OPCODE_DEC
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_IMPLICIT_READ,		OPCODE_INY
			db	ADDRESS_IMMEDIATE_READ,		OPCODE_CMP
			db	ADDRESS_IMPLICIT_READ, 		OPCODE_DEX
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ABSOLUTE_READ,		OPCODE_CPY
			db	ADDRESS_ABSOLUTE_READ,		OPCODE_CMP
			db	ADDRESS_ABSOLUTE_RMW,		OPCODE_DEC
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL

			;---------
			;$d0 - $df
			;---------

			db	ADDRESS_RELATIVE_READ,		OPCODE_BNE
			db	ADDRESS_INDIRECT_Y_READ,	OPCODE_CMP
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ZEROPAGE_X_READ,	OPCODE_CMP
			db	ADDRESS_ZEROPAGE_X_RMW,		OPCODE_DEC
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_IMPLICIT_READ,		OPCODE_CLD
			db	ADDRESS_ABSOLUTE_Y_READ,	OPCODE_CMP
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ABSOLUTE_X_READ,	OPCODE_CMP
			db	ADDRESS_ABSOLUTE_X_RMW,		OPCODE_DEC
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL

			;---------
			;$e0 - $ef
			;---------

			db	ADDRESS_IMMEDIATE_READ,		OPCODE_CPX
			db	ADDRESS_X_INDIRECT_READ,	OPCODE_SBC
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ZEROPAGE_READ,		OPCODE_CPX
			db	ADDRESS_ZEROPAGE_READ,		OPCODE_SBC
			db	ADDRESS_ZEROPAGE_RMW,		OPCODE_INC
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_IMPLICIT_READ,		OPCODE_INX
			db	ADDRESS_IMMEDIATE_READ,		OPCODE_SBC
			db	ADDRESS_IMPLICIT_READ, 		OPCODE_NOP
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ABSOLUTE_READ,		OPCODE_CPX
			db	ADDRESS_ABSOLUTE_READ,		OPCODE_SBC
			db	ADDRESS_ABSOLUTE_RMW,		OPCODE_INC
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL

			;---------
			;$f0 - $ff
			;---------

			db	ADDRESS_RELATIVE_READ,		OPCODE_BEQ
			db	ADDRESS_INDIRECT_Y_READ,	OPCODE_SBC
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ZEROPAGE_X_READ,	OPCODE_SBC
			db	ADDRESS_ZEROPAGE_X_RMW,		OPCODE_INC
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
;			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_IMPLICIT_READ,		OPCODE_SED
			db	ADDRESS_ABSOLUTE_Y_READ,	OPCODE_SBC
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL
			db	ADDRESS_ABSOLUTE_X_READ,	OPCODE_SBC
			db	ADDRESS_ABSOLUTE_X_RMW,		OPCODE_INC
			db	ADDRESS_ILLEGAL,		OPCODE_ILLEGAL



;****************************************************************************
;* Addressing Modes *********************************************************
;****************************************************************************
section .data
ap__address_modes:	
			;---------------
			;ADDRESS_ILLEGAL
			;---------------

			dd	opcode_illegal
			dd	opcode_illegal
			dd	opcode_illegal
			dd	opcode_illegal

			;---------------------
			;ADDRESS_IMPLICIT_READ
			;---------------------

			dd	address_implicit	;INX
			dd	address_illegal
			dd	address_illegal
			dd	address_illegal

			;---------------
			;ACCUMULATOR_RMW
			;---------------

			dd	address_accumulator	;ASL
			dd	address_illegal
			dd	address_illegal
			dd	address_accumulator.rmw

			;----------------------
			;ADDRESS_IMMEDIATE_READ
			;----------------------

			dd	address_immediate	;LDA #$??
			dd	address_illegal
			dd	address_illegal
			dd	address_illegal

			;-------------------------------------------
			;ADDRESS_ZEROPAGE_READ, ADDRESS_ZEROPAGE_RMW
			;-------------------------------------------

			dd	address_zeropage	;LDA $??  ASL $??
			dd	address_illegal
			dd	address_zeropage.read
			dd	address_zeropage.rmw

			;----------------------
			;ADDRESS_ZEROPAGE_WRITE
			;----------------------

			dd	address_zeropage	;STA $??
			dd	address_illegal
			dd	address_zeropage.write
			dd	address_illegal

			;-----------------------------------------------
			;ADDRESS_ZEROPAGE_X_READ, ADDRESS_ZEROPAGE_X_RMW
			;-----------------------------------------------

			dd	address_zeropage.x	;LDA $??,X  ASL $??,X
			dd	address_illegal
			dd	address_zeropage.read
			dd	address_zeropage.rmw

			;------------------------
			;ADDRESS_ZEROPAGE_X_WRITE
			;------------------------

			dd	address_zeropage.x	;STA $??,X
			dd	address_illegal
			dd	address_zeropage.write
			dd	address_illegal

			;-----------------------
			;ADDRESS_ZEROPAGE_Y_READ
			;-----------------------

			dd	address_zeropage.y	;LDX $??,Y
			dd	address_illegal
			dd	address_zeropage.read
			dd	address_illegal

			;------------------------
			;ADDRESS_ZEROPAGE_Y_WRITE
			;------------------------

			dd	address_zeropage.y	;STX $??,Y
			dd	address_illegal
			dd	address_zeropage.write
			dd	address_illegal

			;-------------------------------------------
			;ADDRESS_ABSOLUTE_READ, ADDRESS_ABSOLUTE_RMW
			;-------------------------------------------

			dd	address_absolute	;LDA $????  ASL $????
			dd	address_illegal
			dd	address_absolute.read
			dd	address_absolute.rmw

			;----------------------
			;ADDRESS_ABSOLUTE_WRITE
			;----------------------

			dd	address_absolute	;STA $????
			dd	address_illegal
			dd	address_absolute.write
			dd	address_illegal

			;-----------------------
			;ADDRESS_ABSOLUTE_X_READ
			;-----------------------

			dd	address_absolute.x	;LDA $????,X
			dd	address_absolute.read
			dd	address_absolute.read
			dd	address_illegal

			;------------------------
			;ADDRESS_ABSOLUTE_X_WRITE
			;------------------------

			dd	address_absolute.x	;STA $????
			dd	address_absolute.delay
			dd	address_absolute.write
			dd	address_illegal

			;----------------------
			;ADDRESS_ABSOLUTE_X_RMW
			;----------------------

			dd	address_absolute.x	;ASL $????,X
			dd	address_absolute.delay
			dd	address_absolute.read
			dd	address_absolute.rmw

			;-----------------------
			;ADDRESS_ABSOLUTE_Y_READ
			;-----------------------

			dd	address_absolute.y	;LDA $????,Y
			dd	address_absolute.read
			dd	address_absolute.read
			dd	address_illegal

			;------------------------
			;ADDRESS_ABSOLUTE_Y_WRITE
			;------------------------

			dd	address_absolute.y	;STA $????,Y
			dd	address_absolute.delay
			dd	address_absolute.write
			dd	address_illegal

			;---------------------
			;ADDRESS_INDIRECT_READ
			;---------------------

			dd	address_indirect	;JMP ($????)
			dd	address_illegal
			dd	address_illegal
			dd	address_illegal

			;-----------------------
			;ADDRESS_X_INDIRECT_READ
			;-----------------------

			dd	address_x_indirect	;LDA ($??,X)
			dd	address_illegal
			dd	address_x_indirect.read
			dd	address_illegal

			;------------------------
			;ADDRESS_X_INDIRECT_WRITE
			;------------------------

			dd	address_x_indirect	;STA ($??,X)
			dd	address_illegal
			dd	address_x_indirect.write
			dd	address_illegal

			;-----------------------
			;ADDRESS_INDIRECT_Y_READ
			;-----------------------

			dd	address_indirect_y	;LDA ($??),Y
			dd	address_indirect_y.read
			dd	address_indirect_y.read
			dd	address_illegal

			;------------------------
			;ADDRESS_INDIRECT_Y_WRITE
			;------------------------

			dd	address_indirect_y	;STA ($??),Y
			dd	address_indirect_y.delay
			dd	address_indirect_y.write
			dd	address_illegal

			;---------------------
			;ADDRESS_RELATIVE_READ
			;---------------------

			dd	address_relative	;BEQ $??
			dd	address_illegal
			dd	address_illegal
			dd	address_illegal



;****************************************************************************
;* Opcodes ******************************************************************
;****************************************************************************
ap__opcodes:		dd	opcode_illegal
			dd	opcode_adc
			dd	opcode_and
			dd	opcode_asl
			dd	opcode_bcc
			dd	opcode_bcs
			dd	opcode_beq
			dd	opcode_bit
			dd	opcode_bmi
			dd	opcode_bne
			dd	opcode_bpl
			dd	opcode_brk
			dd	opcode_bvc
			dd	opcode_bvs
			dd	opcode_clc
			dd	opcode_cld
			dd	opcode_cli
			dd	opcode_clv
			dd	opcode_cmp
			dd	opcode_cpx
			dd	opcode_cpy
			dd	opcode_dec
			dd	opcode_dex
			dd	opcode_dey
			dd	opcode_eor
			dd	opcode_inc
			dd	opcode_inx
			dd	opcode_iny
			dd	opcode_jmp
			dd	opcode_jsr
			dd	opcode_lda
			dd	opcode_ldx
			dd	opcode_ldy
			dd	opcode_lsr
			dd	opcode_nop
			dd	opcode_ora
			dd	opcode_pha
			dd	opcode_php
			dd	opcode_pla
			dd	opcode_plp
			dd	opcode_rol
			dd	opcode_ror
			dd	opcode_rti
			dd	opcode_rts
			dd	opcode_sbc
			dd	opcode_sec
			dd	opcode_sed
			dd	opcode_sei
			dd	opcode_sta
			dd	opcode_stx
			dd	opcode_sty
			dd	opcode_tax
			dd	opcode_tay
			dd	opcode_tsx
			dd	opcode_txa
			dd	opcode_txs
			dd	opcode_tya





;****************************************************************************
;****************************************************************************
;*
;* PART 2: Routines for CPU State Information and Loading a CPU State
;*
;****************************************************************************
;****************************************************************************




%ifdef CPU65XX_DEBUG
;****************************************************************************
;* cpu65xx_debug ************************************************************
;****************************************************************************
section .text
cpu65xx_debug:
			push	dword eax
			push	dword ebx
			push	dword ecx
			push	dword edx
			mov	word  ax, [PROGRAM_COUNTER]
			mov	word  bx, [ABSOLUTE_ADDRESS]
			mov	word  cx, [ZEROPAGE_ADDRESS]
			mov	word  [.ab__status], ax
			mov	word  [.ab__status + 2], bx
			mov	word  [.ab__status + 4], cx
			mov	byte  al, [OPCODE]
			mov	byte  bl, [STACK_POINTER]
			mov	byte  cl, [FLAG_REGISTER]
			mov	byte  dl, [CYCLE]
			mov	byte  [.ab__status + 8], al
			mov	byte  [.ab__status + 9], bl
			mov	byte  [.ab__status + 10], cl
			mov	byte  [.ab__status + 11], dl
			mov	byte  al, [ACCUMULATOR]
			mov	byte  bl, [INDEX_X]
			mov	byte  cl, [INDEX_Y]
			mov	byte  [.ab__status + 12], al
			mov	byte  [.ab__status + 13], bl
			mov	byte  [.ab__status + 14], cl
			SYS_WRITE  2, .ab__status, 16
			pop	dword edx
			pop	dword ecx
			pop	dword ebx
			pop	dword eax
			ret

section .data
			align 4
.ab__status:		times 16 db 0
%endif



;****************************************************************************
;* cpu65xx_dump *************************************************************
;****************************************************************************
section .text
cpu65xx_dump:
			cmp	dword [MODUL_ID], eax
			jne	near .exit
			push	dword eax
			push	dword ebx
			push	dword ecx
			push	dword edx
			WRITE_TEXT  .cb__registers, .registers_length
			WRITE_HEX  2, [ACCUMULATOR]
			WRITE_SPACE  1
			WRITE_HEX  2, [INDEX_X]
			WRITE_SPACE  1
			WRITE_HEX  2, [INDEX_Y]
			WRITE_SPACE  1
			WRITE_HEX  2, [STACK_POINTER]
			WRITE_SPACE  1
			WRITE_HEX  4, [PROGRAM_COUNTER]
			WRITE_SPACE  1
			WRITE_FLAG  8, [FLAG_REGISTER]
			WRITE_SPACE  1
			WRITE_HEX  2, [OPCODE]
			WRITE_SPACE  1
			WRITE_HEX  2, [CYCLE]
			WRITE_SPACE  1
			WRITE_HEX  8, [MODUL_ID]
			WRITE_SPACE  1
			WRITE_FLAG  32, [STATE]
			WRITE_RETURN
			pop	dword edx
			pop	dword ecx
			pop	dword ebx
			pop	dword eax
.exit:			ret

section .data
.cb__registers:		db	"-<CPU65XX>---------------------------------------------------------------", 00ah
			db	"AC XR YR SP _PC_ NV-BDIZC OP CY MODUL_ID --------RNI-----DNI-4321------AO", 00ah
.registers_length:	equ	$ - .cb__registers



;****************************************************************************
;* cpu65xx_initialize *******************************************************
;****************************************************************************
section .text
cpu65xx_initialize:
			ret



;****************************************************************************
;* cpu65xx_exit *************************************************************
;****************************************************************************
section .text
cpu65xx_exit:
			ret



;****************************************************************************
;* cpu65xx_save_snapshot ****************************************************
;****************************************************************************
section .text
cpu65xx_save_snapshot:
			ret



;****************************************************************************
;* cpu65xx_load_snapshot ****************************************************
;****************************************************************************
section .text
cpu65xx_load_snapshot:
			ret





;****************************************************************************
;****************************************************************************
;*
;* PART 3: Handlers for external Signals
;*
;****************************************************************************
;****************************************************************************





;****************************************************************************
;* cpu65xx_reset ************************************************************
;****************************************************************************
section .text
cpu65xx_reset:		
			mov	byte  [FLAG_REGISTER], (CPU65XX_FLAG_R | CPU65XX_FLAG_B | CPU65XX_FLAG_I)
			mov	dword ebx, 00000fffch
			call	dword [MEMORY_READ]
			inc	dword ebx
			mov	byte  [PROGRAM_COUNTER], al
			call	dword [MEMORY_READ]
			xor	dword ebx, ebx
			mov	byte  [PROGRAM_COUNTER + 1], al
			mov	dword [ABSOLUTE_ADDRESS], ebx
			xor	dword eax, eax
			mov	dword [ZEROPAGE_ADDRESS], ebx
			or	dword eax, CPU65XX_PROCESSING_RESET
			mov	byte  [CYCLE], bl
			mov	dword [STATE], eax
			mov	byte  [OPCODE], bl
			mov	dword eax, get_opcode
			inc	byte  bh
			mov	dword [EXECUTE_CYCLE], eax
			mov	dword [STACK_POINTER], ebx
			ret



;****************************************************************************
;* cpu65xx_cycle ************************************************************
;****************************************************************************
cpu65xx_cycle:
			CPU65XX_CALL_CYCLE
			ret



;****************************************************************************
;* cpu65xx_signal_set_irq ***************************************************
;****************************************************************************
cpu65xx_signal_set_irq:
			push	dword ebx
			push	dword ecx
			push	dword edx
			mov	byte  cl, al
			mov	dword edx, (CPU65XX_SIGNAL_IRQ1 / 2)
			mov	dword ebx, [STATE]
			shl	dword edx, cl
			and	dword edx, (CPU65XX_SIGNAL_IRQ4 | CPU65XX_SIGNAL_IRQ3 | CPU65XX_SIGNAL_IRQ2 | CPU65XX_SIGNAL_IRQ1)
			;if result zero ==> internal error
			or	dword edx, CPU65XX_SIGNAL_IRQ
			mov	byte  cl, [CYCLE]
			or	dword ebx, edx
			mov	byte  [IRQ_CYCLE], cl
			pop	dword edx
			mov	dword [STATE], ebx
			pop	dword ecx
			pop	dword ebx
			ret



;****************************************************************************
;* cpu65xx_signal_clear_irq *************************************************
;****************************************************************************
cpu65xx_signal_clear_irq:
			push	dword ebx
			push	dword ecx
			push	dword edx
			mov	byte  cl, al
			mov	dword edx, (CPU65XX_SIGNAL_IRQ1 / 2)
			mov	dword ebx, [STATE]
			shl	dword edx, cl
			and	dword edx, (CPU65XX_SIGNAL_IRQ4 | CPU65XX_SIGNAL_IRQ3 | CPU65XX_SIGNAL_IRQ2 | CPU65XX_SIGNAL_IRQ1)
			;if result zero ==> internal error
			not	dword edx
			and	dword ebx, edx
			test	dword ebx, (CPU65XX_SIGNAL_IRQ4 | CPU65XX_SIGNAL_IRQ3 | CPU65XX_SIGNAL_IRQ2 | CPU65XX_SIGNAL_IRQ1)
			jnz	.other_irq_left
			and	dword ebx, (NOT32 - CPU65XX_SIGNAL_IRQ)
.other_irq_left:	pop	dword edx
			mov	dword [STATE], ebx
			pop	dword ecx
			pop	dword ebx
			ret



;****************************************************************************
;* cpu65xx_signal_set_nmi ***************************************************
;****************************************************************************
cpu65xx_signal_set_nmi:
			push	dword eax
			mov	byte  al, [CYCLE]
			or	dword [STATE], CPU65XX_SIGNAL_NMI
			mov	byte  [NMI_CYCLE], al
			pop	dword eax
			ret



;****************************************************************************
;* cpu65xx_signal_clear_nmi *************************************************
;****************************************************************************
;* the nmi handler clears this itself
cpu65xx_signal_clear_nmi:
			and	dword [STATE], (NOT32 - CPU65XX_SIGNAL_NMI)
			ret



;****************************************************************************
;* cpu65xx_signal_set_dma ***************************************************
;****************************************************************************
cpu65xx_signal_set_dma:	
			push	dword eax
			push	dword ebx
			mov	dword eax, [MEMORY_READ]
			mov	dword ebx, .ignore_memory_read
			or	dword [STATE], CPU65XX_SIGNAL_DMA
			mov	dword [MEMORY_READ], ebx
			mov	dword [MEMORY_READ_TEMP], eax
			pop	dword ebx
			pop	dword eax
			ret

.ignore_memory_read:	pop	dword eax
			ret



;****************************************************************************
;* cpu65xx_signal_clear_dma *************************************************
;****************************************************************************
cpu65xx_signal_clear_dma:
			push	dword eax
			mov	dword eax, [MEMORY_READ_TEMP]
			and	dword [STATE], (NOT32 - CPU65XX_SIGNAL_DMA)
			mov	dword [MEMORY_READ], eax
			pop	dword eax
			ret



;****************************************************************************
;* cpu65xx_signal_set_overflow **********************************************
;****************************************************************************
cpu65xx_signal_set_overflow:
			or	byte  [FLAG_REGISTER], CPU65XX_FLAG_V
			ret





;****************************************************************************
;****************************************************************************
;*
;* PART 4: CPU Emulation Code
;*
;****************************************************************************
;****************************************************************************





;****************************************************************************
;****************************************************************************
;*
;* PART 4.1: Adressing Modes
;*
;****************************************************************************
;****************************************************************************




;****************************************************************************
;* DMA Macro ****************************************************************
;****************************************************************************
%macro CHECK_FOR_DMA 0
			test	dword [STATE], CPU65XX_SIGNAL_DMA
			jz	%%no_dma
			ret
%%no_dma:
%endmacro



;****************************************************************************
;* Implicit *****************************************************************
;****************************************************************************
address_implicit:
			jmp	dword [OPCODE_ADDRESS]




;****************************************************************************
;* Accumulator **************************************************************
;****************************************************************************
address_accumulator:
			mov	byte  al, [ACCUMULATOR]
			jmp	dword [OPCODE_ADDRESS]

.rmw:			mov	dword ebx, get_opcode
			mov	byte  [ACCUMULATOR], al
			mov	dword [EXECUTE_CYCLE], ebx
			ret



;****************************************************************************
;* Immediate ****************************************************************
;****************************************************************************
address_immediate:
			mov	dword ebx, [PROGRAM_COUNTER]
			call	dword [MEMORY_READ]
			inc	word  bx
			mov	dword [PROGRAM_COUNTER], ebx
			jmp	dword [OPCODE_ADDRESS]



;****************************************************************************
;* Zeropage *****************************************************************
;****************************************************************************
address_zeropage:
			mov	dword ebx, [PROGRAM_COUNTER]
			call	dword [MEMORY_READ]
			inc	word  bx
			mov	byte  [ZEROPAGE_ADDRESS], al
			mov	dword ecx, [RW_ADDRESS]
			mov	dword [PROGRAM_COUNTER], ebx
			mov	dword [EXECUTE_CYCLE], ecx
			ret

.x:			mov	dword ebx, [PROGRAM_COUNTER]
			call	dword [MEMORY_READ]
			add	byte  al, [INDEX_X]
			inc	word  bx
			mov	byte  [ZEROPAGE_ADDRESS], al
			mov	dword ecx, .delay
			mov	dword [PROGRAM_COUNTER], ebx
			mov	dword [EXECUTE_CYCLE], ecx
			ret

.y:			mov	dword ebx, [PROGRAM_COUNTER]
			call	dword [MEMORY_READ]
			add	byte  al, [INDEX_Y]
			inc	word  bx
			mov	byte  [ZEROPAGE_ADDRESS], al
			mov	dword ecx, .delay
			mov	dword [PROGRAM_COUNTER], ebx
			mov	dword [EXECUTE_CYCLE], ecx
			ret

.delay:			CHECK_FOR_DMA
			mov	dword eax, [RW_ADDRESS]
			mov	dword [EXECUTE_CYCLE], eax
			ret

.read:			mov	dword ebx, [ZEROPAGE_ADDRESS]
			call	dword [MEMORY_READ]
			jmp	dword [OPCODE_ADDRESS]

.write:			mov	dword ebx, [ZEROPAGE_ADDRESS]
			jmp	dword [OPCODE_ADDRESS]

.rmw:			mov	dword ebx, .rmw_1
			mov	byte  [ACCUMULATOR_TEMP], al
			mov	dword [EXECUTE_CYCLE], ebx
			ret

			;---------------------------------------
			;at this cycle the CPU is in write state
			;---------------------------------------

.rmw_1:			;CHECK_FOR_DMA
			mov	dword [EXECUTE_CYCLE], .rmw_2
			ret

.rmw_2:			mov	dword ebx, [ZEROPAGE_ADDRESS]
			mov	dword ecx, get_opcode
			mov	byte  al, [ACCUMULATOR_TEMP]
			call	dword [MEMORY_WRITE]
			mov	dword [EXECUTE_CYCLE], ecx
			ret



;****************************************************************************
;* Absolute *****************************************************************
;****************************************************************************
address_absolute:
			mov	dword ebx, [PROGRAM_COUNTER]
			call	dword [MEMORY_READ]
			inc	word  bx
			mov	byte  [ABSOLUTE_ADDRESS], al
			mov	dword ecx, ._1
			mov	dword [PROGRAM_COUNTER], ebx
			mov	dword [EXECUTE_CYCLE], ecx
			ret

._1:			mov	dword ebx, [PROGRAM_COUNTER]
			call	dword [MEMORY_READ]
			inc	word  bx
			mov	byte  [ABSOLUTE_ADDRESS + 1], al
			mov	dword ecx, [RW_ADDRESS]
			mov	dword [PROGRAM_COUNTER], ebx
			mov	dword [EXECUTE_CYCLE], ecx
			ret

.x:			mov	dword ebx, [PROGRAM_COUNTER]
			call	dword [MEMORY_READ]
			add	byte  al, [INDEX_X]
			inc	word  bx
			setc	byte  ah
			mov	word  [ABSOLUTE_ADDRESS], ax
			mov	dword ecx, .index
			mov	dword [PROGRAM_COUNTER], ebx
			mov	dword [EXECUTE_CYCLE], ecx
			ret

.y:			mov	dword ebx, [PROGRAM_COUNTER]
			call	dword [MEMORY_READ]
			add	byte  al, [INDEX_Y]
			inc	word  bx
			setc	byte  ah
			mov	word  [ABSOLUTE_ADDRESS], ax
			mov	dword ecx, .index
			mov	dword [PROGRAM_COUNTER], ebx
			mov	dword [EXECUTE_CYCLE], ecx
			ret

.index:			mov	dword ebx, [PROGRAM_COUNTER]
			call	dword [MEMORY_READ]
			inc	word  bx
			mov	dword ecx, [DELAY_ADDRESS]
			test	byte  [ABSOLUTE_ADDRESS + 1], 0ffh
			jz	.index_1
			inc	byte  al
			mov	dword ecx, .delay
.index_1:		mov	byte  [ABSOLUTE_ADDRESS + 1], al
			mov	dword [PROGRAM_COUNTER], ebx
			mov	dword [EXECUTE_CYCLE], ecx
			ret

.delay:			CHECK_FOR_DMA
			mov	dword eax, [RW_ADDRESS]
			mov	dword [EXECUTE_CYCLE], eax
			ret

.read:			mov	dword ebx, [ABSOLUTE_ADDRESS]
			call	dword [MEMORY_READ]
			jmp	dword [OPCODE_ADDRESS]

.write:			mov	dword ebx, [ABSOLUTE_ADDRESS]
			jmp	dword [OPCODE_ADDRESS]

.rmw:			mov	dword ebx, .rmw_1
			mov	byte  [ACCUMULATOR_TEMP], al
			mov	dword [EXECUTE_CYCLE], ebx
			ret

			;---------------------------------------
			;at this cycle the CPU is in write state
			;---------------------------------------

.rmw_1:			;CHECK_FOR_DMA
			mov	dword [EXECUTE_CYCLE], .rmw_2
			ret

.rmw_2:			mov	dword ebx, [ABSOLUTE_ADDRESS]
			mov	dword ecx, get_opcode
			mov	byte  al, [ACCUMULATOR_TEMP]
			call	dword [MEMORY_WRITE]
			mov	dword [EXECUTE_CYCLE], ecx
			ret



;****************************************************************************
;* Indirect *****************************************************************
;****************************************************************************
address_indirect:
			mov	dword ebx, [PROGRAM_COUNTER]
			call	dword [MEMORY_READ]
			inc	word  bx
			mov	byte  [ABSOLUTE_ADDRESS], al
			mov	dword ecx, ._1
			mov	dword [PROGRAM_COUNTER], ebx
			mov	dword [EXECUTE_CYCLE], ecx
			ret

._1:			mov	dword ebx, [PROGRAM_COUNTER]
			call	dword [MEMORY_READ]
			inc	word  bx
			mov	byte  [ABSOLUTE_ADDRESS + 1], al
			mov	dword ecx, ._2
			mov	dword [PROGRAM_COUNTER], ebx
			mov	dword [EXECUTE_CYCLE], ecx
			ret

._2:			mov	dword ebx, [ABSOLUTE_ADDRESS]
			call	dword [MEMORY_READ]
			mov	dword ecx, ._3
			mov	byte  [PROGRAM_COUNTER], al
			mov	dword [EXECUTE_CYCLE], ecx
			ret

._3:			mov	dword ebx, [ABSOLUTE_ADDRESS]
			mov	dword ecx, get_opcode
			inc	word  bx
			call	dword [MEMORY_READ]
			mov	byte  [PROGRAM_COUNTER + 1], al
			mov	dword [EXECUTE_CYCLE], ecx
			ret



;****************************************************************************
;* X Indirect ***************************************************************
;****************************************************************************
address_x_indirect:
			mov	dword ebx, [PROGRAM_COUNTER]
			call	dword [MEMORY_READ]
			add	byte  al, [INDEX_X]
			inc	word  bx
			mov	dword ecx, ._1
			mov	byte  [ZEROPAGE_ADDRESS], al
			mov	dword [PROGRAM_COUNTER], ebx
			mov	dword [EXECUTE_CYCLE], ecx
			ret

._1:			CHECK_FOR_DMA
			mov	dword [EXECUTE_CYCLE], ._2
			ret
			
._2:			mov	dword ebx, [ZEROPAGE_ADDRESS]
			call	dword [MEMORY_READ]
			mov	dword ecx, ._3
			mov	byte  [ABSOLUTE_ADDRESS], al
			mov	dword [EXECUTE_CYCLE], ecx
			ret

._3:			mov	dword ebx, [ZEROPAGE_ADDRESS]
			mov	dword ecx, [RW_ADDRESS]
			inc	byte  bl
			call	dword [MEMORY_READ]
			mov	byte  [ABSOLUTE_ADDRESS + 1], al
			mov	dword [EXECUTE_CYCLE], ecx
			ret

.read:			mov	dword ebx, [ABSOLUTE_ADDRESS]
			call	dword [MEMORY_READ]
			jmp	dword [OPCODE_ADDRESS]

.write:			mov	dword ebx, [ABSOLUTE_ADDRESS]
			jmp	dword [OPCODE_ADDRESS]



;****************************************************************************
;* Indirect Y ***************************************************************
;****************************************************************************
address_indirect_y:
			mov	dword ebx, [PROGRAM_COUNTER]
			call	dword [MEMORY_READ]
			inc	word  bx
			mov	byte  [ZEROPAGE_ADDRESS], al
			mov	dword ecx, ._1
			mov	dword [PROGRAM_COUNTER], ebx
			mov	dword [EXECUTE_CYCLE], ecx
			ret

._1:			mov	dword ebx, [ZEROPAGE_ADDRESS]
			call	dword [MEMORY_READ]
			mov	dword ecx, ._2
			mov	byte  [ABSOLUTE_ADDRESS], al
			mov	dword [EXECUTE_CYCLE], ecx
			ret

._2:			mov	dword ebx, [ZEROPAGE_ADDRESS]
			mov	dword ecx, [DELAY_ADDRESS]
			inc	byte  bl
			call	dword [MEMORY_READ]
			mov	byte  ah, al
			mov	byte  al, [ABSOLUTE_ADDRESS]
			add	byte  al, [INDEX_Y]
			jnc	._3
			inc	byte  ah
			mov	dword ecx, .delay
._3:			mov	word  [ABSOLUTE_ADDRESS], ax
			mov	dword [EXECUTE_CYCLE], ecx
			ret

.delay:			CHECK_FOR_DMA
			mov	dword eax, [RW_ADDRESS]
			mov	dword [EXECUTE_CYCLE], eax
			ret

.read:			mov	dword ebx, [ABSOLUTE_ADDRESS]
			call	dword [MEMORY_READ]
			jmp	dword [OPCODE_ADDRESS]

.write:			mov	dword ebx, [ABSOLUTE_ADDRESS]
			jmp	dword [OPCODE_ADDRESS]



;****************************************************************************
;* Relative *****************************************************************
;****************************************************************************
address_relative:
			mov	dword ebx, [PROGRAM_COUNTER]
			call	dword [MEMORY_READ]
			inc	word  bx
			mov	dword [PROGRAM_COUNTER], ebx
			jmp	dword [OPCODE_ADDRESS]


;****************************************************************************
;* address_illegal **********************************************************
;****************************************************************************
address_illegal:
			or	dword [STATE], CPU65XX_ILLEGAL_ADDRESS_MODE
			mov	dword eax, [MODUL_ID]
			jmp	main_exception






;****************************************************************************
;****************************************************************************
;*
;* PART 4.2: Opcodes
;*
;****************************************************************************
;****************************************************************************






;****************************************************************************
;* opcode_adc ***************************************************************
;****************************************************************************
opcode_adc:
			test	byte  [FLAG_REGISTER], CPU65XX_FLAG_D
			jz	.no_decimal
.decimal:		test	byte  [FLAG_REGISTER], CPU65XX_FLAG_C
			jz	.decimal_1
			stc
.decimal_1:		adc	byte  al, [ACCUMULATOR]
			daa
			mov	byte  [ACCUMULATOR], al
			setc	byte  cl
			jmp	opcode_sbc.common
.no_decimal:		test	byte  [FLAG_REGISTER], CPU65XX_FLAG_C
			jz	.no_decimal_1
			stc
.no_decimal_1:		adc	byte  [ACCUMULATOR], al
			setc	byte  cl
			jmp	opcode_sbc.common



;****************************************************************************
;* opcode_sbc ***************************************************************
;****************************************************************************
opcode_sbc:
			test	byte  [FLAG_REGISTER], CPU65XX_FLAG_D
			jz	.no_decimal
.decimal:		test	byte  [FLAG_REGISTER], CPU65XX_FLAG_C
			jz	.decimal_1
			stc
.decimal_1:		cmc
			mov	byte  cl, al
			mov	byte  al, [ACCUMULATOR]
			sbb	byte  al, cl
			das
			mov	byte  [ACCUMULATOR], al
			setnc	byte  cl
			jmp	.common
.no_decimal:		test	byte  [FLAG_REGISTER], CPU65XX_FLAG_C
			jz	.no_decimal_1
			stc
.no_decimal_1:		cmc
			sbb	byte  [ACCUMULATOR], al
			setnc	byte  cl
.common:		setz	byte  ch
			seto	byte  dl
			sets	byte  dh
			shl	byte  ch, CPU65XX_FLAG_Z_BIT
			mov	byte  bl, [FLAG_REGISTER]
			shl	byte  dl, CPU65XX_FLAG_V_BIT
			shl	byte  dh, CPU65XX_FLAG_N_BIT
			or	byte  cl, ch
			or	byte  dl, dh
			and	byte  bl, (NOT8 - CPU65XX_FLAG_N - CPU65XX_FLAG_V - CPU65XX_FLAG_Z - CPU65XX_FLAG_C)
			or	byte  cl, dl
			or	byte  bl, cl
			mov	dword eax, get_opcode
			mov	byte  [FLAG_REGISTER], bl
			mov	dword [EXECUTE_CYCLE], eax
			ret



;****************************************************************************
;* opcode_cmp ***************************************************************
;****************************************************************************
opcode_cmp:
			mov	byte  ah, [ACCUMULATOR]
			jmp	opcode_cpy.common


;****************************************************************************
;* opcode_cpx ***************************************************************
;****************************************************************************
opcode_cpx:
			mov	byte  ah, [INDEX_X]
			jmp	opcode_cpy.common



;****************************************************************************
;* opcode_cpy ***************************************************************
;****************************************************************************
opcode_cpy:
			mov	byte  ah, [INDEX_Y]
.common:		sub	byte  ah, al
			setnc	byte  cl
			setz	byte  ch
			sets	byte  dl
			shl	byte  ch, CPU65XX_FLAG_Z_BIT
			mov	byte  bl, [FLAG_REGISTER]
			shl	byte  dl, CPU65XX_FLAG_N_BIT
			or	byte  cl, ch
			and	byte  bl, (NOT8 - CPU65XX_FLAG_N - CPU65XX_FLAG_Z - CPU65XX_FLAG_C)
			or	byte  cl, dl
			or	byte  bl, cl
			mov	dword eax, get_opcode
			mov	byte  [FLAG_REGISTER], bl
			mov	dword [EXECUTE_CYCLE], eax
			ret



;****************************************************************************
;* opcode_and ***************************************************************
;****************************************************************************
opcode_and:
			and	byte  [ACCUMULATOR], al
			jmp	opcode_ora.common



;****************************************************************************
;* opcode_eor ***************************************************************
;****************************************************************************
opcode_eor:
			xor	byte [ACCUMULATOR], al
			jmp	opcode_ora.common



;****************************************************************************
;* opcode_ora ***************************************************************
;****************************************************************************
opcode_ora:
			or	byte  [ACCUMULATOR], al
.common:		setz	byte  cl
			sets	byte  dl
			shl	byte  cl, CPU65XX_FLAG_Z_BIT
			mov	byte  bl, [FLAG_REGISTER]
			shl	byte  dl, CPU65XX_FLAG_N_BIT
			and	byte  bl, (NOT8 - CPU65XX_FLAG_N - CPU65XX_FLAG_Z)
			or	byte  cl, dl
			or	byte  bl, cl
			mov	dword eax, get_opcode
			mov	byte  [FLAG_REGISTER], bl
			mov	dword [EXECUTE_CYCLE], eax
			ret



;****************************************************************************
;* opcode_asl ***************************************************************
;****************************************************************************
opcode_asl:
			shl	byte  al, 1
			jmp	opcode_ror.common



;****************************************************************************
;* opcode_lsr ***************************************************************
;****************************************************************************
opcode_lsr:
			shr	byte  al, 1
			jmp	opcode_ror.common


;****************************************************************************
;* opcode_rol ***************************************************************
;****************************************************************************
opcode_rol:
			test	byte  [FLAG_REGISTER], CPU65XX_FLAG_C
			jz	._1
			stc
._1:			rcl	byte  al, 1
			jmp	opcode_ror.common


;****************************************************************************
;* opcode_ror ***************************************************************
;****************************************************************************
opcode_ror:
			test	byte  [FLAG_REGISTER], CPU65XX_FLAG_C
			jz	._1
			stc
._1:			rcr	byte  al, 1
.common:		setc	byte  cl
			test	byte  al, al
			setz	byte  ch
			sets	byte  dl
			shl	byte  ch, CPU65XX_FLAG_Z_BIT
			mov	byte  bl, [FLAG_REGISTER]
			shl	byte  dl, CPU65XX_FLAG_N_BIT
			or	byte  cl, ch
			and	byte  bl, (NOT8 - CPU65XX_FLAG_N - CPU65XX_FLAG_Z - CPU65XX_FLAG_C)
			or	byte  cl, dl
			or	byte  bl, cl
			mov	byte  [FLAG_REGISTER], bl
			jmp	dword [RMW_ADDRESS]



;****************************************************************************
;* opcode_dec ***************************************************************
;****************************************************************************
opcode_dec:
			dec	byte  al
			jmp	opcode_inc.common


;****************************************************************************
;* opcode_inc ***************************************************************
;****************************************************************************
opcode_inc:
			inc	byte  al
.common:		setz	byte  cl
			sets	byte  dl
			shl	byte  cl, CPU65XX_FLAG_Z_BIT
			mov	byte  bl, [FLAG_REGISTER]
			shl	byte  dl, CPU65XX_FLAG_N_BIT
			and	byte  bl, (NOT8 - CPU65XX_FLAG_N - CPU65XX_FLAG_Z)
			or	byte  cl, dl
			or	byte  bl, cl
			mov	byte  [FLAG_REGISTER], bl
			jmp	dword [RMW_ADDRESS]



;****************************************************************************
;* opcode_dex ***************************************************************
;****************************************************************************
opcode_dex:
			dec	byte  [INDEX_X]
			jmp	opcode_iny.common



;****************************************************************************
;* opcode_dey ***************************************************************
;****************************************************************************
opcode_dey:
			dec	byte  [INDEX_Y]
			jmp	opcode_iny.common



;****************************************************************************
;* opcode_inx ***************************************************************
;****************************************************************************
opcode_inx:
			inc	byte  [INDEX_X]
			jmp	opcode_iny.common


;****************************************************************************
;* opcode_iny ***************************************************************
;****************************************************************************
opcode_iny:
			inc	byte  [INDEX_Y]
.common:		setz	byte  cl
			sets	byte  dl
			shl	byte  cl, CPU65XX_FLAG_Z_BIT
			mov	byte  bl, [FLAG_REGISTER]
			shl	byte  dl, CPU65XX_FLAG_N_BIT
			and	byte  bl, (NOT8 - CPU65XX_FLAG_N - CPU65XX_FLAG_Z)
			or	byte  cl, dl
			mov	dword eax, get_opcode
			or	byte  bl, cl
			mov	dword [EXECUTE_CYCLE], eax
			mov	byte  [FLAG_REGISTER], bl
			ret



;****************************************************************************
;* opcode_bcc ***************************************************************
;****************************************************************************
opcode_bcc:
			test	byte  [FLAG_REGISTER], CPU65XX_FLAG_C
			jz	opcode_bvs.common
			mov	dword [EXECUTE_CYCLE], get_opcode
			ret



;****************************************************************************
;* opcode_bcs ***************************************************************
;****************************************************************************
opcode_bcs:
			test	byte  [FLAG_REGISTER], CPU65XX_FLAG_C
			jnz	opcode_bvs.common
			mov	dword [EXECUTE_CYCLE], get_opcode
			ret



;****************************************************************************
;* opcode_beq ***************************************************************
;****************************************************************************
opcode_beq:
			test	byte  [FLAG_REGISTER], CPU65XX_FLAG_Z
			jnz	opcode_bvs.common
			mov	dword [EXECUTE_CYCLE], get_opcode
			ret



;****************************************************************************
;* opcode_bmi ***************************************************************
;****************************************************************************
opcode_bmi:
			test	byte  [FLAG_REGISTER], CPU65XX_FLAG_N
			jnz	opcode_bvs.common
			mov	dword [EXECUTE_CYCLE], get_opcode
			ret



;****************************************************************************
;* opcode_bne ***************************************************************
;****************************************************************************
opcode_bne:
			test	byte  [FLAG_REGISTER], CPU65XX_FLAG_Z
			jz	opcode_bvs.common
			mov	dword [EXECUTE_CYCLE], get_opcode
			ret



;****************************************************************************
;* opcode_bpl ***************************************************************
;****************************************************************************
opcode_bpl:
			test	byte  [FLAG_REGISTER], CPU65XX_FLAG_N
			jz	opcode_bvs.common
			mov	dword [EXECUTE_CYCLE], get_opcode
			ret



;****************************************************************************
;* opcode_bvc ***************************************************************
;****************************************************************************
opcode_bvc:
			test	byte  [FLAG_REGISTER], CPU65XX_FLAG_V
			jz	opcode_bvs.common
			mov	dword [EXECUTE_CYCLE], get_opcode
			ret



;****************************************************************************
;* opcode_bvs ***************************************************************
;****************************************************************************
opcode_bvs:
			test	byte  [FLAG_REGISTER], CPU65XX_FLAG_V
			jnz	.common
			mov	dword [EXECUTE_CYCLE], get_opcode
			ret
.common:		cbw
			mov	word  [ABSOLUTE_ADDRESS], ax
			mov	dword [EXECUTE_CYCLE], ._1
			ret

._1:			CHECK_FOR_DMA
			mov	dword eax, [ABSOLUTE_ADDRESS]
			mov	dword ebx, [PROGRAM_COUNTER]
			add	word  ax, bx
			mov	dword [PROGRAM_COUNTER], eax
			cmp	byte  ah, bh
			je	._3
			mov	dword [EXECUTE_CYCLE], ._2
			ret
._2:			CHECK_FOR_DMA
._3:			mov	dword [EXECUTE_CYCLE], get_opcode
			ret


;****************************************************************************
;* opcode_clc ***************************************************************
;****************************************************************************
opcode_clc:
			CHECK_FOR_DMA
			mov	byte  al, [FLAG_REGISTER]
			mov	dword ebx, get_opcode
			and	byte  al, (NOT8 - CPU65XX_FLAG_C)
			mov	dword [EXECUTE_CYCLE], ebx
			mov	byte  [FLAG_REGISTER], al
			ret



;****************************************************************************
;* opcode_cld ***************************************************************
;****************************************************************************
opcode_cld:
			CHECK_FOR_DMA
			mov	byte  al, [FLAG_REGISTER]
			mov	dword ebx, get_opcode
			and	byte  al, (NOT8 - CPU65XX_FLAG_D)
			mov	dword [EXECUTE_CYCLE], ebx
			mov	byte  [FLAG_REGISTER], al
			ret



;****************************************************************************
;* opcode_cli ***************************************************************
;****************************************************************************
opcode_cli:
			CHECK_FOR_DMA
			mov	byte  al, [FLAG_REGISTER]
			mov	dword ebx, get_opcode
			and	byte  al, (NOT8 - CPU65XX_FLAG_I)
			mov	dword [EXECUTE_CYCLE], ebx
			mov	byte  cl, [CYCLE]
			mov	byte  [FLAG_REGISTER], al
			mov	byte  [IRQ_CYCLE], cl
			ret



;****************************************************************************
;* opcode_clv ***************************************************************
;****************************************************************************
opcode_clv:
			CHECK_FOR_DMA
			mov	byte  al, [FLAG_REGISTER]
			mov	dword ebx, get_opcode
			and	byte  al, (NOT8 - CPU65XX_FLAG_V)
			mov	dword [EXECUTE_CYCLE], ebx
			mov	byte  [FLAG_REGISTER], al
			ret



;****************************************************************************
;* opcode_sec ***************************************************************
;****************************************************************************
opcode_sec:
			CHECK_FOR_DMA
			mov	byte  al, [FLAG_REGISTER]
			mov	dword ebx, get_opcode
			or	byte  al, CPU65XX_FLAG_C
			mov	dword [EXECUTE_CYCLE], ebx
			mov	byte  [FLAG_REGISTER], al
			ret



;****************************************************************************
;* opcode_sed ***************************************************************
;****************************************************************************
opcode_sed:
			CHECK_FOR_DMA
			mov	byte  al, [FLAG_REGISTER]
			mov	dword ebx, get_opcode
			or	byte  al, CPU65XX_FLAG_D
			mov	dword [EXECUTE_CYCLE], ebx
			mov	byte  [FLAG_REGISTER], al
			ret



;****************************************************************************
;* opcode_sei ***************************************************************
;****************************************************************************
opcode_sei:
			CHECK_FOR_DMA
			mov	byte  al, [FLAG_REGISTER]
			mov	dword ebx, get_opcode
			or	byte  al, CPU65XX_FLAG_I
			mov	dword [EXECUTE_CYCLE], ebx
			mov	byte  [FLAG_REGISTER], al
			ret


;****************************************************************************
;* opcode_tax ***************************************************************
;****************************************************************************
opcode_tax:
			mov	byte  al, [ACCUMULATOR]
			mov	byte  [INDEX_X], al
			jmp	opcode_tya.common



;****************************************************************************
;* opcode_tay ***************************************************************
;****************************************************************************
opcode_tay:
			mov	byte  al, [ACCUMULATOR]
			mov	byte  [INDEX_Y], al
			jmp	opcode_tya.common



;****************************************************************************
;* opcode_tsx ***************************************************************
;****************************************************************************
opcode_tsx:
			mov	byte  al, [STACK_POINTER]
			mov	byte  [INDEX_X], al
			jmp	opcode_tya.common



;****************************************************************************
;* opcode_txa ***************************************************************
;****************************************************************************
opcode_txa:
			mov	byte  al, [INDEX_X]
			mov	byte  [ACCUMULATOR], al
			jmp	opcode_tya.common



;****************************************************************************
;* opcode_txs ***************************************************************
;****************************************************************************
opcode_txs:
			mov	byte  al, [INDEX_X]
			mov	dword [EXECUTE_CYCLE], get_opcode
			mov	byte  [STACK_POINTER], al
			ret


;****************************************************************************
;* opcode_tya ***************************************************************
;****************************************************************************
opcode_tya:
			mov	byte  al, [INDEX_Y]
			mov	byte  [ACCUMULATOR], al
.common:		test	byte  al, al
			setz	byte  cl
			sets	byte  dl
			shl	byte  cl, CPU65XX_FLAG_Z_BIT
			mov	byte  bl, [FLAG_REGISTER]
			shl	byte  dl, CPU65XX_FLAG_N_BIT
			and	byte  bl, (NOT8 - CPU65XX_FLAG_N - CPU65XX_FLAG_Z)
			or	byte  cl, dl
			mov	dword eax, get_opcode
			or	byte  bl, cl
			mov	dword [EXECUTE_CYCLE], eax
			mov	byte  [FLAG_REGISTER], bl
			ret



;****************************************************************************
;* opcode_lda ***************************************************************
;****************************************************************************
opcode_lda:
			mov	byte  [ACCUMULATOR], al
			jmp	opcode_ldy.common


;****************************************************************************
;* opcode_ldx ***************************************************************
;****************************************************************************
opcode_ldx:
			mov	byte  [INDEX_X], al
			jmp	opcode_ldy.common


;****************************************************************************
;* opcode_ldy ***************************************************************
;****************************************************************************
opcode_ldy:
			mov	byte  [INDEX_Y], al
.common:		test	byte  al, al
			setz	byte  cl
			sets	byte  dl
			shl	byte  cl, CPU65XX_FLAG_Z_BIT
			mov	byte  bl, [FLAG_REGISTER]
			shl	byte  dl, CPU65XX_FLAG_N_BIT
			and	byte  bl, (NOT8 - CPU65XX_FLAG_N - CPU65XX_FLAG_Z)
			or	byte  cl, dl
			mov	dword eax, get_opcode
			or	byte  bl, cl
			mov	dword [EXECUTE_CYCLE], eax
			mov	byte  [FLAG_REGISTER], bl
			ret


;****************************************************************************
;* opcode_sta ***************************************************************
;****************************************************************************
opcode_sta:
			mov	byte  al, [ACCUMULATOR]
			call	dword [MEMORY_WRITE]
			mov	dword [EXECUTE_CYCLE], get_opcode
			ret



;****************************************************************************
;* opcode_stx ***************************************************************
;****************************************************************************
opcode_stx:
			mov	byte  al, [INDEX_X]
			call	dword [MEMORY_WRITE]
			mov	dword [EXECUTE_CYCLE], get_opcode
			ret



;****************************************************************************
;* opcode_sty ***************************************************************
;****************************************************************************
opcode_sty:
			mov	byte  al, [INDEX_Y]
			call	dword [MEMORY_WRITE]
			mov	dword [EXECUTE_CYCLE], get_opcode
			ret



;****************************************************************************
;* opcode_jmp ***************************************************************
;****************************************************************************
opcode_jmp:
			mov	dword ebx, [PROGRAM_COUNTER]
			call	dword [MEMORY_READ]
			mov	dword ecx, ._1
			mov	byte  [ABSOLUTE_ADDRESS], al
			mov	dword [EXECUTE_CYCLE], ecx
			ret

._1:			mov	dword ebx, [PROGRAM_COUNTER]
			mov	dword ecx, get_opcode
			inc	word  bx
			call	dword [MEMORY_READ]
			mov	byte  ah, al
			inc	word  bx
			mov	byte  al, [ABSOLUTE_ADDRESS]
			mov	dword [PROGRAM_COUNTER], ebx
			mov	dword [EXECUTE_CYCLE], ecx
			mov	word  [PROGRAM_COUNTER], ax
			ret


;****************************************************************************
;* opcode_jsr ***************************************************************
;****************************************************************************
opcode_jsr:
			mov	dword ebx, [PROGRAM_COUNTER]
			call	dword [MEMORY_READ]
			inc	word  bx
			mov	dword ecx, ._1
			mov	dword [PROGRAM_COUNTER], ebx
			mov	byte  [ABSOLUTE_ADDRESS], al
			mov	dword [EXECUTE_CYCLE], ecx
			ret

._1:			mov	dword ebx, [PROGRAM_COUNTER]
			call	dword [MEMORY_READ]
			mov	dword ecx, ._2
			mov	byte  [ABSOLUTE_ADDRESS + 1], al
			mov	dword [EXECUTE_CYCLE], ecx
			ret

._2:			mov	dword ebx, [STACK_POINTER]
			mov	dword ecx, ._3
			mov	byte  al, [PROGRAM_COUNTER + 1]
			call	dword [MEMORY_WRITE]
			dec	byte  bl
			mov	dword [EXECUTE_CYCLE], ecx
			mov	dword [STACK_POINTER], ebx
			ret

._3:			mov	dword ebx, [STACK_POINTER]
			mov	dword ecx, ._4
			mov	byte  al, [PROGRAM_COUNTER]
			call	dword [MEMORY_WRITE]
			dec	byte  bl
			mov	dword [EXECUTE_CYCLE], ecx
			mov	dword [STACK_POINTER], ebx
			ret

._4:			CHECK_FOR_DMA
			mov	dword eax, [ABSOLUTE_ADDRESS]
			mov	dword [EXECUTE_CYCLE], get_opcode
			mov	dword [PROGRAM_COUNTER], eax
			ret


;****************************************************************************
;* opcode_bit ***************************************************************
;****************************************************************************
opcode_bit:
			mov	byte  bl, [FLAG_REGISTER]
			and	byte  bl, (3fh - CPU65XX_FLAG_Z)
			test	byte  [ACCUMULATOR], al
			jnz	.bit1
			or	byte  bl, CPU65XX_FLAG_Z
.bit1:			and	byte  al, 0c0h
			or	byte  bl, al
			mov	dword [EXECUTE_CYCLE], get_opcode
			mov	byte  [FLAG_REGISTER], bl
			ret



;****************************************************************************
;* opcode_brk ***************************************************************
;****************************************************************************
opcode_brk:
			CHECK_FOR_DMA
			inc	word  [PROGRAM_COUNTER]
			mov	dword [EXECUTE_CYCLE], ._1
			ret

._1:			mov	dword ebx, [STACK_POINTER]
			mov	byte  al, [PROGRAM_COUNTER + 1]
			call	dword [MEMORY_WRITE]
			dec	byte  bl
			mov	dword [EXECUTE_CYCLE], ._2
			mov	dword [STACK_POINTER], ebx
			ret

._2:			mov	dword ebx, [STACK_POINTER]
			mov	byte  al, [PROGRAM_COUNTER]
			call	dword [MEMORY_WRITE]
			dec	byte  bl
			mov	dword [EXECUTE_CYCLE], ._3
			mov	dword [STACK_POINTER], ebx
			ret

._3:			mov	dword ebx, [STACK_POINTER]
			mov	byte  al, [FLAG_REGISTER]
			call	dword [MEMORY_WRITE]
			dec	byte  bl
			mov	dword [EXECUTE_CYCLE], ._4
			mov	dword [STACK_POINTER], ebx
			ret

._4:			mov	dword ebx, 00000fffeh
			call	dword [MEMORY_READ]
			mov	byte  [PROGRAM_COUNTER], al
			mov	dword [EXECUTE_CYCLE], ._5
			ret

._5:			mov	dword ebx, 00000ffffh
			call	dword [MEMORY_READ]
			mov	byte  bl, [FLAG_REGISTER]
			mov	byte  [PROGRAM_COUNTER + 1], al
			or	byte  bl, (CPU65XX_FLAG_B | CPU65XX_FLAG_I)
			mov	dword [EXECUTE_CYCLE], get_opcode
			mov	byte  [FLAG_REGISTER], bl
			ret



;****************************************************************************
;* opcode_nop ***************************************************************
;****************************************************************************
opcode_nop:
			mov	dword [EXECUTE_CYCLE], get_opcode
			ret



;****************************************************************************
;* opcode_pha ***************************************************************
;****************************************************************************
opcode_pha:
			mov	dword ebx, [STACK_POINTER]
			mov	byte  al, [ACCUMULATOR]
			call	dword [MEMORY_WRITE]
			dec	byte  bl
			mov	dword [EXECUTE_CYCLE], ._1
			mov	dword [STACK_POINTER], ebx
			ret

._1:			CHECK_FOR_DMA
			mov	dword [EXECUTE_CYCLE], get_opcode
			ret



;****************************************************************************
;* opcode_php ***************************************************************
;****************************************************************************
opcode_php:
			mov	dword ebx, [STACK_POINTER]
			mov	byte  al, [FLAG_REGISTER]
			call	dword [MEMORY_WRITE]
			dec	byte  bl
			mov	dword [EXECUTE_CYCLE], ._1
			mov	dword [STACK_POINTER], ebx
			ret

._1:			CHECK_FOR_DMA
			mov	dword [EXECUTE_CYCLE], get_opcode
			ret



;****************************************************************************
;* opcode_pla ***************************************************************
;****************************************************************************
opcode_pla:
			mov	dword ebx, [STACK_POINTER]
			inc	byte  bl
			call	dword [MEMORY_READ]
			mov	byte  [ACCUMULATOR], al
			mov	dword [STACK_POINTER], ebx
			test	byte  al, al
			setz	byte  cl
			sets	byte  dl
			shl	byte  cl, CPU65XX_FLAG_Z_BIT
			mov	byte  bl, [FLAG_REGISTER]
			shl	byte  dl, CPU65XX_FLAG_N_BIT
			and	byte  bl, (NOT8 - CPU65XX_FLAG_N - CPU65XX_FLAG_Z)
			or	byte  cl, dl
			mov	dword eax, ._1
			or	byte  bl, cl
			mov	dword [EXECUTE_CYCLE], eax
			mov	byte  [FLAG_REGISTER], bl
			ret

._1:			CHECK_FOR_DMA
			mov	dword [EXECUTE_CYCLE], ._2
			ret

._2:			CHECK_FOR_DMA
			mov	dword [EXECUTE_CYCLE], get_opcode
			ret



;****************************************************************************
;* opcode_plp ***************************************************************
;****************************************************************************
opcode_plp:
			mov	dword ebx, [STACK_POINTER]
			inc	byte  bl
			call	dword [MEMORY_READ]
			or	byte  al, (CPU65XX_FLAG_R | CPU65XX_FLAG_B)
			mov	byte  [FLAG_REGISTER], al
			mov	dword [STACK_POINTER], ebx
			mov	dword [EXECUTE_CYCLE], ._1
			ret

._1:			CHECK_FOR_DMA
			mov	dword [EXECUTE_CYCLE], ._2
			ret

._2:			CHECK_FOR_DMA
			mov	dword [EXECUTE_CYCLE], get_opcode
			ret



;****************************************************************************
;* opcode_rti ***************************************************************
;****************************************************************************
opcode_rti:
			mov	dword ebx, [STACK_POINTER]
			inc	byte  bl
			call	dword [MEMORY_READ]
			or	byte  al, (CPU65XX_FLAG_R | CPU65XX_FLAG_B)
			mov	byte  [FLAG_REGISTER], al
			mov	dword [STACK_POINTER], ebx
			mov	dword [EXECUTE_CYCLE], ._1
			ret

._1:			mov	dword ebx, [STACK_POINTER]
			inc	byte  bl
			call	dword [MEMORY_READ]
			mov	byte  [PROGRAM_COUNTER], al
			mov	dword [STACK_POINTER], ebx
			mov	dword [EXECUTE_CYCLE], ._2
			ret

._2:			mov	dword ebx, [STACK_POINTER]
			inc	byte  bl
			call	dword [MEMORY_READ]
			mov	byte  [PROGRAM_COUNTER + 1], al
			mov	dword [STACK_POINTER], ebx
			mov	dword [EXECUTE_CYCLE], ._3
			ret

._3:			CHECK_FOR_DMA
			mov	dword [EXECUTE_CYCLE], ._4
			ret

._4:			CHECK_FOR_DMA
			mov	dword [EXECUTE_CYCLE], get_opcode
			ret



;****************************************************************************
;* opcode_rts ***************************************************************
;****************************************************************************
opcode_rts:
			mov	dword ebx, [STACK_POINTER]
			inc	byte  bl
			call	dword [MEMORY_READ]
			mov	byte  [PROGRAM_COUNTER], al
			mov	dword [STACK_POINTER], ebx
			mov	dword [EXECUTE_CYCLE], ._1
			ret

._1:			mov	dword ebx, [STACK_POINTER]
			inc	byte  bl
			call	dword [MEMORY_READ]
			mov	byte  [PROGRAM_COUNTER + 1], al
			mov	dword [STACK_POINTER], ebx
			mov	dword [EXECUTE_CYCLE], ._2
			ret

._2:			CHECK_FOR_DMA
			mov	dword eax, [PROGRAM_COUNTER]
			mov	dword ebx, ._3
			inc	word ax
			mov	dword [EXECUTE_CYCLE], ebx
			mov	dword [PROGRAM_COUNTER], eax
			ret

._3:			CHECK_FOR_DMA
			mov	dword [EXECUTE_CYCLE], ._4
			ret

._4:			CHECK_FOR_DMA
			mov	dword [EXECUTE_CYCLE], get_opcode
			ret



;****************************************************************************
;* opcode_illegal ***********************************************************
;****************************************************************************
opcode_illegal:
			or	dword [STATE], CPU65XX_ILLEGAL_OPCODE
			mov	dword eax, [MODUL_ID]
			jmp	main_exception






;****************************************************************************
;****************************************************************************
;*
;* PART 4.3: The following Routines are not directly Opcodes
;*
;****************************************************************************
;****************************************************************************





;****************************************************************************
;* get_opcode ***************************************************************
;****************************************************************************
;* fetch a byte from memory and interpret it as opcode
;*
get_opcode:
;%ifdef CPU65XX_DEBUG
;			cmp	dword esi, cpu65xx__vc1541
;			jne	.skip_debug
;			call	cpu65xx_debug
;.skip_debug:
;%endif
			mov	dword eax, [STATE]
			mov	byte  bl, [CYCLE]
			and	dword eax, (NOT32 - CPU65XX_PROCESSING_RESET - CPU65XX_PROCESSING_NMI - CPU65XX_PROCESSING_IRQ)
			mov	byte  [CYCLE], 1

			;--------------------------------
			;check if we have to handle a nmi
			;--------------------------------

			test	dword eax, CPU65XX_SIGNAL_NMI
			jz	.skip_nmi_handler
			test	dword eax, CPU65XX_NMI_DELAY_OK
			jz	.check_nmi_delay
			and	dword eax, (NOT32 - CPU65XX_NMI_DELAY_OK)
.process_nmi:		or	dword eax, CPU65XX_PROCESSING_NMI
			mov	dword [STATE], eax
			jmp	processing_nmi
.check_nmi_delay:	mov	byte  cl, [NMI_CYCLE]
			add	byte  cl, 2
			cmp	byte  bl, cl
			ja	.process_nmi
			or	dword eax, CPU65XX_NMI_DELAY_OK
			
			;---------------------------------
			;check if we have to handle an irq
			;---------------------------------

.skip_nmi_handler:	test	byte  [FLAG_REGISTER], CPU65XX_FLAG_I
			jnz	.skip_irq_handler
			test	dword eax, CPU65XX_SIGNAL_IRQ
			jz	.skip_irq_handler
			test	dword eax, CPU65XX_IRQ_DELAY_OK
			jz	.check_irq_delay
			and	dword eax, (NOT32 - CPU65XX_IRQ_DELAY_OK)
.process_irq:		or	dword eax, CPU65XX_PROCESSING_IRQ
			mov	dword [STATE], eax
			jmp	processing_irq
.check_irq_delay:	mov	byte  cl, [IRQ_CYCLE]
			add	byte  cl, 2
			cmp	byte  bl, cl
			ja	.process_irq
			or	dword eax, CPU65XX_IRQ_DELAY_OK

			;-------------------
			;get the next opcode
			;-------------------

.skip_irq_handler:	mov	dword [STATE], eax
			mov	dword ebx, [PROGRAM_COUNTER]
			call	dword [MEMORY_READ]
			xor	dword ecx, ecx
			mov	byte  [OPCODE], al
			inc	word  bx
			mov	byte  cl, al
			xor	dword edx, edx
			mov	dword [PROGRAM_COUNTER], ebx
			mov	byte  dl, [ab__opcode_details + 2 * ecx]
			shl	dword edx, 4
			mov	byte  cl, [ab__opcode_details + 2 * ecx + 1]
			mov	dword eax, [ap__address_modes + edx]
			mov	dword ebx, [ap__address_modes + edx + 4]
			mov	dword ebp, [ap__opcodes + 4 * ecx]
			mov	dword [EXECUTE_CYCLE], eax
			mov	dword [DELAY_ADDRESS], ebx
			mov	dword [OPCODE_ADDRESS], ebp
			mov	dword eax, [ap__address_modes + edx + 8]
			mov	dword ebx, [ap__address_modes + edx + 12]
			mov	dword [RW_ADDRESS], eax
			mov	dword [RMW_ADDRESS], ebx
			ret



;****************************************************************************
;* processing_irq ***********************************************************
;****************************************************************************
processing_irq:
			;----------------------
			;seems to have 7 Cycles
			;----------------------

			CHECK_FOR_DMA
			mov	dword [EXECUTE_CYCLE], ._0
			ret

._0:			CHECK_FOR_DMA
			mov	dword [EXECUTE_CYCLE], ._1
			ret

._1:			mov	dword ebx, [STACK_POINTER]
			mov	dword ecx, ._2
			mov	byte  al, [PROGRAM_COUNTER + 1]
			call	dword [MEMORY_WRITE]
			dec	byte  bl
			mov	dword [EXECUTE_CYCLE], ecx
			mov	dword [STACK_POINTER], ebx
			ret

._2:			mov	dword ebx, [STACK_POINTER]
			mov	dword ecx, ._3
			mov	byte  al, [PROGRAM_COUNTER]
			call	dword [MEMORY_WRITE]
			dec	byte  bl
			mov	dword [EXECUTE_CYCLE], ecx
			mov	dword [STACK_POINTER], ebx
			ret

._3:			mov	dword ebx, [STACK_POINTER]
			mov	byte  al, [FLAG_REGISTER]
			mov	dword ecx, ._4
			and	byte  al, (NOT8 - CPU65XX_FLAG_B)
			call	dword [MEMORY_WRITE]
			dec	byte  bl
			mov	dword [EXECUTE_CYCLE], ecx
			mov	dword [STACK_POINTER], ebx
			ret

._4:			mov	dword ebx, 00000fffeh
			call	dword [MEMORY_READ]
			mov	dword ecx, ._5
			mov	byte  [PROGRAM_COUNTER], al
			mov	dword [EXECUTE_CYCLE], ecx
			ret

._5:			mov	dword ebx, 00000ffffh
			call	dword [MEMORY_READ]
			mov	byte  bl, [FLAG_REGISTER]
			mov	byte  [PROGRAM_COUNTER + 1], al
			or	byte  bl, CPU65XX_FLAG_I
			mov	dword [EXECUTE_CYCLE], get_opcode
			mov	byte  [FLAG_REGISTER], bl
			ret



;****************************************************************************
;* processing_nm ************************************************************
;****************************************************************************
processing_nmi:
			;----------------------
			;seems to have 7 Cycles
			;----------------------

			CHECK_FOR_DMA
			mov	dword eax, [STATE]
			mov	dword ecx, ._0
			and	dword eax, (NOT32 - CPU65XX_SIGNAL_NMI)
			mov	dword [EXECUTE_CYCLE], ecx
			mov	dword [STATE], eax
			ret

._0:			CHECK_FOR_DMA
			mov	dword [EXECUTE_CYCLE], ._1
			ret

._1:			mov	dword ebx, [STACK_POINTER]
			mov	dword ecx, ._2
			mov	byte  al, [PROGRAM_COUNTER + 1]
			call	dword [MEMORY_WRITE]
			dec	byte  bl
			mov	dword [EXECUTE_CYCLE], ecx
			mov	dword [STACK_POINTER], ebx
			ret

._2:			mov	dword ebx, [STACK_POINTER]
			mov	dword ecx, ._3
			mov	byte  al, [PROGRAM_COUNTER]
			call	dword [MEMORY_WRITE]
			dec	byte  bl
			mov	dword [EXECUTE_CYCLE], ecx
			mov	dword [STACK_POINTER], ebx
			ret

._3:			mov	dword ebx, [STACK_POINTER]
			mov	byte  al, [FLAG_REGISTER]
			mov	dword ecx, ._4
			and	byte  al, (NOT8 - CPU65XX_FLAG_B)
			call	dword [MEMORY_WRITE]
			dec	byte  bl
			mov	dword [EXECUTE_CYCLE], ecx
			mov	dword [STACK_POINTER], ebx
			ret

._4:			mov	dword ebx, 00000fffah
			call	dword [MEMORY_READ]
			mov	dword ecx, ._5
			mov	byte  [PROGRAM_COUNTER], al
			mov	dword [EXECUTE_CYCLE], ecx
			ret

._5:			mov	dword ebx, 00000fffbh
			call	dword [MEMORY_READ]
			mov	byte  bl, [FLAG_REGISTER]
			mov	byte  [PROGRAM_COUNTER + 1], al
			or	byte  bl, CPU65XX_FLAG_I
			mov	dword [EXECUTE_CYCLE], get_opcode
			mov	byte  [FLAG_REGISTER], bl
			ret





;****************************************************************************
;****************************************************************************
;*
;* PART 5: Code for testing the CPU Emulation
;*
;****************************************************************************
;****************************************************************************





%ifdef CPU65XX_TEST
%include "system/sys_calls.inc"
global _start






;****************************************************************************
;****************************************************************************
;*
;* 5.1 Routine needed from main.inc
;*
;****************************************************************************
;****************************************************************************





section .text
main_exception:
			call	cpu65xx_dump
			SYS_EXIT  1





;****************************************************************************
;****************************************************************************
;*
;* 5.2 Memory Access Routines
;*
;****************************************************************************
;****************************************************************************





;****************************************************************************
;* 64 KB Memory *************************************************************
;****************************************************************************
section .bss
ab__memory		resb	010000h



;****************************************************************************
;* memory_initialize ********************************************************
;****************************************************************************
section .text
memory_initialize:
			mov	dword ebx, .data
			xor	dword edx, edx
.get_count:
			mov	word  cx, [ebx]
			add	dword ebx, 02h
			test	word  cx, cx
			jz	.end
			mov	word  dx, [ebx]
			add	dword ebx, 02h
.write_bytes:
			mov	byte  al, [ebx]
			mov	byte [ab__memory + edx], al
			inc	dword ebx
			inc	word  dx
			dec	word  cx
			jnz	.write_bytes
			jmp	.get_count
.end:
			ret

section .data
.data:			dw	6
			dw	0fffch
			dw	08000h, 08000h, 08000h

			dw	.pass_1_end - .pass_1_start
			dw	08000h
.pass_1_start:		db	058h			;cli
			db	0a2h, 0ffh		;ldx #$ff
			db	09ah			;txs
			db	0a2h, 000h		;ldx #$00
			db	08ah			;txa
			db	09dh, 000h, 010h	;sta $1000,x
			db	0e8h			;inx
			db	010h, 0f9h		;bpl PC-$07
			db	0e6h, 000h		;inc $00
			db	04ch, 000h, 081h	;jmp $8100
.pass_1_end:
			dw	.pass_2_end - .pass_2_start
			dw	08100h
.pass_2_start:		db	0a9h, 080h		;lda #$80
			db	030h, 001h		;bmi PC+$01
			db	00ah			;asl
			db	08dh, 000h, 011h	;sta $1100
			db	038h			;sec
			db	090h, 001h		;bcc PC+$01
			db	06ah			;ror
			db	08dh, 001h, 011h	;sta $1101
			db	0e6h, 000h		;inc $00
			db	04ch, 000h, 082h	;jmp $8200
			db	0bah			;tsx
			db	0e8h			;inx
			db	0bdh, 000h, 001h	;lda $0100,x
			db	08dh, 000h, 012h	;sta $1200
			db	0e8h			;inx
			db	0bdh, 000h, 001h	;lda $0100,x
			db	08dh, 001h, 012h	;sta $1201
			db	060h			;rts
.pass_2_end:
			dw	.pass_3_end - .pass_3_start
			dw	08200h
.pass_3_start:		db	020h, 014h, 081h	;jsr $8114
			db	0e6h, 000h		;inc $00
			db	04ch, 000h, 083h	;jmp $8300
.pass_3_end:
			dw	.pass_4_end - .pass_4_start
			dw	08300h
.pass_4_start:		db	0f8h			;sed
			db	0a9h, 00ah		;lda #$0a
			db	018h			;clc
			db	069h, 001h		;adc #$01
			db	08dh, 000h, 013h	;sta $1300
			db	0a9h, 0a9h		;lda #$a9
			db	018h			;clc
			db	069h, 001h		;adc #$01
			db	08dh, 001h, 013h	;sta $1301
			db	0a9h, 0ffh		;lda #$ff
			db	018h			;clc
			db	069h, 000h		;adc #$00
			db	08dh, 002h, 013h	;sta $1302
			db	0e6h, 000h		;inc $00
.pass_4_end:
			dw	0


;****************************************************************************
;* memory_read **************************************************************
;****************************************************************************
section .text
memory_read:
			mov	byte  al, [ab__memory + ebx]
			ret



;****************************************************************************
;* memory_write *************************************************************
;****************************************************************************
section .text
memory_write:
			mov	byte  [ab__memory + ebx], al
			ret





;****************************************************************************
;****************************************************************************
;*
;* 5.3 Output Routines
;*
;****************************************************************************
;****************************************************************************





;****************************************************************************
;* write_pass ***************************************************************
;****************************************************************************
section .text
write_pass:
			inc	dword [.d__pass_counter]
			WRITE_TEXT  .cb__pass, 5
			WRITE_HEX  8, [.d__pass_counter]
			WRITE_TEXT  .cb__points, 5
			ret
section .data
.cb__pass:		db	"Pass "
.cb__points:		db	" ... "
.d__pass_counter:	dd	0



;****************************************************************************
;* write_passed *************************************************************
;****************************************************************************
section .text
write_passed:
			WRITE_TEXT  .cb__passed, 7
			ret

section .data
.cb__passed:		db	"passed", 00ah



;****************************************************************************
;* write_failed *************************************************************
;****************************************************************************
section .text
write_failed:
			WRITE_TEXT  .cb__failed, 7
			ret

section .data
.cb__failed:		db	"failed", 00ah





;****************************************************************************
;****************************************************************************
;*
;* 5.4 The Test Program
;*
;****************************************************************************
;****************************************************************************





section .text
_start:
			WRITE_TEXT  .cb__message, .message_length

			call	cpu65xx_initialize
			call	memory_initialize

			mov	dword esi, .cpu65xx_test
			call	cpu65xx_reset


.test_pass_1:		call	write_pass
			xor	dword ebx, ebx
			call	memory_read
			mov	byte  [.b__end_flag], al
			mov	dword [.counter], 000001000h
.pass_1_run:		CALL_CPU65XX_CYCLE
			xor	dword ebx, ebx
			call	memory_read
			cmp	byte  [.b__end_flag], al
			jne	.pass_1_end
			dec	dword [.counter]
			jnz	.pass_1_run
			jmp	.pass_1_failed
.pass_1_end:		mov	dword ebx, 00000107fh
.pass_1_control:	call	memory_read
			cmp	byte  al, bl
			jne	.pass_1_failed
			dec	byte  bl
			jns	.pass_1_control
			call	write_passed
			jmp	.test_pass_2
.pass_1_failed:		call	write_failed

.test_pass_2:		call	write_pass
			xor	dword ebx, ebx
			call	memory_read
			mov	byte  [.b__end_flag], al
			mov	dword [.counter], 000000020h
.pass_2_run:		CALL_CPU65XX_CYCLE
			xor	dword ebx, ebx
			call	memory_read
			cmp	byte  [.b__end_flag], al
			jne	.pass_2_end
			dec	dword [.counter]
			jnz	.pass_2_run
			jmp	.pass_2_failed
.pass_2_end:		mov	dword ebx, 000001100h
                	call	memory_read
			cmp	byte  al, 080h
			jne	.pass_2_failed
			inc	byte  bl
			call	memory_read
			cmp	byte  al, 0c0h
			jne	.pass_2_failed
			call	write_passed
			jmp	.test_pass_3
.pass_2_failed:		call	write_failed

.test_pass_3:		call	write_pass
			xor	dword ebx, ebx
			call	memory_read
			mov	byte  [.b__end_flag], al
			mov	dword [.counter], 000000030h
.pass_3_run:		CALL_CPU65XX_CYCLE
			xor	dword ebx, ebx
			call	memory_read
			cmp	byte  [.b__end_flag], al
			jne	.pass_3_end
			dec	dword [.counter]
			jnz	.pass_3_run
			jmp	.pass_3_failed
.pass_3_end:		mov	dword ebx, 000001200h
                	call	memory_read
			cmp	byte  al, 002h
			jne	.pass_3_failed
			inc	byte  bl
			call	memory_read
			cmp	byte  al, 082h
			jne	.pass_3_failed
			call	write_passed
			jmp	.test_pass_4
.pass_3_failed:		call	write_failed

.test_pass_4:		call	write_pass
			xor	dword ebx, ebx
			call	memory_read
			mov	byte  [.b__end_flag], al
			mov	dword [.counter], 000000040h
.pass_4_run:		CALL_CPU65XX_CYCLE
			xor	dword ebx, ebx
			call	memory_read
			cmp	byte  [.b__end_flag], al
			jne	.pass_4_end
			dec	dword [.counter]
			jnz	.pass_4_run
			jmp	.pass_4_failed
.pass_4_end:		mov	dword ebx, 000001300h
                	call	memory_read
			push	dword ebx
			WRITE_HEX  2
			pop	dword ebx
;			cmp	byte  al, 011h
;			jne	.pass_4_failed
			inc	byte  bl
			call	memory_read
			push	dword ebx
			WRITE_HEX  2
			pop	dword ebx
;			cmp	byte  al, 011h
;			jne	.pass_4_failed
			inc	byte  bl
			call	memory_read
			push	dword ebx
			WRITE_HEX  2
			pop	dword ebx
;			cmp	byte  al, 080h
;			jne	.pass_4_failed
			call	write_passed
			jmp	.test_pass_5
.pass_4_failed:		call	write_failed

.test_pass_5:
;			rdtsc
;			mov	dword ebx, eax
;			rdtsc
;			sub	dword eax, ebx
;			mov	byte bl, 007h
;			call	write_hex
			SYS_EXIT

section .data
.cb__message:		db	"CPU 65XX Emulation Selftest", 00ah
.message_length:	equ	$ - .cb__message

.cpu65xx_test:		istruc	cpu65xx_data
			at d__cpu_modul_id,	dd 065020000h
			at p__cpu_execute_cycle,dd opcode_illegal
			at p__cpu_memory_read,	dd memory_read
			at p__cpu_memory_write,	dd memory_write
			iend

.b__end_flag:		db	0
align 4
.counter:		dd	0


%endif
;********************************************* karsten.scheibler@bigfoot.de *

