;****************************************************************************
;****************************************************************************
;*
;* disc.nasm
;*
;****************************************************************************
;****************************************************************************


;TODO
;* allow SYNC's also on non byte aligned positions
;* read rubbish, if the read speed differs from the write speed
;* include the mechanic characteristics of the step- and drivemotor 
;  [delaycounter]
;* read rubbish if drivemotor is turned off
;* clean up code




;%define DISC_DEBUG
;%define DISC_INCLUDE_D64





%include "defines.inc"
%include "main.inc"
%include "link.inc"
%include "cpu65xx.inc"
%include "via6522.inc"

%define DISC_NASM
%include "disc.inc"



;****************************************************************************
;* assign's & defines's *****************************************************
;****************************************************************************
%define MODUL_ID		DISC_MODUL_ID
%define STATE			DISC_STATE
%define CYCLES_PER_TRACK	DISC_CYCLES_PER_TRACK
%define BYTES_PER_TRACK		DISC_BYTES_PER_TRACK
%define CYCLE_COUNTER		DISC_CYCLE_COUNTER
%define TRACK_OFFSET		DISC_TRACK_OFFSET
%define SPEED_TABLE		DISC_SPEED_TABLE
%define TRACK			DISC_TRACK
%define DATA_BYTE		DISC_DATA_BYTE
%define LAST_BYTE		DISC_LAST_BYTE
%define HEAD_POSITION		DISC_HEAD_POSITION
%define GCR_IMAGE		DISC_GCR_IMAGE

%assign DATA_RW_MODE		001h
%assign NO_BYTE_READY		002h
%assign SYNC_FOUND		004h



;****************************************************************************
;* disc_initialize **********************************************************
;****************************************************************************
section .text
disc_initialize:
			mov	byte  al, 055h
			mov	dword ebx, DISC_GCR_IMAGE_SIZE - 1
.fill:			mov	byte  [GCR_IMAGE + ebx], al
			dec	dword ebx
			jns	.fill
%ifdef DISC_INCLUDE_D64
			mov	dword eax, ab__disc_image
			call	cbm_gcr_encode_d64
			ret

			align 4
ab__disc_image:		incbin "../../../../disk-images/turrican-side1.d64"
%else
			ret
%endif



;****************************************************************************
;* disc_exit ****************************************************************
;****************************************************************************
section .text
disc_exit:

%ifdef DISC_DEBUG
%include "sys_calls.inc"
			SYS_WRITE  STDERR, GCR_IMAGE, DISC_GCR_IMAGE_SIZE
%endif
			ret



;****************************************************************************
;* disc_save_snapshot *******************************************************
;****************************************************************************
section .text
disc_save_snapshot:
			ret



;****************************************************************************
;* disc_load_snapshot *******************************************************
;****************************************************************************
section .text
disc_load_snapshot:
			ret



;****************************************************************************
;* disc_reset ***************************************************************
;****************************************************************************
section .text
disc_reset:
			mov	dword eax, VC1541_CLOCK / 5	;300rpm / 60 seconds = 5
			mov	dword ebx, 6250
			mov	dword ecx, 6667
			mov	dword [CYCLES_PER_TRACK], eax
			mov	dword [SPEED_TABLE], ebx
			mov	dword [BYTES_PER_TRACK], ebx
			mov	dword [SPEED_TABLE + 4], ecx
			xor	dword eax, eax
			mov	dword ebx, 7143
			mov	dword ecx, 7692
			mov	dword [SPEED_TABLE + 8], ebx
			mov	dword [SPEED_TABLE + 12], ecx
			mov	dword [STATE], eax
			mov	dword [CYCLE_COUNTER], eax
			mov	dword [TRACK_OFFSET], eax
			mov	byte  [TRACK], al
			ret



;****************************************************************************
;* disc_cycle ***************************************************************
;****************************************************************************
section .text
disc_cycle:
			mov	dword eax, [STATE]
			test	dword eax, (NO_BYTE_READY | SYNC_FOUND)
			jnz	.no_byte_ready
			call	cpu65xx_signal_set_overflow
.no_byte_ready:		xor	dword ebx, ebx
			mov	dword ecx, [TRACK_OFFSET]
			mov	byte  bl, [TRACK]
			inc	dword ecx
			shl	dword ebx, 13		;TRACK * 02000h Bytes
			cmp	dword ecx, [BYTES_PER_TRACK]
			jb	.no_clear
			xor	dword ecx, ecx
.no_clear:		add	dword ebx, ecx
			mov	dword [TRACK_OFFSET], ecx
			test	dword eax, DATA_RW_MODE
			jz	.read
.write:			and	dword eax, (NOT32 - SYNC_FOUND)
			mov	dword [STATE], eax
			mov	byte  al, [DATA_BYTE]
			mov	byte  [LAST_BYTE], al
			xor	byte  [GCR_IMAGE + ebx], al
			mov	byte  [GCR_IMAGE + DISC_TRACK_SIZE + ebx], al
			xor	byte  [GCR_IMAGE + 2 * DISC_TRACK_SIZE + ebx], al
			mov	byte  al, 1
			call	link_disc_sync
			ret
.read:			and	dword eax, (NOT32 - SYNC_FOUND)
			mov	byte  bl, [GCR_IMAGE + DISC_TRACK_SIZE + ebx]
			mov	byte  cl, 1
			cmp	byte  bl, 0ffh
			jne	.read_1
			cmp	byte  [LAST_BYTE], 0ffh
			jne	.read_1
			or	dword eax, SYNC_FOUND
			xor	byte  cl, cl
.read_1:		mov	dword [STATE], eax
			mov	byte  al, [LAST_BYTE]
			mov	byte  [LAST_BYTE], bl
			call	link_disc_data
			mov	byte  al, cl
			call	link_disc_sync
			ret



;****************************************************************************
;* disc_dump ****************************************************************
;****************************************************************************
section .text
disc_dump:
			ret



;****************************************************************************
;* disc_signal_data *********************************************************
;****************************************************************************
section .text
disc_signal_data:
			mov	byte  [DATA_BYTE], al
;%ifdef DISC_DEBUG
;%include "sys_calls.inc"
;			push	dword eax
;			push	dword ebx
;			push	dword ecx
;			push	dword edx
;			SYS_WRITE  STDERR, DATA_BYTE, 1
;			pop	dword edx
;			pop	dword ecx
;			pop	dword ebx
;			pop	dword eax
;%endif
			ret



;****************************************************************************
;* disc_signal_rw ***********************************************************
;****************************************************************************
section .text
disc_signal_rw:
			test	byte  al, 001h
			jnz	.read
.write:			or	dword [STATE], DATA_RW_MODE
			ret
.read:			and	dword [STATE], (NOT32 - DATA_RW_MODE)
			ret



;****************************************************************************
;* disc_signal_byte_ready ***************************************************
;****************************************************************************
section .text
disc_signal_byte_ready:
			test	byte  al, 001h
			jz	.disable
.enable:		and	dword [STATE], (NOT32 - NO_BYTE_READY)
			ret
.disable:		or	dword [STATE], NO_BYTE_READY
			ret



;****************************************************************************
;* disc_signal_head *********************************************************
;****************************************************************************
section .text
disc_signal_head:
			push	dword eax
			and	byte  al, 3
			mov	byte  ah, [HEAD_POSITION]
			inc	byte  ah
			and	byte  ah, 3
.compare_1:		cmp	byte  al, ah
			jne	.compare_2
			mov	byte  ah, [TRACK]
			cmp	byte  ah, DISC_LAST_TRACK
			je	.end
			inc	byte  ah
			mov	byte  [TRACK], ah
			jmp	.end
.compare_2:		sub	byte  ah, 2
			and	byte  ah, 3
			cmp	byte  al, ah
			jne	.end
			mov	byte  ah, [TRACK]
			cmp	byte  ah, 0
			je	.end
			dec	byte  ah
			mov	byte  [TRACK], ah
.end:			mov	byte  [HEAD_POSITION], al
			pop	dword eax
			ret



;****************************************************************************
;* disc_signal_speed ********************************************************
;****************************************************************************
section .text
disc_signal_speed:
			push	dword eax
			and	dword eax, 3
			mov	dword eax, [SPEED_TABLE + 4 * eax]
			mov	dword [BYTES_PER_TRACK], eax
			pop	dword eax
			ret



;****************************************************************************
;* cbm_gcr_encode_dword *****************************************************
;****************************************************************************
section .text
cbm_gcr_encode_dword:
			push	dword eax
			push	dword ebx
			push	dword ecx
			push	dword edx
			push	dword ebp
			xor	dword ebx, ebx
			mov	byte  cl, 006h
			xor	dword ebp, ebp
.loop:			mov	byte  bl, al
			shl	dword ebp, 5
			shr	byte  bl, 4
			or	dword ebp, [.ad__gcr_codes + 4 * ebx]
			mov	byte  bl, al
			shl	dword ebp, 5
			and	byte  bl, 00fh
			or	dword ebp, [.ad__gcr_codes + 4 * ebx]
			mov	dword edx, ebp
			shr	dword eax, 8
			shl	dword edx, cl
			mov	byte  [edi], dh
			inc	dword edi
			sub	byte  cl, 2
			jns	.loop
			mov	byte  [edi], dl
			inc	dword edi
			pop	dword ebp
			pop	dword edx
			pop	dword ecx
			pop	dword ebx
			pop	dword eax
			ret

			align 4
.ad__gcr_codes:		dd	00ah, 00bh, 012h, 013h
			dd	00eh, 00fh, 016h, 017h
			dd	009h, 019h, 01ah, 01bh
			dd	00dh, 01dh, 01eh, 015h



;****************************************************************************
;* cbm_gcr_encode_sector ****************************************************
;****************************************************************************
section .text
cbm_gcr_encode_sector:
			push	dword eax
			push	dword ebx
			push	dword ecx
			mov	dword ecx, eax
			push	dword esi
			push	dword edi

			;---------------
			;Header: 5 x $FF
			;---------------

			mov	byte  al, 0ffh
			mov	byte  [edi], al
			mov	byte  [edi + 1], al
			mov	byte  [edi + 2], al
			mov	byte  [edi + 3], al
			mov	byte  [edi + 4], al
			add	dword edi, 5

			;------------------------------------
			;Header: $08, Checksum, Sector, Track
			;------------------------------------

			mov	dword eax, ecx
			shl	dword eax, 16
			mov	byte  ah, ch
			xor	byte  ah, cl
			mov	byte  al, 008h
			call	cbm_gcr_encode_dword

			;--------------------------
			;Header: $30, $30, $0f, $0f
			;--------------------------

			mov	dword eax, 00f0f3030h
			call	cbm_gcr_encode_dword

			;---------------
			;Header: 8 x $55
			;---------------

			mov	byte  al, 055h
			mov	byte  [edi], al
			mov	byte  [edi + 1], al
			mov	byte  [edi + 2], al
			mov	byte  [edi + 3], al
			mov	byte  [edi + 4], al
			mov	byte  [edi + 5], al
			mov	byte  [edi + 6], al
			mov	byte  [edi + 7], al
			add	dword edi, 8

			;--------------
			;Block: 5 x $FF
			;--------------

			mov	byte  al, 0ffh
			mov	byte  [edi], al
			mov	byte  [edi + 1], al
			mov	byte  [edi + 2], al
			mov	byte  [edi + 3], al
			mov	byte  [edi + 4], al
			add	dword edi, 5

			;-------------------------
			;Block: calculate Checksum
			;-------------------------

			xor	dword ebx, ebx
			xor	byte  ch, ch
.checksum:		xor	byte  ch, [esi + ebx]
			inc	byte  bl
			jnz	.checksum

			;-------------------------
			;Block: $07, 3 x Databytes
			;-------------------------

			mov	byte  ah, [esi + 2]
			mov	byte  al, [esi + 1]
			shl	dword eax, 16
			mov	byte  ah, [esi]
			mov	byte  al, 007h
			call	cbm_gcr_encode_dword
			add	dword esi, 3

			;--------------------
			;Block: 252 Databytes
			;--------------------

			mov	byte  cl, 03fh
.loop:			mov	byte  ah, [esi + 3]
			mov	byte  al, [esi + 2]
			shl	dword eax, 16
			mov	byte  ah, [esi + 1]
			mov	byte  al, [esi]
			call	cbm_gcr_encode_dword
			add	dword esi, 4
			dec	byte  cl
			jnz	.loop

			;---------------------------------------
			;Block: 1 x Databyte, Checksum, $00, $00
			;---------------------------------------

			xor	dword eax, eax
			mov	byte  al, [esi]
			mov	byte  ah, ch
			call	cbm_gcr_encode_dword

			;--------------
			;Block: 8 x $55
			;--------------

			mov	byte  al, 055h
			mov	byte  [edi], al
			mov	byte  [edi + 1], al
			mov	byte  [edi + 2], al
			mov	byte  [edi + 3], al
			mov	byte  [edi + 4], al
			mov	byte  [edi + 5], al
			mov	byte  [edi + 6], al
			mov	byte  [edi + 7], al

			pop	dword edi
			pop	dword esi
			pop	dword ecx
			pop	dword ebx
			pop	dword eax
			ret



;****************************************************************************
;* cbm_gcr_encode_d64 *******************************************************
;****************************************************************************
section .text
disc_encode_d64:
cbm_gcr_encode_d64:
			push	dword eax
			push	dword ebx
			push	dword ecx
			push	dword esi
			push	dword edi
			mov	dword esi, eax
			mov	dword eax, 00100h
			xor	dword ebx, ebx
			mov	dword ecx, GCR_IMAGE + DISC_TRACK_SIZE
			mov	dword edi, ecx
.encode:		cmp	byte  [.ab__sectors + ebx], 0
			je	.end
			call	cbm_gcr_encode_sector
			inc	byte  al
			add	dword esi, 0100h
			add	dword edi, 0168h
			cmp	byte  al, [.ab__sectors + ebx]
			jne	.encode
			inc	byte  ah
			inc	dword ebx
			add	dword ecx, (2 * DISC_TRACK_SIZE)
			xor	byte  al, al
			mov	dword edi, ecx
			jmp	.encode
.end:			pop	dword edi
			pop	dword esi
			pop	dword ecx
			pop	dword ebx
			pop	dword eax
			ret

.ab__sectors:		db	21, 21, 21, 21, 21, 21, 21, 21
			db	21, 21, 21, 21, 21, 21, 21, 21
			db	21, 19, 19, 19, 19, 19, 19, 19
			db	18, 18, 18, 18, 18, 18, 17, 17
			db	17, 17, 17, 0
;********************************************* karsten.scheibler@bigfoot.de *

