;****************************************************************************
;****************************************************************************
;*
;* link.nasm
;*
;****************************************************************************
;****************************************************************************


;TODO:
;
;* ATNout from the VC1541 isn't connected to the serial bus at the moment,
;  examine the correct wiring





%include "defines.inc"
%include "main.inc"
%include "cpu65xx.inc"
%include "vic6569.inc"
%include "cia6526.inc"
%include "via6522.inc"
%include "disc.inc"

%define LINK_NASM
%include "link.inc"



;****************************************************************************
;* assign's & defines's *****************************************************
;****************************************************************************
%assign C64_IRQ_CIA1		001h
%assign C64_IRQ_VIC		002h
%assign VC1541_IRQ_VIA1		001h
%assign VC1541_IRQ_VIA2		002h



;****************************************************************************
;* local data ***************************************************************
;****************************************************************************
section .data
p__c64_key_codes:	dd	0
cd__c64_key_codes:	times 16 dd 0
b__c64_serial_bus:	db	0
b__vc1541_serial_bus:	db	0
flag__atn_cleared:	db	0



;****************************************************************************
;* link_nothing *************************************************************
;****************************************************************************
section .text
link_nothing:
			ret



;****************************************************************************
;* link_keyboard ************************************************************
;****************************************************************************
section .text
link_keyboard:		
			push	dword eax
			push	dword ebx
			push	dword ecx
			push	dword edx
			xor	dword ebx, ebx
			xor	dword ecx, ecx
.next_key_code:		mov	byte  cl, al
			mov	dword edx, [.ad__c64_key_codes + 4 * ecx]
			mov	dword [cd__c64_key_codes + 4 * ebx], edx
			inc	byte  bl
			shr	dword eax, 8
			jnz	.next_key_code
			xor	dword ecx, ecx
			mov	dword eax, cd__c64_key_codes
			mov	dword [cd__c64_key_codes + 4 * ebx], ecx
			jmp	link_cia1_prb.common

			align 4
.ad__c64_key_codes:	
			;-------
			;a ... o
                        ;-------

			dd	00000h, 00204h, 00810h, 00410h
			dd	00404h, 00240h, 00420h, 00804h
			dd	00820h, 01002h, 01004h, 01020h
			dd	02004h, 01010h, 01080h, 01040h

			;-------------------------------
			;p ... z SHIFT_LEFT  SHIFT_RIGHT
                        ;-------------------------------

			dd	02002h, 08040h, 00402h, 00220h
			dd	00440h, 00840h, 00880h, 00202h
			dd	00480h, 00802h, 00210h, 00280h
			dd	04001h, 04010h, 04040h, 08002h

			;----------------------------------
			;SPACE  CLR/HOME  INST/DEL  CONTROL
			;RUN_STOP  RETURN  C=  CRSR_DOWN  
			;CRSR_RIGHT  F1  * +  
			;F3  -  F5  F7
			;----------------------------------

			dd	08010h, 04008h, 00101h, 08004h
			dd	08080h, 00102h, 08020h, 00180h
			dd	00104h, 00110h, 04002h, 02001h
			dd	00120h, 02008h, 00140h, 00108h

			;---------------------
			;0 ... 9 : ; , = . / @
			;---------------------

			dd	01008h, 08001h, 08008h, 00201h
			dd	00208h, 00401h, 00408h, 00801h
			dd	00808h, 01001h, 02020h, 04004h
			dd	02080h, 04020h, 02010h, 04080h
			dd	02040h



;****************************************************************************
;* link_cia1_irq ************************************************************
;****************************************************************************
section .text
link_cia1_irq:
			push	dword eax
			test	byte  al, al
			jnz	.set
.clear:			mov	byte  al, C64_IRQ_CIA1
			call	cpu65xx_signal_clear_irq
			pop	dword eax
			ret
.set:			mov	byte  al, C64_IRQ_CIA1
			call	cpu65xx_signal_set_irq
			pop	dword eax
			ret



;****************************************************************************
;* link_cia1_pra, link_cia1_prb *********************************************
;****************************************************************************
section .text
link_cia1_pra:
link_cia1_prb:
			push	dword eax
			push	dword ebx
			push	dword ecx
			push	dword edx
.common:		push	dword edi
			mov	dword ebx, cd__c64_key_codes
			mov	dword edi, cia6526__c64_1
			mov	byte  ah, [CIA6526_LATCH_PRA]
			mov	byte  dh, [CIA6526_LATCH_PRB]
			xor	byte  ah, NOT8
			xor	byte  dh, NOT8
			and	byte  ah, [CIA6526_REGISTERS + CIA6526_DDRA]
			and	byte  dh, [CIA6526_REGISTERS + CIA6526_DDRB]
			xor	byte  al, al
			xor	byte  dl, dl
.next_key_code:		mov	dword ecx, [ebx]
			test	dword ecx, ecx
			jz	.end
			test	byte  ch, ah
			jz	._1
			or	byte  dl, cl
._1:			test	byte  cl, dh
			jz	._2
			or	byte  al, ch
._2:			add	byte  bl, 4
			jnz	.next_key_code
.end:			xor	byte  al, NOT8
			xor	byte  dl, NOT8
			call	cia6526_signal_pra
			mov	byte  al, dl
			call	cia6526_signal_prb
			pop	dword edi
			pop	dword edx
			pop	dword ecx
			pop	dword ebx
			pop	dword eax
			ret



;****************************************************************************
;* link_cia1_sp *************************************************************
;****************************************************************************
section .text
link_cia1_sp:
			ret



;****************************************************************************
;* link_cia1_pc *************************************************************
;****************************************************************************
section .text
link_cia1_pc:
			ret



;****************************************************************************
;* link_cia2_irq ************************************************************
;****************************************************************************
section .text
link_cia2_irq:
			push	dword eax
			test	byte  al, al
			jz	.end
.set:			call	cpu65xx_signal_set_nmi
.end:			pop	dword eax
			ret



;****************************************************************************
;* link_cia2_pra ************************************************************
;****************************************************************************
section .text
link_cia2_pra:
			push	dword eax
			push	dword ebx
			mov	byte  al, [CIA6526_LATCH_PRA]
			mov	byte  bl, [CIA6526_REGISTERS + CIA6526_DDRA]
			and	byte  al, bl
			xor	byte  bl, NOT8
			or	byte  al, bl
			xor	byte  al, NOT8
			and	byte  al, 3
			call	vic6569_signal_high_bits
			mov	byte  al, [CIA6526_LATCH_PRA]
			mov	byte  bl, [CIA6526_REGISTERS + CIA6526_DDRA]
			and	byte  al, bl
			xor	byte  bl, NOT8
			or	byte  al, bl
			and	byte  al, 038h
			shr	byte  al, 3
			mov	byte  [b__c64_serial_bus], al
			jmp	serial_bus



;****************************************************************************
;* link_cia2_prb ************************************************************
;****************************************************************************
section .text
link_cia2_prb:
			ret



;****************************************************************************
;* link_cia2_sp *************************************************************
;****************************************************************************
section .text
link_cia2_sp:
			ret



;****************************************************************************
;* link_cia2_pc *************************************************************
;****************************************************************************
section .text
link_cia2_pc:
			ret



;****************************************************************************
;* link_vic_irq *************************************************************
;****************************************************************************
section .text
link_vic_irq:
			push	dword eax
			test	byte  al, al
			jnz	.set
.clear:			mov	byte  al, C64_IRQ_VIC
			call	cpu65xx_signal_clear_irq
			pop	dword eax
			ret
.set:			mov	byte  al, C64_IRQ_VIC
			call	cpu65xx_signal_set_irq
			pop	dword eax
			ret



;****************************************************************************
;* link_via1_irq ************************************************************
;****************************************************************************
section .text
link_via1_irq:
			push	dword eax
			test	byte  al, al
			jnz	.set
.clear:			mov	byte  al, VC1541_IRQ_VIA1
			call	cpu65xx_signal_clear_irq
			pop	dword eax
			ret
.set:			mov	byte  al, VC1541_IRQ_VIA1
			call	cpu65xx_signal_set_irq
			pop	dword eax
			ret



;****************************************************************************
;* link_via1_pb *************************************************************
;****************************************************************************
section .text
link_via1_pb:
			push	dword eax
			push	dword ebx
			mov	byte  al, [VIA6522_LATCH_PB]
			mov	byte  bl, [VIA6522_REGISTERS + VIA6522_DDRB]
			and	byte  al, bl
			xor	byte  bl, NOT8
			or	byte  al, bl
			xor	byte  bl, bl
			test	byte  al, 010h
			jz	.no_atn
			or	byte  bl, 1
.no_atn:		test	byte  al, 008h
			jz	.no_clk
			or	byte  bl, 2
.no_clk:		test	byte  al, 002h
			jz	.no_data
			or	byte  bl, 4
.no_data:		mov	byte  [b__vc1541_serial_bus], bl
			jmp	serial_bus



;****************************************************************************
;* link_via1_pa *************************************************************
;****************************************************************************
section .text
link_via1_pa:
			ret



;****************************************************************************
;* link_via1_cb2 ************************************************************
;****************************************************************************
section .text
link_via1_cb2:
			ret



;****************************************************************************
;* link_via1_ca2 ************************************************************
;****************************************************************************
section .text
link_via1_ca2:
			ret



;****************************************************************************
;* link_via2_irq ************************************************************
;****************************************************************************
section .text
link_via2_irq:
			push	dword eax
			test	byte  al, al
			jnz	.set
.clear:			mov	byte  al, VC1541_IRQ_VIA2
			call	cpu65xx_signal_clear_irq
			pop	dword eax
			ret
.set:			mov	byte  al, VC1541_IRQ_VIA2
			call	cpu65xx_signal_set_irq
			pop	dword eax
			ret



;****************************************************************************
;* link_via2_pb *************************************************************
;****************************************************************************
section .text
link_via2_pb:
			push	dword eax
			mov	byte  al, [VIA6522_REGISTERS + VIA6522_PB]
			mov	byte  ah, al
			and	byte  al, 003h
			shr	byte  ah, 5
			call	disc_signal_head
			and	byte  ah, 003h
			mov	byte  al, ah
			call	disc_signal_speed
			pop	dword eax
			ret



;****************************************************************************
;* link_via2_pa *************************************************************
;****************************************************************************
section .text
link_via2_pa:
			push	dword eax
			push	dword ebx
			mov	byte  al, [VIA6522_LATCH_PA]
			mov	byte  bl, [VIA6522_REGISTERS + VIA6522_DDRA]
			and	byte  al, bl
			xor	byte  bl, NOT8
			or	byte  al, bl
			call	disc_signal_data
			pop	dword ebx
			pop	dword eax
			ret



;****************************************************************************
;* link_via2_cb2 ************************************************************
;****************************************************************************
section .text
link_via2_cb2:
			call	disc_signal_rw
			ret



;****************************************************************************
;* link_via2_ca2 ************************************************************
;****************************************************************************
section .text
link_via2_ca2:
			call	disc_signal_byte_ready
			ret



;****************************************************************************
;* link_disc_sync ***********************************************************
;****************************************************************************
section .text
link_disc_sync:
			push	dword eax
			push	dword edi
			mov	dword edi, via6522__vc1541_2
			mov	byte  ah, [VIA6522_SIGNAL_PB]
			and	byte  ah, 07fh
			shl	byte  al, 7
			or	byte  al, ah
			or	byte  al, 010h
			call	via6522_signal_pb
			pop	dword edi
			pop	dword eax
			ret



;****************************************************************************
;* link_disc_data ***********************************************************
;****************************************************************************
section .text
link_disc_data:
			push	dword edi
			mov	dword edi, via6522__vc1541_2
			call	via6522_signal_pa
			pop	dword edi
			ret



;****************************************************************************
;* link_vc1541_led **********************************************************
;****************************************************************************
section .text
link_vc1541_led:
			push	dword eax
			push	dword ebx
			push	dword edi
			mov	dword edi, via6522__vc1541_2
			mov	byte  al, [VIA6522_LATCH_PB]
			mov	byte  bl, [VIA6522_REGISTERS + VIA6522_DDRB]
			and	byte  al, bl
			xor	byte  bl, NOT8
			or	byte  al, bl
			shr	byte  al, 2
			and	byte  al, 1
			call	vic6569_led1
			mov	byte  al, [VIA6522_LATCH_PB]
			mov	byte  bl, [VIA6522_REGISTERS + VIA6522_DDRB]
			and	byte  al, bl
			xor	byte  bl, NOT8
			or	byte  al, bl
			shr	byte  al, 3
			and	byte  al, 1
			call	vic6569_led2
			mov	byte  al, [DISC_TRACK]
			call	vic6569_number1
			pop	dword edi
			pop	dword ebx
			pop	dword eax
			ret



;****************************************************************************
;* serial_bus ***************************************************************
;****************************************************************************
section .text
serial_bus:
			push	dword edi
			mov	byte  al, [b__c64_serial_bus]
			mov	byte  bl, [b__vc1541_serial_bus]
			test	byte  bl, 001h
			jz	.skip_set
			mov	byte  [flag__atn_cleared], 1
.skip_set:		test	byte  al, 001h
			jz	.clear
			test	byte  [flag__atn_cleared], NOT8
			jnz	.no_data_influence
			or	byte  al, 4
			jmp	.no_data_influence
.clear:			mov	byte  [flag__atn_cleared], 0
.no_data_influence:	and	byte  bl, 0feh
			or	byte  al, bl
			mov	dword edi, cia6526__c64_2
			mov	byte  bl, al
			xor	byte  al, NOT8
			shl	byte  al, 5
			and	byte  al, 0c0h
			call	cia6526_signal_pra
			mov	byte  al, bl
			mov	dword edi, via6522__vc1541_1
			and	byte  al, 001h
			call	via6522_signal_ca1
			xor	byte  al, al
			test	byte  bl, 1
			jz	.no_atn
			or	byte  al, 080h
.no_atn:		test	byte  bl, 2
			jz	.no_clk
			or	byte  al, 004h
.no_clk:		test	byte  bl, 4
			jz	.no_data
			or	byte  al, 001h
.no_data:		call	via6522_signal_pb
			pop	dword edi
			pop	dword ebx
			pop	dword eax
			ret
;********************************************* karsten.scheibler@bigfoot.de *

