;****************************************************************************
;****************************************************************************
;*
;* main.nasm
;*
;****************************************************************************
;****************************************************************************





;%define MAIN_TEST_VIA





%include "sys_calls.inc"
%include "c64_memory.inc"
%include "cpu65xx.inc"
%include "vic6569.inc"
%include "sid6581.inc"
%include "cia6526.inc"
%include "vc1541_memory.inc"
%include "via6522.inc"
%include "disc.inc"
%include "write.inc"
%include "tty.inc"
%include "link.inc"
%include "cmdline.inc"

%define MAIN_NASM
%include "main.inc"



global _start



;****************************************************************************
;* main_exit ****************************************************************
;****************************************************************************
section .text
main_exit:		call	tty_exit
			call	c64_memory_exit
			call	cpu65xx_exit
			call	vic6569_exit
			call	sid6581_exit
			call	cia6526_exit
			call	vc1541_memory_exit
			call	via6522_exit
			call	disc_exit
			SYS_EXIT



;****************************************************************************
;* main_suspend *************************************************************
;****************************************************************************
section .text
main_suspend:		jmp	main_exit



;****************************************************************************
;* main_exception ***********************************************************
;****************************************************************************
section .text
main_exception:
			push	dword eax
			call	tty_clear_screen
			WRITE_TEXT  .cb__exception, .exception_length
			pop	dword eax
			WRITE_HEX  8
			WRITE_RETURN
			mov	dword esi, cpu65xx__c64
			call	cpu65xx_dump
			call	vic6569_dump
			call	sid6581_dump
			mov	dword edi, cia6526__c64_1
			call	cia6526_dump
			mov	dword edi, cia6526__c64_2
			call	cia6526_dump
			mov	dword esi, cpu65xx__vc1541
			call	cpu65xx_dump
			mov	dword edi, via6522__vc1541_1
			call	via6522_dump
			mov	dword edi, via6522__vc1541_2
			call	via6522_dump
			call	disc_dump
			WRITE_RETURN
			WRITE_TEXT  .cb__c64_cycles, .c64_cycles_length
			WRITE_HEX  8, [q__c64_cycles + 4]
			WRITE_HEX  8, [q__c64_cycles]
			WRITE_RETURN
			WRITE_TEXT  .cb__vc1541_cycles, .vc1541_cycles_length
			WRITE_HEX  8, [q__vc1541_cycles + 4]
			WRITE_HEX  8, [q__vc1541_cycles]
			WRITE_RETURN
			jmp	main_exit

section .data
.cb__exception:		db	"An Exception occured in Modul "
.exception_length:	equ	$ - .cb__exception
.cb__c64_cycles:	db	"Overall emulated C64 Cycles:    "
.c64_cycles_length:	equ	$ - .cb__c64_cycles
.cb__vc1541_cycles:	db	"Overall emulated VC1541 Cycles: "
.vc1541_cycles_length:	equ	$ - .cb__vc1541_cycles



;****************************************************************************
;* main_save_snapshot *******************************************************
;****************************************************************************
section .text
main_save_snapshot:
;			call	c64_memory_save_snapshot
;			mov	dword esi, cpu65xx__c64_cpu
;			call	cpu65xx_save_snapshot
;			call	vic6569_save_snapshot
;			call	sid6581_save_snapshot
;			mov	dword edi, cia6526__c64_cia1
;			call	cia6526_save_snapshot
;			mov	dword edi, cia6526__c64_cia2
;			call	cia6526_save_snapshot
;			mov	dword esi, cpu65xx__vc1541_cpu
;			call	vc1541_memory_save_snapshot
;			call	cpu65xx_save_snapshot
;			mov	dword edi, via6522__vc1541_via1
;			call	via6522_save_snapshot
;			mov	dword edi, via6522__vc1541_via2
;			call	via6522_save_snapshot
;			call	disc_save_snapshot
			;save data
			ret



;****************************************************************************
;* main_load_snapshot *******************************************************
;****************************************************************************
section .text
main_load_snapshot:
			ret



;****************************************************************************
;* main_start ***************************************************************
;****************************************************************************
section .text
_start:
main_start:
			;--------------------------------
			;initialize the emulation modules
			;--------------------------------

			call	c64_memory_initialize
			call	cpu65xx_initialize
			call	vic6569_initialize
			call	sid6581_initialize
			call	cia6526_initialize
			call	vc1541_memory_initialize
			call	via6522_initialize
			call	disc_initialize

			;-----------------------------
			;process commandline arguments
			;-----------------------------

			CMDLINE_PARSE

			;---------------------------
			;press enter to continue ...
			;---------------------------

			WRITE_TEXT  .cb__press_any_key, .text_length
			call	tty_press_enter

			;-------------------------------
			;initialize the keyboard handler
			;-------------------------------

			call	tty_initialize

			;--------------------------------------------------
			;reset all emulated devices. remember: cpu, cia and
			;via are emulated with the same programcode but
			;with different data for each device, the according
			;pointer is loaded into esi for the cpu and edi for
			;cia and via
			;--------------------------------------------------

			C64_MEMORY_CALL_RESET  c64_memory__c64, 0dddd0001h
			CPU65XX_CALL_RESET  cpu65xx__c64, 065100001h, c64_memory_read, c64_memory_write
			VIC6569_CALL_RESET  vic6569__c64, 065690001h
			SID6581_CALL_RESET  sid6581__c64, 065810001h
			CIA6526_CALL_RESET  cia6526__c64_1, 065260001h, link_cia1_irq, link_cia1_pra, link_cia1_prb
			CIA6526_CALL_RESET  cia6526__c64_2, 065260002h, link_cia2_irq, link_cia2_pra, link_cia2_prb
%ifdef MAIN_TEST_VIA
			VIA6522_CALL_RESET  via6522__c64_1, 06522ffffh, link_cia2_irq, link_nothing, link_nothing, link_nothing, link_nothing
%endif
			VC1541_MEMORY_CALL_RESET  vc1541_memory__vc1541, 0eeee0001h
			CPU65XX_CALL_RESET  cpu65xx__vc1541, 065020001h, vc1541_memory_read, vc1541_memory_write
			VIA6522_CALL_RESET  via6522__vc1541_1, 065220001h, link_via1_irq, link_via1_pb, link_via1_cb2, link_via1_pa, link_via1_ca2
			VIA6522_CALL_RESET  via6522__vc1541_2, 065220002h, link_via2_irq, link_via2_pb, link_via2_cb2, link_via2_pa, link_via2_ca2
			DISC_CALL_RESET  disc__vc1541, 0ffff0001h

			;----------------------------
			;initialize the clockcounters
			;----------------------------

			mov	dword eax, C64_CLOCK
			mov	dword ebx, VC1541_CLOCK
			mov	dword [d__c64_clock], eax
			mov	dword [d__c64_clock_counter], eax
			mov	dword [d__vc1541_clock], ebx

			;-----------------------------
			;enter to the clock_cycle_loop
			;-----------------------------

.clock_cycle_loop:
			mov	dword eax, [d__c64_clock_counter]
			cmp	dword eax, [d__vc1541_clock]
			jae	.run_c64_cycle
.run_vc1541_cycle:	
			;----------------------
			;one vc1541 clock cycle
			;----------------------

			add	dword eax, [d__c64_clock]
			mov	dword esi, cpu65xx__vc1541
			mov	dword [d__c64_clock_counter], eax
			CPU65XX_CALL_CYCLE
			VIA6522_CALL_CYCLE  via6522__vc1541_1
			VIA6522_CALL_CYCLE  via6522__vc1541_2
			DISC_CALL_CYCLE
			add	dword [q__vc1541_cycles], 1
			adc	dword [q__vc1541_cycles + 4], 0
			jmp	.clock_cycle_loop
.run_c64_cycle:		
			;-------------------
			;one c64 clock cycle
			;-------------------

			sub	dword eax, [d__vc1541_clock]
			mov	dword esi, cpu65xx__c64
			mov	dword [d__c64_clock_counter], eax
			CPU65XX_CALL_CYCLE
			VIC6569_CALL_CYCLE
			SID6581_CALL_CYCLE
			CIA6526_CALL_CYCLE  cia6526__c64_1
			CIA6526_CALL_CYCLE  cia6526__c64_2
%ifdef MAIN_TEST_VIA
			VIA6522_CALL_CYCLE  via6522__c64_1
%endif
			add	dword [q__c64_cycles], 1
			mov	dword eax, [d__c64_clock_counter]
			adc	dword [q__c64_cycles + 4], 0
			jmp	.run_vc1541_cycle

			align 4
.cb__press_any_key:	db	"Here would be in later versions displayed some informations ...", 10
			db	"Press ENTER to continue ...", 10
.text_length:		equ	$ - .cb__press_any_key


;****************************************************************************
;* some counters ************************************************************
;****************************************************************************
section .data
			align 4
q__c64_cycles:		dd	0, 0
q__vc1541_cycles:	dd	0, 0
d__c64_clock:		dd	0
d__vc1541_clock:	dd	0
d__c64_clock_counter:	dd	0



;****************************************************************************
;* Data of all needed Modules ***********************************************
;****************************************************************************
section .bss
			alignb 32
c64_memory__c64:	resb	c64_memory_data_size

			alignb 32
cpu65xx__c64:		resb	cpu65xx_data_size

			alignb 32
vic6569__c64:		resb	vic6569_data_size

			alignb 32
sid6581__c64:		resb	sid6581_data_size

			alignb 32
cia6526__c64_1:		resb	cia6526_data_size

			alignb 32
cia6526__c64_2:		resb	cia6526_data_size

			alignb 32
via6522__c64_1:		resb	via6522_data_size

			alignb 32
vc1541_memory__vc1541:	resb	vc1541_memory_data_size
			
			alignb 32
cpu65xx__vc1541:	resb	cpu65xx_data_size

			alignb 32
via6522__vc1541_1:	resb	via6522_data_size

			alignb 32
via6522__vc1541_2:	resb	via6522_data_size

			alignb 32
disc__vc1541:		resb	disc_data_size
;********************************************* karsten.scheibler@bigfoot.de *

