;****************************************************************************
;****************************************************************************
;*
;* vc1541_memory.nasm
;*
;****************************************************************************
;****************************************************************************




%ifndef VC1541_MEMORY_TEST
%include "via6522.inc"
%include "main.inc"
%endif

%define VC1541_MEMORY_NASM
%include "vc1541_memory.inc"





;****************************************************************************
;****************************************************************************
;*
;* PART 1: 2KB RAM, Kernel ROM
;*
;****************************************************************************
;****************************************************************************





section .data
;ab__vc1541_kernel_rom:	incbin "../roms/vc1541-2_kernel.rom"
ab__vc1541_kernel_rom:	incbin "../roms/vc1541_kernel.rom"

section .bss
ab__vc1541_ram:		resb	00800h





;****************************************************************************
;****************************************************************************
;*
;* PART 2: The Memory Access Routines
;*
;****************************************************************************
;****************************************************************************





;****************************************************************************
;* vc1541_memory_initialize *************************************************
;****************************************************************************
section .text
vc1541_memory_initialize:
			call	vc1541_memory_reset
			ret



;****************************************************************************
;* vc1541_memory_exit *******************************************************
;****************************************************************************
section .text
vc1541_memory_exit:
			ret



;****************************************************************************
;* vc1541_memory_save_snapshot **********************************************
;****************************************************************************
section .text
vc1541_memory_save_snapshot:
			ret



;****************************************************************************
;* vc1541_memory_loas_snapshot **********************************************
;****************************************************************************
section .text
vc1541_memory_load_snapshot:
			ret



;****************************************************************************
;* vc1541_memory_reset ******************************************************
;****************************************************************************
section .text
vc1541_memory_reset:
			push	dword eax
			push	dword ecx
			push	dword edi
			xor	dword eax, eax
			mov	dword ecx, 00200h
			mov	dword edi, ab__vc1541_ram
			cld
			rep stosd
			pop	dword edi
			pop	dword ecx
			pop	dword eax
			ret



;****************************************************************************
;* vc1541_memory_read *******************************************************
;****************************************************************************
section .text
vc1541_memory_read:
			and	dword ebx, 0ffffh
			cmp	dword ebx, 0c000h
			jae	.read_kernel_rom
			cmp	dword ebx, 00800h
			jb	.read_ram_1
			cmp	dword ebx, 01000h
			jb	.read_ram_2
			cmp	dword ebx, 02000h
			jae	.read_rubbish
			cmp	dword ebx, 01c00h
			jae	.read_via2
			cmp	dword ebx, 01800h
			jae	.read_via1
.read_rubbish:		mov	byte  al, bh
			ret
.read_ram_1:		mov	byte  al, [ab__vc1541_ram + ebx]
			ret
.read_ram_2:		mov	byte  al, [ab__vc1541_ram - 0800h + ebx]
			ret
.read_via1:		mov	dword edi, via6522__vc1541_1
			call	via6522_read
			ret
.read_via2:		mov	dword edi, via6522__vc1541_2
			call	via6522_read
			ret
.read_kernel_rom:	mov	byte  al, [ab__vc1541_kernel_rom - 0c000h + ebx]
			ret



;****************************************************************************
;* vc1541_memory_write ******************************************************
;****************************************************************************
section .text
vc1541_memory_write:
			and	dword ebx, 0ffffh
			cmp	dword ebx, 00800h
			jb	.write_ram_1
			cmp	dword ebx, 01000h
			jb	.write_ram_2
			cmp	dword ebx, 02000h
			jae	.write_ignore
			cmp	dword ebx, 01c00h
			jae	.write_via2
			cmp	dword ebx, 01800h
			jae	.write_via1
.write_ignore:		ret
.write_ram_1:		mov	byte  [ab__vc1541_ram + ebx], al
			ret
.write_ram_2:		mov	byte  [ab__vc1541_ram - 00800h + ebx], al
			ret
.write_via1:		mov	dword edi, via6522__vc1541_1
			call	via6522_write
			ret
.write_via2:		mov	dword edi, via6522__vc1541_2
			call	via6522_write
			ret





;****************************************************************************
;****************************************************************************
;*
;* PART 3: Code for Selftest
;*
;****************************************************************************
;****************************************************************************





%ifdef VC1541_MEMORY_TEST
%error "No Test Code implemented!"
%endif
;********************************************* karsten.scheibler@bigfoot.de *

