;****************************************************************************
;****************************************************************************
;*
;* via6522.nasm
;*
;****************************************************************************
;****************************************************************************


;TODO:
;
;* Timerin-/output on PB7
;* latching of PA/PB
;* serial register
;* register handshakes




;%define VIA6522_DEBUG





%include "defines.inc"
%include "main.inc"
%include "link.inc"

%define VIA6522_NASM
%include "via6522.inc"



;****************************************************************************
;* assign's & defines's *****************************************************
;****************************************************************************
%assign PB		VIA6522_PB
%assign PA		VIA6522_PA
%assign DDRB		VIA6522_DDRB
%assign DDRA		VIA6522_DDRA
%assign T1C_LO		VIA6522_T1C_LO
%assign T1C_HI		VIA6522_T1C_HI
%assign T1L_LO		VIA6522_T1L_LO
%assign T1L_HI		VIA6522_T1L_HI
%assign T2C_LO		VIA6522_T2C_LO
%assign T2C_HI		VIA6522_T2C_HI
%assign SR		VIA6522_SR
%assign ACR		VIA6522_ACR
%assign PCR		VIA6522_PCR
%assign IFR		VIA6522_IFR
%assign IER		VIA6522_IER
%assign PA_NH		VIA6522_PA_NH	;PA no handshake

%assign STATE_CB1	001h
%assign STATE_CA1	002h
%assign GENERATE_IRQ	004h
%assign LOAD_T1		008h
%assign LOAD_T2		010h
%assign UNDERFLOW_T1	020h
%assign UNDERFLOW_T2	040h




%define MODUL_ID	VIA6522_MODUL_ID
%define LINK_IRQ	VIA6522_LINK_IRQ
%define LINK_PB		VIA6522_LINK_PB
%define LINK_CB2	VIA6522_LINK_CB2
%define LINK_PA		VIA6522_LINK_PA
%define LINK_CA2	VIA6522_LINK_CA2
%define STATE		VIA6522_STATE
%define LATCH_PB	VIA6522_LATCH_PB
%define SIGNAL_PB	VIA6522_SIGNAL_PB
%define LATCH_PA	VIA6522_LATCH_PA
%define SIGNAL_PA	VIA6522_SIGNAL_PA
%define LATCH_T2_LO	VIA6522_LATCH_T2_LO
%define REGISTERS	VIA6522_REGISTERS
%define IRQ_MASK	VIA6522_IRQ_MASK



;****************************************************************************
;* via6522_initialize *******************************************************
;****************************************************************************
section .text
via6522_initialize:
			ret



;****************************************************************************
;* via6522_exit *************************************************************
;****************************************************************************
section .text
via6522_exit:
			ret



;****************************************************************************
;* via6522_save_snapshot ****************************************************
;****************************************************************************
section .text
via6522_save_snapshot:
			ret



;****************************************************************************
;* via6522_load_snapshot ****************************************************
;****************************************************************************
section .text
via6522_load_snapshot:
			ret



%ifdef VIA6522_DEBUG
%include "sys_calls.inc"
;****************************************************************************
;* via6522_debug ************************************************************
;****************************************************************************
section .text
via6522_debug:
.read:			push	dword eax
			mov	byte  ah, 0ffh
			mov	byte  al, ah
			jmp	.common
.write:			push	dword eax
			xor	byte  ah, ah
.common:		cmp	dword edi, via6522__vc1541_2
			jne	.end
			mov	byte  [.ab__status], ah
			mov	byte  [.ab__status + 1], al
			mov	word  [.ab__status + 2], bx
			push	dword ebx
			push	dword ecx
			push	dword edx
			SYS_WRITE  2, .ab__status, 4
			pop	dword edx
			pop	dword ecx
			pop	dword ebx
.end:			pop	dword eax
			ret

section .data
			align 4
.ab__status:		times 4 db 0
%endif



;****************************************************************************
;* via6522_reset ************************************************************
;****************************************************************************
section .text
via6522_reset:
			xor	dword eax, eax
			xor	dword ecx, ecx
			push	dword edi
			mov	byte  cl, 010h
			lea	dword edi, [REGISTERS]
			cld
			rep stosb
			pop	dword edi
			mov	dword [STATE], eax
			mov	byte  [LATCH_PB], al
			mov	byte  [SIGNAL_PB], al
			mov	byte  [LATCH_PA], al
			mov	byte  [SIGNAL_PA], al
			mov	byte  [LATCH_T2_LO], al
			mov	byte  [IRQ_MASK], al
			ret



;****************************************************************************
;* via6522_cycle ************************************************************
;****************************************************************************
section .text
via6522_cycle:
			mov	dword eax, [STATE]
			test	dword eax, GENERATE_IRQ
			jz	.no_irq
			mov	byte  bl, [REGISTERS + IFR]
			and	dword eax, (NOT32 - GENERATE_IRQ)
			or	byte  bl, 080h
			push	dword eax
			mov	byte  [REGISTERS + IFR], bl
			mov	byte  al, 1
			call	dword [LINK_IRQ]
			pop	dword eax
.no_irq:		xor	dword ebx, ebx
			xor	dword ecx, ecx
			mov	word  bx, [REGISTERS + T1C_LO]
			mov	word  cx, [REGISTERS + T2C_LO]
			mov	byte  dl, [REGISTERS + IFR]
.test_t1:		test	dword eax, LOAD_T1
			jz	.test_t2
			mov	word  bx, [REGISTERS + T1L_LO]
			and	dword eax, (NOT32 - LOAD_T1)
			inc	dword ebx
.test_t2:		test	dword eax, LOAD_T2
			jz	.count_t1
			mov	byte  cl, [LATCH_T2_LO]
			and	dword eax, (NOT32 - LOAD_T2)
			mov	byte  ch, [REGISTERS + T2C_HI]
			inc	dword ecx
.count_t1:		dec	dword ebx
			jns	.count_t2
			test	byte  [REGISTERS + ACR], 040h
			jnz	.continuous_mode
			test	dword eax, UNDERFLOW_T1
			jnz	.count_t2
			or	dword eax, UNDERFLOW_T1
			jmp	.common
.continuous_mode:	or	dword eax, LOAD_T1
.common:		or	byte  dl, 040h
			test	byte  [IRQ_MASK], 040h
			jz	.count_t2
			or	dword eax, GENERATE_IRQ
.count_t2:		dec	dword ecx
			jns	.end
			test	dword eax, UNDERFLOW_T2
			jnz	.end
			or	dword eax, UNDERFLOW_T2
			or	byte  dl, 020h
			test	byte  [IRQ_MASK], 020h
			jz	.end
			or	dword eax, GENERATE_IRQ
.end:			mov	word  [REGISTERS + T1C_LO], bx
			mov	word  [REGISTERS + T2C_LO], cx
			mov	byte  [REGISTERS + IFR], dl
			mov	dword [STATE], eax
			ret



;****************************************************************************
;* via6522_dump *************************************************************
;****************************************************************************
section .text
via6522_dump:
			ret



;****************************************************************************
;* Macro: CLEAR_IRQ *********************************************************
;****************************************************************************
%macro CLEAR_IRQ 1-2
			test	byte  [REGISTERS + IFR], %1
			jz	%%no_irq
%if %0 > 1
			push	dword eax
%endif
			mov	byte  al, [REGISTERS + IFR]
			and	byte  al, (NOT8 - %1)
			test	byte  al, [IRQ_MASK]
			jnz	%%irq_left
			and	byte  al, 07fh
			push	dword eax
			xor	byte  al, al
			call	dword [LINK_IRQ]
			pop	dword eax
%%irq_left:		mov	byte  [REGISTERS + IFR], al
%if %0 > 1
			pop	dword eax
%endif
%%no_irq:
%endmacro



;****************************************************************************
;* via6522_read *************************************************************
;****************************************************************************
section .text
via6522_read:
%ifdef VIA6522_DEBUG
			call	via6522_debug.read
%endif
			mov	dword ebp, ebx
			and	dword ebp, 00000000fh
			jmp	dword [.ap__registers + 4 * ebp]
.ap__registers:		dd	.read_pb
			dd	.read_pa
			dd	.read_register
			dd	.read_register
			dd	.read_t1c_lo
			dd	.read_register
			dd	.read_register
			dd	.read_register
			dd	.read_t2c_lo
			dd	.read_register
			dd	.read_register
			dd	.read_register
			dd	.read_register
			dd	.read_register
			dd	.read_ier
			dd	.read_register

			;----------------------
			;ordinary register read
			;----------------------

.read_register:		mov	byte  al, [REGISTERS + ebp]
			ret

			;-------
			;read PB
			;-------

.read_pb:		CLEAR_IRQ  010h
			mov	byte  al, [REGISTERS + PB]
			ret

			;-------
			;read PA
			;-------

.read_pa:		CLEAR_IRQ  002h
			mov	byte  al, [REGISTERS + PA]
			ret

			;----------------------------------
			;read T1C_LO and reset the IRQ Flag
			;----------------------------------

.read_t1c_lo:		CLEAR_IRQ  040h
			mov	byte  al, [REGISTERS + T1C_LO]
			ret

			;----------------------------------
			;read T2C_LO and reset the IRQ Flag
			;----------------------------------

.read_t2c_lo:		CLEAR_IRQ  020h
			mov	byte  al, [REGISTERS + T2C_LO]
			ret

			;--------
			;read IER
			;--------

.read_ier:		mov	byte  al, [IRQ_MASK]
			or	byte  al, 080h
			ret



;****************************************************************************
;* via6522_write ************************************************************
;****************************************************************************
section .text
via6522_write:
%ifdef VIA6522_DEBUG
			call	via6522_debug.write
%endif
			mov	dword ebp, ebx
			and	dword ebp, 00000000fh
			jmp	dword [.ap__registers + 4 * ebp]
.ap__registers:		dd	.write_pb
			dd	.write_pa
			dd	.write_ddrb
			dd	.write_ddra
			dd	.write_t1c_lo
			dd	.write_t1c_hi
			dd	.write_register
			dd	.write_register
			dd	.write_t2c_lo
			dd	.write_t2c_hi
			dd	.write_ignore
			dd	.write_register
			dd	.write_pcr
			dd	.write_ifr
			dd	.write_ier
			dd	.write_pa_nh

			;-------------
			;ignore writes
			;-------------

.write_ignore:		ret

			;-----------------------
			;ordinary register write
			;-----------------------

.write_register:	mov	byte  [REGISTERS + ebp], al
			ret

			;------
			;set PB
			;------

.write_pb:		;handshake!!
			mov	byte  [LATCH_PB], al
			CLEAR_IRQ  010h, PUSH_EAX
			mov	byte  ah, [REGISTERS + DDRB]
			and	byte  al, ah
			xor	byte  ah, NOT8
			and	byte  ah, [SIGNAL_PB]
			or	byte  al, ah
			mov	byte  [REGISTERS + PB], al
			call	dword [LINK_PB]
			ret

			;------
			;set PA
			;------

.write_pa:		;handshake!!
			mov	byte  [LATCH_PA], al
			CLEAR_IRQ  002h, PUSH_EAX
			mov	byte  ah, [REGISTERS + DDRA]
			and	byte  al, ah
			xor	byte  ah, NOT8
			and	byte  ah, [SIGNAL_PA]
			or	byte  al, ah
			mov	byte  [REGISTERS + PA], al
			call	dword [LINK_PA]
			ret

			;--------
			;set DDRB
			;--------

.write_ddrb:		mov	byte  [REGISTERS + DDRB], al
			mov	byte  ah, [LATCH_PB]
			and	byte  ah, al
			xor	byte  al, NOT8
			and	byte  al, [SIGNAL_PB]
			or	byte  ah, al
			mov	byte  [REGISTERS + PB], ah
			call	dword [LINK_PB]
			ret

			;--------
			;set DDRA
			;--------

.write_ddra:		mov	byte  [REGISTERS + DDRA], al
			mov	byte  ah, [LATCH_PA]
			and	byte  ah, al
			xor	byte  al, NOT8
			and	byte  al, [SIGNAL_PA]
			or	byte  ah, al
			mov	byte  [REGISTERS + PA], ah
			call	dword [LINK_PA]
			ret

			;------------------------------
			;write to T1C_LO goes to T1L_LO
			;------------------------------

.write_t1c_lo:		mov	byte  [REGISTERS + T1L_LO], al
			ret

			;------------------------------------
			;write to T1C_HI which goes to T1L_HI
			;and reset the IRQ Flag and load the
			;counter from the latch
			;------------------------------------

.write_t1c_hi:		mov	byte  [REGISTERS + T1L_HI], al
			mov	dword ebp, [STATE]
			CLEAR_IRQ  040h
			and	dword ebp, (NOT32 - UNDERFLOW_T1)
			or	dword ebp, LOAD_T1
			mov	dword [STATE], ebp
			ret

			;-----------------------------------
			;write to T2C_LO goes to LATCH_T2_LO
			;-----------------------------------

.write_t2c_lo:		mov	byte  [LATCH_T2_LO], al
			ret

			;--------------------------------------
			;write to T2C_HI and reset the IRQ Flag
			;and load the counter from the latch
			;--------------------------------------

.write_t2c_hi:		mov	byte  [REGISTERS + T2C_HI], al
			mov	dword ebp, [STATE]
			CLEAR_IRQ  020h
			and	dword ebp, (NOT32 - UNDERFLOW_T2)
			or	dword ebp, LOAD_T2
			mov	dword [STATE], ebp
			ret

			;-------
			;set PCR
			;-------

.write_pcr:		mov	byte  [REGISTERS + PCR], al
			push	dword ebx
			mov	byte  ah, al
			and	byte  al, 00eh
			xor	dword ebx, ebx
			cmp	byte  al, 00eh
			jne	.write_pcr_1
			mov	byte  bl, 1
.write_pcr_1:		and	byte  ah, 0e0h
			cmp	byte  ah, 0e0h
			jne	.write_pcr_2
			mov	byte  bh, 1
.write_pcr_2:		mov	byte  al, bl
			call	dword [LINK_CA2]
			mov	byte  al, bh
			call	dword [LINK_CB2]
			pop	dword ebx
			ret

			;-------
			;set IFR
			;-------

.write_ifr:		and	byte  al, 07fh
			jz	.write_ifr_2
			xor	byte  al, 07fh
			and	byte  al, [REGISTERS + IFR]
			test	byte  al, [IRQ_MASK]
			jnz	.write_ifr_1
			push	dword eax
			xor	byte  al, al
			call	dword [LINK_IRQ]
			pop	dword eax
.write_ifr_1:		mov	byte  [REGISTERS + IFR], al
.write_ifr_2:		ret

			;-------
			;set IER
			;-------

.write_ier:		test	byte  al, 080h
			jz	.write_ier_1
			and	byte  al, 07fh
			or	byte  al, [IRQ_MASK]
			jmp	.write_ier_2
.write_ier_1:		xor	byte  al, 07fh
			and	byte  al, [IRQ_MASK]
.write_ier_2:		mov	byte  [IRQ_MASK], al
			test	byte  al, [REGISTERS + IFR]
			jnz	.write_ier_3
			and	byte  [REGISTERS + IFR], 07fh
.write_ier_3:		ret

			;------------------------
			;set PA with no handshake
			;------------------------

.write_pa_nh:		mov	byte  [LATCH_PA], al
			mov	byte  ah, [REGISTERS + DDRA]
			and	byte  al, ah
			not	byte  ah
			and	byte  ah, [SIGNAL_PA]
			or	byte  al, ah
			mov	byte  [REGISTERS + PA], al
			call	dword [LINK_PA]
			ret



;****************************************************************************
;* via6522_signal_pb ********************************************************
;****************************************************************************
section .text
via6522_signal_pb:
			push	dword eax
			mov	byte  [SIGNAL_PB], al
			mov	byte  ah, [REGISTERS + DDRB]
			xor	byte  ah, NOT8
			and	byte  al, ah
			mov	byte  ah, [LATCH_PB]
			and	byte  ah, [REGISTERS + DDRB]
			or	byte  ah, al
			mov	byte  [REGISTERS + PB], ah
			pop	dword eax
			ret



;****************************************************************************
;* via6522_signal_pa ********************************************************
;****************************************************************************
section .text
via6522_signal_pa:
			push	dword eax
			mov	byte  [SIGNAL_PA], al
			mov	byte  ah, [REGISTERS + DDRA]
			xor	byte  ah, NOT8
			and	byte  al, ah
			mov	byte  ah, [LATCH_PA]
			and	byte  ah, [REGISTERS + DDRA]
			or	byte  ah, al
			mov	byte  [REGISTERS + PA], ah
			pop	dword eax
			ret



;****************************************************************************
;* via6522_signal_cb1 *******************************************************
;****************************************************************************
section .text
via6522_signal_cb1:
			push	dword eax
			push	dword ebx
			mov	byte  bl, al
			mov	dword eax, [STATE]
			xor	byte  bh, bh
			test	dword eax, STATE_CB1
			jz	.no_xor_1
			xor	byte  bl, 1
			mov	byte  bh, 1
.no_xor_1:		test	byte  bl, 1
			jz	.exit
			test	byte  [REGISTERS + PCR], 010h
			jnz	.no_xor_2
			xor	byte  bh, 1
.no_xor_2:		cmp	byte  bh, 1
			je	.exit
			mov	byte  al, 010h
			test	byte  [REGISTERS + IER], al
			jz	.no_irq
			or	dword [STATE], GENERATE_IRQ
.no_irq:		or	byte  al, [REGISTERS + IFR]
			mov	byte  [REGISTERS + IFR], al
.exit:			pop	dword ebx
			pop	dword eax
			ret



;****************************************************************************
;* via6522_signal_ca1 *******************************************************
;****************************************************************************
section .text
via6522_signal_ca1:
			push	dword eax
			push	dword ebx
			mov	byte  bl, al
			mov	dword eax, [STATE]
			xor	byte  bh, bh
			test	dword eax, STATE_CA1
			jz	.no_xor_1
			xor	byte  bl, 1
			mov	byte  bh, 1
.no_xor_1:		test	byte  bl, 1
			jz	.exit
			test	byte  [REGISTERS + PCR], 001h
			jnz	.no_xor_2
			xor	byte  bh, 1
.no_xor_2:		cmp	byte  bh, 1
			je	.exit
			mov	byte  al, 002h
			test	byte  [REGISTERS + IER], al
			jz	.no_irq
			or	dword [STATE], GENERATE_IRQ
.no_irq:		or	byte  al, [REGISTERS + IFR]
			mov	byte  [REGISTERS + IFR], al
.exit:			pop	dword ebx
			pop	dword eax
			ret
;********************************************* karsten.scheibler@bigfoot.de *

