;****************************************************************************
;****************************************************************************
;*
;* vic6569.nasm
;*
;****************************************************************************
;****************************************************************************


;TODO:
;
;* initialize over the write routines
;* implement the Lightpen
;* check $d019 behavior
;* implement sprites
;* implement Border Flags
;* implement X Scrolling
;* implement Sprite Collisions





;%define VIC6569_DEBUG





%ifdef VIC6569_DEBUG
%include "sys_calls.inc"
%endif

%include "defines.inc"
%include "sync.inc"
%include "fb.inc"
%include "main.inc"
%include "link.inc"
%include "cpu65xx.inc"
%include "c64_memory.inc"

%define VIC6569_NASM
%include "vic6569.inc"



;****************************************************************************
;* assign's & defines's *****************************************************
;****************************************************************************
%define	MODUL_ID		VIC6569_MODUL_ID
%define EXECUTE_BUS_CYCLE	VIC6569_EXECUTE_BUS_CYCLE
%define EXECUTE_SEQUENCER_CYCLE	VIC6569_EXECUTE_SEQUENCER_CYCLE
%define X_RESOLUTION		VIC6569_X_RESOLUTION
%define X_PADDING		VIC6569_X_PADDING
%define Y_RESOLUTION		VIC6569_Y_RESOLUTION
%define FB_ADDRESS		VIC6569_FB_ADDRESS
%define X_START			VIC6569_X_START
%define X_END			VIC6569_X_END
%define Y_START			VIC6569_Y_START
%define Y_END			VIC6569_Y_END
%define SCREEN_START		VIC6569_SCREEN_START
%define SCREEN			VIC6569_SCREEN
%define FRAME			VIC6569_FRAME
%define LINE_CYCLE		VIC6569_LINE_CYCLE
%define LINE_COUNTER		VIC6569_LINE_COUNTER
%define LINE_LATCH		VIC6569_LINE_LATCH
%define HIGH_BITS		VIC6569_HIGH_BITS
%define VIDEO_COUNTER		VIC6569_VIDEO_COUNTER
%define VIDEO_COUNTER_BASE	VIC6569_VIDEO_COUNTER_BASE
%define ROW_COUNTER		VIC6569_ROW_COUNTER
%define CACHE_OFFSET		VIC6569_CACHE_OFFSET
%define STATE			VIC6569_STATE
%define	BORDER_LEFT		VIC6569_BORDER_LEFT
%define BORDER_RIGHT		VIC6569_BORDER_RIGHT
%define BORDER_TOP		VIC6569_BORDER_TOP
%define BORDER_BUTTOM		VIC6569_BORDER_BUTTOM
%define BUS_CYCLE_COUNTER	VIC6569_BUS_CYCLE_COUNTER
%define SEQUENCER_CYCLE_COUNTER	VIC6569_SEQUENCER_CYCLE_COUNTER
%define DMA_DELAY_COUNTER	VIC6569_DMA_DELAY_COUNTER
%define VIDEO_BYTE		VIC6569_VIDEO_BYTE
%define COLOR_RAM_BYTE		VIC6569_COLOR_RAM_BYTE
%define DATA_BYTE		VIC6569_DATA_BYTE
%define COLOR_LOOKUP_TABLE	VIC6569_COLOR_LOOKUP_TABLE
%define VIDEO_CACHE		VIC6569_VIDEO_CACHE
%define COLOR_RAM_CACHE		VIC6569_COLOR_RAM_CACHE
%define REGISTERS		VIC6569_REGISTERS
%define FRAME_DATA		VIC6569_FRAME_DATA



;****************************************************************************
;* vic6569_initialize *******************************************************
;****************************************************************************
section .text
vic6569_initialize:
			call	fb_open
			call	fb_get_parameters
			mov	dword [X_RESOLUTION], eax
			mov	dword [X_PADDING], ebx
			mov	dword [Y_RESOLUTION], ecx
			mov	dword [FB_ADDRESS], edx

			;check resolution, 8bpp ? some virtual 
			;resolution settings ?

			;------------
			;!!!! set screen geometry manually
			;----------------

			xor	dword eax, eax
			mov	dword ebx, 312
			xor	dword ecx, ecx
			mov	dword edx, 504
			mov	dword [Y_START], eax
			mov	dword [Y_END], ebx
			mov	dword [X_START], ecx
			mov	dword [X_END], edx
			ret



;****************************************************************************
;* vic6569_exit *************************************************************
;****************************************************************************
section .text
vic6569_exit:
;			call	fb_close
			ret



;****************************************************************************
;* vic6569_reset ************************************************************
;****************************************************************************
section .text
vic6569_reset:
			call	fb_clear
			call	vic6569_center_screen
			mov	dword eax, [SCREEN_START]
			lea	dword ebx, [FRAME_DATA]
			mov	dword [SCREEN], eax
			mov	dword [FRAME], ebx
			xor	dword eax, eax
			mov	dword [LINE_CYCLE], eax
			mov	dword [LINE_COUNTER], eax
			mov	dword [HIGH_BITS], eax
			mov	dword [VIDEO_COUNTER], eax
			mov	dword [VIDEO_COUNTER_BASE], eax
			mov	dword [ROW_COUNTER], eax
			mov	dword [STATE], eax
			mov	byte  [BUS_CYCLE_COUNTER], al
			mov	byte  [SEQUENCER_CYCLE_COUNTER], al
			push	dword edi
			xor	dword ecx, ecx
			lea	dword edi, [REGISTERS]
			mov	byte  cl, 47
			cld
			rep stosb
			mov	dword ecx, (VIC6569_LINES * VIC6569_LINE_LENGTH) / 4
			lea	dword edi, [FRAME_DATA]
			rep stosd
			pop	dword edi
			mov	dword eax, 31
			mov	dword ebx, 334
			mov	dword [BORDER_LEFT], eax
			mov	dword [BORDER_RIGHT], ebx
			mov	dword eax, 55
			mov	dword ebx, 246
			mov	dword [BORDER_TOP], eax
			mov	dword [BORDER_BUTTOM], ebx
			mov	dword eax, sequencer_cycle.cycle0
			mov	dword ebx, bus_cycle.cycle0
			mov	dword [EXECUTE_SEQUENCER_CYCLE], eax
			mov	dword [EXECUTE_BUS_CYCLE], ebx
			ret



;****************************************************************************
;* vic6569_cycle ************************************************************
;****************************************************************************
section .text
vic6569_cycle:
			VIC6569_CALL_CYCLE
			ret



%ifdef VIC6569_DEBUG
;****************************************************************************
;* vic6569_debug ************************************************************
;****************************************************************************
section .text
vic6569_debug:
			push	dword eax
			push	dword ebx
			push	dword ecx
			push	dword edx
			SYS_WRITE  2, .d__counter, 4
			pop	dword edx
			pop	dword ecx
			pop	dword ebx
			pop	dword eax
			ret

section .data
			align 4
.d__counter:		dd	0
%endif



;****************************************************************************
;* vic6569_dump *************************************************************
;****************************************************************************
section .text
vic6569_dump:
			ret



;****************************************************************************
;* vic6569_read *************************************************************
;****************************************************************************
section .text
vic6569_read:
			mov	dword ebp, ebx
			and	dword ebp, 03fh
			cmp	dword ebp, 02eh
			ja	.read_rubbish
			xor	byte  ah, ah
			mov	byte  al, [REGISTERS + ebp]
			cmp	dword ebp, 01eh
			jne	.read1
			mov	byte  [REGISTERS + ebp], ah
			ret
.read1:			cmp	dword ebp, 01fh
			jne	.read2
			mov	byte  [REGISTERS + ebp], ah
.read2:			ret
.read_rubbish:		mov	byte  al, 0ffh
			ret

 
;****************************************************************************
;* vic6569_write ************************************************************
;****************************************************************************
section .text
vic6569_write:
			mov	dword ebp, ebx
			and	dword ebp, 03fh
			jmp	dword [.ap__registers + 4 * ebp]
.ap__registers:		dd	.write
			dd	.write
			dd	.write
			dd	.write
			dd	.write
			dd	.write
			dd	.write
			dd	.write
			dd	.write
			dd	.write
			dd	.write
			dd	.write
			dd	.write
			dd	.write
			dd	.write
			dd	.write

			dd	.write
			dd	.write_reg11
			dd	.write_reg12
			dd	.write_ignore
			dd	.write_ignore
			dd	.write
			dd	.write_reg16
			dd	.write
			dd	.write_reg18
			dd	.write_reg19
			dd	.write_reg1a
			dd	.write
			dd	.write
			dd	.write
			dd	.write_ignore
			dd	.write_ignore

			dd	.write_color
			dd	.write_color
			dd	.write_color
			dd	.write_color
			dd	.write_color
			dd	.write_color
			dd	.write_color
			dd	.write_color
			dd	.write_color
			dd	.write_color
			dd	.write_color
			dd	.write_color
			dd	.write_color
			dd	.write_color
			dd	.write_color
			dd	.write_ignore

			dd	.write_ignore
			dd	.write_ignore
			dd	.write_ignore
			dd	.write_ignore
			dd	.write_ignore
			dd	.write_ignore
			dd	.write_ignore
			dd	.write_ignore
			dd	.write_ignore
			dd	.write_ignore
			dd	.write_ignore
			dd	.write_ignore
			dd	.write_ignore
			dd	.write_ignore
			dd	.write_ignore
			dd	.write_ignore

.write:			mov	byte  [REGISTERS + ebp], al
.write_ignore:		ret

.write_reg11:		mov	byte  ah, al
			push	dword ebx
			and	byte  al, 07fh
			mov	byte  bl, [REGISTERS + 011h]
			rol	byte  ah, 1
			and	byte  bl, 080h
			and	byte  ah, 001h
			or	byte  al, bl
			mov	byte  [REGISTERS + 011h], al
			mov	byte  [LINE_LATCH + 1], ah
			mov	dword ebx, 55
			mov	dword ebp, 246
			test	byte  al, 008h
			jz	.write_reg11_1
			mov	dword ebx, 51
			mov	dword ebp, 250
.write_reg11_1:		mov	dword [BORDER_TOP], ebx
			mov	dword [BORDER_BUTTOM], ebp
			mov	dword ebx, [LINE_COUNTER]
			mov	dword ebp, [STATE]
			cmp	dword ebx, 030h
			ja	.write_reg11_3
			and	dword ebp, (NOT32 - VIC6569_DISPLAY_ENABLE)
			test	byte  al, 010h
			jz	.write_reg11_2
			or	dword ebp, VIC6569_DISPLAY_ENABLE
.write_reg11_2:		mov	dword [STATE], ebp
.write_reg11_3:		pop	dword ebx
			ret

.write_reg12:		mov	byte  [LINE_LATCH], al
			ret

			;!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
			;TODO: state of the reset bit ?
			;!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

.write_reg16:		or	byte  al, 0c0h
			push	dword ebx
			mov	byte  [REGISTERS + 016h], al
			mov	dword ebx, 31
			mov	dword ebp, 334
			test	byte  al, 008h
			jz	.write_reg16_1
			mov	dword ebx, 24
			mov	dword ebp, 343
.write_reg16_1:		mov	dword [BORDER_LEFT], ebx
			mov	dword [BORDER_RIGHT], ebp
			pop	dword ebx
			ret

.write_reg18:		or	byte  al, 001h
			mov	byte  [REGISTERS + 018h], al
			ret

			;!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
			;TODO: correct behavior of this register
			;!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

.write_reg19:		;and	byte  al, 00fh
;			or	byte  al, 070h
;			mov	byte  [REGISTERS + 019h], al
			xor	byte  al, al
			mov	byte  [REGISTERS + 019h], al
			call	link_vic_irq
			ret

.write_reg1a:		or	byte  al, 0f0h
			mov	byte  [REGISTERS + 01ah], al
			ret

.write_color:		or	byte  al, 0f0h
			mov	byte  [REGISTERS + ebp], al
			ret



;****************************************************************************
;* vic6569_signal_high_bits *************************************************
;****************************************************************************
section .text
vic6569_signal_high_bits:
			push	dword eax
			shl	dword eax, 14
			and	dword eax, 0c000h
			mov	dword [HIGH_BITS], eax
			pop	dword eax
			ret



;****************************************************************************
;* vic6569_center_screen ****************************************************
;****************************************************************************
section .text
vic6569_center_screen:
			mov	dword ebx, [Y_END]
			sub	dword ebx, [Y_START]
			mov	dword edx, [X_END]
			sub	dword edx, [X_START]
			mov	dword eax, [Y_RESOLUTION]
			mov	dword ecx, [X_RESOLUTION]
			sub	dword eax, ebx
			sub	dword ecx, edx
			shr	dword eax, 1
			shr	dword ecx, 1
			mov	dword ebx, [X_RESOLUTION]
			add	dword ebx, [X_PADDING]
			mul	dword ebx
			add	dword eax, ecx
			add	dword eax, [FB_ADDRESS]
			mov	dword [SCREEN_START], eax
			ret



;****************************************************************************
;* vic6569_led1 *************************************************************
;****************************************************************************
section .text
vic6569_led1:
			push	dword ebx
			push	dword ecx
			mov	dword ebx, [FB_ADDRESS]
			xor	dword ecx, ecx
			test	byte  al, al
			jz	.led_off
			mov	dword ecx, 011111111h
.led_off:		mov	dword [ebx], ecx
			mov	dword [ebx + 4], ecx
			add	dword ebx, [X_RESOLUTION]
			add	dword ebx, [X_PADDING]
			mov	dword [ebx], ecx
			mov	dword [ebx + 4], ecx
			pop	dword ecx
			pop	dword ebx
			ret



;****************************************************************************
;* vic6569_led2 *************************************************************
;****************************************************************************
section .text
vic6569_led2:
			push	dword ebx
			push	dword ecx
			mov	dword ebx, [FB_ADDRESS]
			xor	dword ecx, ecx
			test	byte  al, al
			jz	.led_off
			mov	dword ecx, 011111111h
.led_off:		mov	dword [ebx + 12], ecx
			mov	dword [ebx + 16], ecx
			add	dword ebx, [X_RESOLUTION]
			add	dword ebx, [X_PADDING]
			mov	dword [ebx + 12], ecx
			mov	dword [ebx + 16], ecx
			pop	dword ecx
			pop	dword ebx
			ret



;****************************************************************************
;* vic6569_number1 **********************************************************
;****************************************************************************
section .text
vic6569_number1:
			push	dword eax
			push	dword ebx
			push	dword ecx
			push	dword edx
			push	dword ebp
			mov	byte  ah, al

			;-------
			;digit 1
			;-------

			mov	dword ebp, [FB_ADDRESS]
			shr	byte  al, 4
			xor	dword ebx, ebx
			add	byte  al, 030h
			cmp	byte  al, 03ah
			jb	.skip_1
			sub	byte  al, 039h
.skip_1:		mov	byte  bl, al
			shl	dword ebx, 3
			mov	byte  cl, 8
.loop_1:		mov	byte  al, [ab__c64_character_rom + ebx]
			mov	byte  ch, 8
.loop_2:		mov	byte  dl, 011h
			shl	byte  al, 1
			jc	.draw_1
			mov	byte  dl, 010h
.draw_1:		mov	byte  [ebp + 24], dl
			inc	dword ebp
			dec	byte  ch
			jnz	.loop_2
			sub	dword ebp, 8
			add	dword ebp, [X_RESOLUTION]
			add	dword ebp, [X_PADDING]
			inc	dword ebx
			dec	byte  cl
			jnz	.loop_1

			;-------
			;digit 2
			;-------

			mov	dword ebp, [FB_ADDRESS]
			and	byte  ah, 00fh
			xor	dword ebx, ebx
			add	byte  ah, 030h
			cmp	byte  ah, 03ah
			jb	.skip_2
			sub	byte  ah, 039h
.skip_2:		mov	byte  bl, ah
			shl	dword ebx, 3
			mov	byte  cl, 8
.loop_3:		mov	byte  al, [ab__c64_character_rom + ebx]
			mov	byte  ch, 8
.loop_4:		mov	byte  dl, 011h
			shl	byte  al, 1
			jc	.draw_2
			mov	byte  dl, 010h
.draw_2:		mov	byte  [ebp + 32], dl
			inc	dword ebp
			dec	byte  ch
			jnz	.loop_4
			sub	dword ebp, 8
			add	dword ebp, [X_RESOLUTION]
			add	dword ebp, [X_PADDING]
			inc	dword ebx
			dec	byte  cl
			jnz	.loop_3

			pop	dword ebp
			pop	dword edx
			pop	dword ecx
			pop	dword ebx
			pop	dword eax
			ret



;****************************************************************************
;* read_ram *****************************************************************
;****************************************************************************
section .text
read_ram:
			cmp	dword ebx, 01000h
			jb	.no_character_rom
			cmp	dword ebx, 02000h
			jb	.character_rom_1
			cmp	dword ebx, 09000h
			jb	.no_character_rom
			cmp	dword ebx, 0a000h
			jb	.character_rom_2
.no_character_rom:	mov	byte  al, [ab__c64_ram + ebx]
			ret
.character_rom_1:	mov	byte  al, [ab__c64_character_rom - 01000h + ebx]
			ret
.character_rom_2:	mov	byte  al, [ab__c64_character_rom - 09000h + ebx]
			ret

;			test	dword ebx, 04000h
;			jnz	.no_character_rom
;			test	dword ebx, 02000h
;			jnz	.no_character_rom
;			test	dword ebx, 01000h
;			jz	.no_character_rom
;			and	dword ebx, 00fffh
;			mov	byte  al, [ab__c64_character_rom + ebx]
;			ret
;.no_character_rom:	mov	byte  al, [ab__c64_ram + ebx]
;			ret



;****************************************************************************
;* Macros needed by bus_cycle ***********************************************
;****************************************************************************
%macro CHECK_BAD_LINE 0
			test	dword [STATE], VIC6569_DISPLAY_ENABLE
			jz	%%no_dma
			test	dword [STATE], VIC6569_BAD_LINE
			jnz	%%no_dma
			mov	dword eax, [LINE_COUNTER]
			cmp	dword eax, 030h
			jb	%%no_dma
			cmp	dword eax, 0f7h
			ja	%%no_dma
			mov	byte  bl, [REGISTERS + 011h]
			and	byte  al, 007h
			and	byte  bl, 007h
			cmp	byte  al, bl
			jne	%%no_dma
			mov	byte  [DMA_DELAY_COUNTER], 4
			or	dword [STATE], (VIC6569_BAD_LINE | VIC6569_DISPLAY_MODE)
			call	cpu65xx_signal_set_dma
%ifdef VIC6569_DEBUG
			inc	dword [vic6569_debug.d__counter]
%endif
%%no_dma:
%endmacro



;****************************************************************************
;* bus_cycle ****************************************************************
;****************************************************************************
section .text
bus_cycle:
.cycle0:		mov	dword eax, [LINE_COUNTER]
			mov	dword ebx, [SCREEN]
			inc	dword eax
			mov	dword ecx, [FRAME]
			cmp	dword eax, VIC6569_LINES
			jb	.cycle0_next_line

			;new frame

.cycle0_new_frame:	call	sync_new_frame
%ifdef VIC6569_DEBUG
			call	vic6569_debug
			mov	dword [vic6569_debug.d__counter], 0
%endif
			mov	dword eax, [STATE]
			xor	dword ebx, ebx
			and	dword eax, (NOT32 - VIC6569_DISPLAY_ENABLE)
			test	byte  [REGISTERS + 011h], 010h
			jz	.cycle0_new_frame_1
			or	dword eax, VIC6569_DISPLAY_ENABLE
.cycle0_new_frame_1:	mov	dword [VIDEO_COUNTER_BASE], ebx
			mov	dword [STATE], eax
			mov	dword ebx, [SCREEN_START]
			lea	dword ecx, [FRAME_DATA]
			xor	dword eax, eax
			mov	dword ebp, .cycle1
			jmp	.cycle0_end

			;next line

.cycle0_next_line:	add	dword ebx, [X_RESOLUTION]
			add	dword ecx, VIC6569_LINE_LENGTH
			add	dword ebx, [X_PADDING]
			mov	dword ebp, .cycle2
			cmp	dword eax, [LINE_LATCH]
			jne	.cycle0_end
			mov	byte  dl, [REGISTERS + 019h]
			or	byte  dl, 1
			test	byte  [REGISTERS + 01ah], 1
			jz	.cycle0_no_irq
			or	byte  dl, 080h
			push	dword eax
			mov	byte  al, 1
			call	link_vic_irq
			pop	dword eax
.cycle0_no_irq:		mov	byte  [REGISTERS + 019h], dl

.cycle0_end:		mov	dword [LINE_COUNTER], eax
			mov	dword [SCREEN], ebx
			mov	byte  [REGISTERS + 012h], al
			mov	byte  bl, [REGISTERS + 011h]
			mov	dword [FRAME], ecx
			ror	byte  ah, 1
			and	byte  bl, 07fh
			xor	dword ecx, ecx
			or	byte  bl, ah
			mov	dword [LINE_CYCLE], ecx
			mov	byte  [REGISTERS + 011h], bl
			mov	byte  [BUS_CYCLE_COUNTER], 10
			mov	dword [EXECUTE_BUS_CYCLE], ebp
;			mov	dword [EXECUTE_SEQUENCER_CYCLE], sequencer_cycle.cycle0
			ret


			;------------------------------------------------
			;for new_frame:
			;if LINE_COUNTER = LINE_LATCH and RasterIRQ is on
			;generate an IRQ
			;------------------------------------------------

.cycle1:		mov	dword eax, [LINE_COUNTER]
			cmp	dword eax, [LINE_LATCH]
			jne	.cycle1_end
			mov	byte  bl, [REGISTERS + 019h]
			or	byte  bl, 1
			test	byte  [REGISTERS + 01ah], 1
			jz	.cycle1_no_irq
			or	byte  bl, 080h
			mov	byte  al, 1
			call	link_vic_irq
.cycle1_no_irq:		mov	byte  [REGISTERS + 019h], bl
.cycle1_end:		dec	byte  [BUS_CYCLE_COUNTER]
			mov	dword [EXECUTE_BUS_CYCLE], .cycle2
			ret



			;-------------------------------
			;Cycle 1 - 10 (2 - 10 in Line 0)
			;-------------------------------

.cycle2:		dec	byte  [BUS_CYCLE_COUNTER]
			jnz	.cycle2_end
			mov	dword [EXECUTE_BUS_CYCLE], .cycle11
.cycle2_end:		ret


			;----------------------------------
			;Cycle 11, check for bad line state
			;----------------------------------

.cycle11:		CHECK_BAD_LINE
			mov	dword [EXECUTE_BUS_CYCLE], .cycle12
			ret



			;----------------------------------
			;Cycle 12, check for bad line state
			;----------------------------------

.cycle12:		CHECK_BAD_LINE
			mov	dword [EXECUTE_BUS_CYCLE], .cycle13
			ret



			;----------------------------------------
			;Cycle 13, check for bad line state
			;move VIDEO_COUNTER_BASE to VIDEO_COUNTER
			;----------------------------------------

.cycle13:		CHECK_BAD_LINE
			mov	dword eax, [VIDEO_COUNTER_BASE]
			xor	dword ebx, ebx
			mov	dword [VIDEO_COUNTER], eax
			mov	dword [CACHE_OFFSET], ebx
			test	dword [STATE], VIC6569_BAD_LINE
			jz	.cycle13_end
			mov	dword [ROW_COUNTER], ebx
.cycle13_end:		mov	dword [EXECUTE_BUS_CYCLE], .cycle14
			ret



			;----------------------------------
			;Cycle 14, check for bad line state
			;----------------------------------

.cycle14:		CHECK_BAD_LINE
			mov	byte  [BUS_CYCLE_COUNTER], 40
			mov	dword [EXECUTE_BUS_CYCLE], .cycle15
			ret




			;------------------------------------------------
			;Cycle 15 - 54, if this line is a bad line
			;fetch ColorRAM and Videomatrix
			;fetch Charset/Bitmap and store it into DATA_BYTE
			;load VIDEO_BYTE, COLOR_RAM_BYTE from the Caches
			;------------------------------------------------

.cycle15:		mov	dword ebp, [CACHE_OFFSET]
			mov	dword ebx, [VIDEO_COUNTER]
			test	dword [STATE], VIC6569_DISPLAY_MODE
			jz	near .cycle15_idle

			;---------------------------------
			;get VIDEO_BYTE and COLOR_RAM_BYTE
			;---------------------------------

			test	dword [STATE], VIC6569_BAD_LINE
			jz	.cycle15_no_badline
			test	dword [STATE], VIC6569_ACCESS_GRANTED
			jz	.cycle15_read_rubbish
			mov	byte  cl, [REGISTERS + 018h]
			shl	dword ecx, 6
			and	dword ecx, 03c00h
			mov	byte  al, [ab__c64_color_ram + ebx]
			or	dword ecx, [HIGH_BITS]
			mov	byte  [COLOR_RAM_CACHE + ebp], al
			mov	byte  [COLOR_RAM_BYTE], al
			or	dword ebx, ecx
			call	read_ram
			mov	byte  [VIDEO_CACHE + ebp], al
			mov	byte  [VIDEO_BYTE], al
			jmp	.cycle15_data
.cycle15_read_rubbish:	mov	byte  al, 0ffh
			mov	byte  [COLOR_RAM_CACHE + ebp], al
			mov	byte  [COLOR_RAM_BYTE], al
			mov	byte  [VIDEO_CACHE + ebp], al
			mov	byte  [VIDEO_BYTE], al
			jmp	.cycle15_data
.cycle15_no_badline:	mov	byte  al, [COLOR_RAM_CACHE + ebp]
			mov	byte  bl, [VIDEO_CACHE + ebp]
			mov	byte  [COLOR_RAM_BYTE], al
			mov	byte  [VIDEO_BYTE], bl

			;-------------
			;get DATA_BYTE
			;-------------

.cycle15_data:		CHECK_BAD_LINE
			mov	dword ebx, [VIDEO_COUNTER]
			mov	byte  cl, [REGISTERS + 018h]
			shl	dword ecx, 10
			test	byte  [REGISTERS + 011h], 020h
			jz	.cycle15_data_charset
			shl	dword ebx, 3
			and	dword ecx, 02000h
			jmp	.cycle15_data_common
.cycle15_data_charset:	xor	dword ebx, ebx
			and	dword ecx, 03800h
			mov	byte  bl, [VIDEO_BYTE]
			shl	dword ebx, 3
.cycle15_data_common:	or	dword ebx, ecx
			or	dword ebx, [ROW_COUNTER]
			or	dword ebx, [HIGH_BITS]
			test	byte  [REGISTERS + 011h], 040h
			jz	.cycle15_data_no_ecm
			and	dword ebx, 0f9ffh
.cycle15_data_no_ecm:	call	read_ram
			mov	dword ebx, [VIDEO_COUNTER]
			mov	byte  [DATA_BYTE], al
			inc	dword ebp
			inc	dword ebx
			mov	dword [CACHE_OFFSET], ebp
			mov	dword [VIDEO_COUNTER], ebx
			jmp	.cycle15_exit

			;-------------------
			;make an idle access
			;-------------------

.cycle15_idle:		mov	dword ebx, [HIGH_BITS]
			test	byte  [REGISTERS + 011h], 040h
			jz	.cycle15_idle_no_ecm
			or	dword ebx, 039ffh
			jmp	.cycle15_idle_common
.cycle15_idle_no_ecm:	or	dword ebx, 03fffh
.cycle15_idle_common:	xor	byte  cl, cl
			mov	byte  al, [ab__c64_ram + ebx]
			mov	byte  [VIDEO_BYTE], cl
			mov	byte  [COLOR_RAM_BYTE], cl
			mov	byte  [DATA_BYTE], al

.cycle15_exit:		dec	byte  [BUS_CYCLE_COUNTER]
			jnz	.cycle15_end
			and	dword [STATE], (NOT32 - VIC6569_BAD_LINE - VIC6569_ACCESS_GRANTED)
			call	cpu65xx_signal_clear_dma
			mov	dword [EXECUTE_BUS_CYCLE], .cycle55
.cycle15_end:		ret



			;--------
			;Cycle 55
			;--------

.cycle55:		mov	dword [EXECUTE_BUS_CYCLE], .cycle56
			ret



			;--------
			;Cycle 56
			;--------

.cycle56:		mov	dword [EXECUTE_BUS_CYCLE], .cycle57
			ret



			;-------------------------------------------------
			;Cycle 57, check if RC=7 if yes mode VIDEO_COUNTER
			;to VIDEO_COUNTER_BASE switch to idle mode
			;-------------------------------------------------

.cycle57:		mov	dword eax, [ROW_COUNTER]
			cmp	byte  al, 7
			jb	.cycle57_exit
			xor	dword eax, eax
			mov	dword ebx, [VIDEO_COUNTER]
			and	dword [STATE], (NOT32 - VIC6569_DISPLAY_MODE)
			mov	dword [VIDEO_COUNTER_BASE], ebx
.cycle57_exit:		test	dword [STATE], VIC6569_DISPLAY_MODE
			jz	.cycle57_no_increment
			inc	dword eax
.cycle57_no_increment:	mov	dword [ROW_COUNTER], eax
			mov	byte  [BUS_CYCLE_COUNTER], 5
			mov	dword [EXECUTE_BUS_CYCLE], .cycle58
			ret



			;-------------
			;Cycle 58 - 62
			;-------------

.cycle58:		dec	byte  [BUS_CYCLE_COUNTER]
			jnz	.cycle58_end
			mov	dword [EXECUTE_BUS_CYCLE], .cycle0
.cycle58_end:		ret



;****************************************************************************
;* sequencer_cycle **********************************************************
;****************************************************************************
section .text
sequencer_cycle:
.cycle0:		call	display_border
			mov	byte  [SEQUENCER_CYCLE_COUNTER], 15
			mov	dword [EXECUTE_SEQUENCER_CYCLE], .cycle1
			ret

.cycle1:		call	display_border
			dec	byte  [SEQUENCER_CYCLE_COUNTER]
			jnz	.cycle1_end
			mov	byte  [SEQUENCER_CYCLE_COUNTER], 40
			mov	dword [EXECUTE_SEQUENCER_CYCLE], .cycle16
.cycle1_end:		ret

.cycle16:		test	dword [STATE], VIC6569_DISPLAY_ENABLE
			jz	.cycle16_border
			mov	dword eax, [LINE_COUNTER]
			cmp	dword eax, [BORDER_TOP]
			jb	.cycle16_border
			cmp	dword eax, [BORDER_BUTTOM]
			jbe	.cycle16_screen
.cycle16_border:	call	display_border
			jmp	.cycle16_exit
.cycle16_screen:	call	display
.cycle16_exit:		dec	byte  [SEQUENCER_CYCLE_COUNTER]
			jnz	.cycle16_end
			mov	byte  [SEQUENCER_CYCLE_COUNTER], 7
			mov	dword [EXECUTE_SEQUENCER_CYCLE], .cycle56
.cycle16_end:		ret

.cycle56:		call	display_border
			dec	byte  [SEQUENCER_CYCLE_COUNTER]
			jnz	.cycle56_end
			mov	dword [EXECUTE_SEQUENCER_CYCLE], .cycle0
.cycle56_end:		ret



;****************************************************************************
;* display_border ***********************************************************
;****************************************************************************
section .text
display_border:
			mov	byte  al, [REGISTERS + 020h]
			mov	dword ebx, [SCREEN]
			and	byte  al, 00fh
			mov	dword edx, [FRAME]
			or	byte  al, 010h
			mov	dword ebp, [LINE_CYCLE]
			mov	byte  ah, al
			mov	dword ecx, eax
			and	dword eax, NOT16
			shl	dword ecx, 16
			or	dword eax, ecx
			cmp	dword [edx + 8 * ebp], eax
			je	.skip1
			mov	dword [edx + 8 * ebp], eax
			mov	dword [ebx + 8 * ebp], eax
.skip1:			cmp	dword [edx + 8 * ebp + 4], eax
			je	.skip2
			mov	dword [edx + 8 * ebp + 4], eax
			mov	dword [ebx + 8 * ebp + 4], eax
.skip2:			inc	dword ebp
			mov	dword [LINE_CYCLE], ebp
			ret



;****************************************************************************
;* display ******************************************************************
;****************************************************************************
section .text
display:
			xor	dword ebp, ebp
			test	byte  [REGISTERS + 011h], 020h
			jz	.no_bitmap
			or	dword ebp, 1
.no_bitmap:		test	byte  [REGISTERS + 011h], 040h
			jz	.no_ecm
			or	dword ebp, 2
.no_ecm:		test	byte  [REGISTERS + 016h], 010h
			jz	.no_mcm
			or	dword ebp, 4
.no_mcm:		jmp	dword [.ap__functions + 4 * ebp]

			align 4
.ap__functions:		dd	display_text
			dd	display_bitmap
			dd	display_text_ecm
			dd	display_bitmap_ecm
			dd	display_text_mcm
			dd	display_bitmap_mcm
			dd	display_text_mcm_ecm
			dd	display_bitmap_mcm_ecm



;****************************************************************************
;* display_text *************************************************************
;****************************************************************************
section .text
display_text:
			xor	dword ecx, ecx
			mov	dword ebx, [SCREEN]
			mov	byte  cl, [REGISTERS + 021h]
			mov	dword ebp, [LINE_CYCLE]
			and	byte  cl, 00fh
			mov	byte  al, [DATA_BYTE]
			or	byte  cl, 010h
			mov	byte  ah, [COLOR_RAM_BYTE]
			mov	byte  ch, cl
			and	byte  ah, 00fh
			mov	dword edx, ecx
			or	byte  ah, 010h
			jmp	display_bitmap.common



;****************************************************************************
;* display_text_ecm *********************************************************
;****************************************************************************
section .text
display_text_ecm:
			xor	dword ecx, ecx
			mov	dword ebx, [SCREEN]
			mov	byte  cl, [VIDEO_BYTE]
			shr	byte  cl, 6
			mov	byte  cl, [REGISTERS + 021h + ecx]
			mov	dword ebp, [LINE_CYCLE]
			and	byte  cl, 00fh
			mov	byte  al, [DATA_BYTE]
			or	byte  cl, 010h
			mov	byte  ah, [COLOR_RAM_BYTE]
			mov	byte  ch, cl
			and	byte  ah, 00fh
			mov	dword edx, ecx
			or	byte  ah, 010h
			jmp	display_bitmap.common



;****************************************************************************
;* display_text_mcm *********************************************************
;****************************************************************************
section .text
display_text_mcm:
			test	byte  [COLOR_RAM_BYTE], 008h
			jz	display_text
			mov	byte  dh, [COLOR_RAM_BYTE]
			mov	byte  cl, [REGISTERS + 021h]
			mov	byte  dl, [REGISTERS + 022h]
			mov	byte  ch, [REGISTERS + 023h]
			and	byte  dh, 007h
			jmp	display_bitmap_mcm.common



;****************************************************************************
;* display_text_mcm_ecm *****************************************************
;****************************************************************************
section .text
display_text_mcm_ecm:
			test	byte  [COLOR_RAM_BYTE], 008h
			jz	.no_mcm
			xor	dword ecx, ecx
			xor	dword edx, edx
			jmp	display_bitmap_mcm.common
.no_mcm:		mov	byte  al, [DATA_BYTE]
			mov	dword ebx, [SCREEN]
			mov	dword ecx, 010101010h
			xor	byte  ah, ah
			mov	dword edx, ecx
			mov	dword ebp, [LINE_CYCLE]
			jmp	display_bitmap.common



;****************************************************************************
;* display_bitmap ***********************************************************
;****************************************************************************
section .text
display_bitmap:
			xor	dword ecx, ecx
			mov	dword ebx, [SCREEN]
			mov	byte  cl, [VIDEO_BYTE]
			mov	dword ebp, [LINE_CYCLE]
			and	byte  cl, 00fh
			mov	byte  al, [DATA_BYTE]
			or	byte  cl, 010h
			mov	byte  ah, [VIDEO_BYTE]
			mov	byte  ch, cl
			shr	byte  ah, 4
			mov	dword edx, ecx
			or	byte  ah, 010h
.common:		shl	dword edx, 16
			push	dword esi
			or	dword ecx, edx
			mov	dword esi, [FRAME]
			mov	dword edx, ecx
			shr	byte  al, 1
			jnc	.bit0
			mov	byte  dh, ah
.bit0:			shr	byte  al, 1
			jnc	.bit1
			mov	byte  dl, ah
.bit1:			rol	dword edx, 16
			shr	byte  al, 1
			jnc	.bit2
			mov	byte  dh, ah
.bit2:			shr	byte  al, 1
			jnc	.bit3
			mov	byte  dl, ah
.bit3:			cmp	dword [esi + 8 * ebp + 4], edx
			je	.skip1
			mov	dword [esi + 8 * ebp + 4], edx
			mov	dword [ebx + 8 * ebp + 4], edx
.skip1:			shr	byte  al, 1
			jnc	.bit4
			mov	byte  ch, ah
.bit4:			shr	byte  al, 1
			jnc	.bit5
			mov	byte  cl, ah
.bit5:			rol	dword ecx, 16
			shr	byte  al, 1
			jnc	.bit6
			mov	byte  ch, ah
.bit6:			shr	byte  al, 1
			jnc	.bit7
			mov	byte  cl, ah
.bit7:			cmp	dword [esi + 8 * ebp], ecx
			je	.skip2
			mov	dword [esi + 8 * ebp], ecx
			mov	dword [ebx + 8 * ebp], ecx
.skip2:			pop	dword esi
			inc	dword ebp
			mov	dword [LINE_CYCLE], ebp
			ret



;****************************************************************************
;* display_bitmap_ecm *******************************************************
;****************************************************************************
section .text
display_bitmap_ecm:
			mov	byte  al, [DATA_BYTE]
			mov	dword ebx, [SCREEN]
			mov	dword ecx, 010101010h
			xor	byte  ah, ah
			mov	dword edx, ecx
			mov	dword ebp, [LINE_CYCLE]
			jmp	display_bitmap.common



;****************************************************************************
;* display_bitmap_mcm *******************************************************
;****************************************************************************
section .text
display_bitmap_mcm:
			mov	byte  dl, [VIDEO_BYTE]
			mov	byte  cl, [REGISTERS + 021h]
			shr	byte  dl, 4
			mov	byte  ch, [VIDEO_BYTE]
			mov	byte  dh, [COLOR_RAM_BYTE]
.common:		mov	dword ebx, [SCREEN]
			and	dword ecx, 00f0fh
			and	dword edx, 00f0fh
			mov	dword ebp, [LINE_CYCLE]
			or	dword ecx, 01010h
			or	dword edx, 01010h
			mov	byte  [COLOR_LOOKUP_TABLE], cl
			mov	byte  [COLOR_LOOKUP_TABLE + 1], dl
			mov	byte  [COLOR_LOOKUP_TABLE + 2], ch
			mov	byte  al, [DATA_BYTE]
			xor	dword ecx, ecx
			mov	byte  [COLOR_LOOKUP_TABLE + 3], dh
			mov	byte  cl, al
			shr	byte  al, 2
			and	byte  cl, 3
			mov	byte  dl, [COLOR_LOOKUP_TABLE + ecx]
			mov	byte  cl, al
			mov	byte  dh, dl
			shr	byte  al, 2
			and	byte  cl, 3
			shl	dword edx, 16
			push	dword esi
			mov	byte  dl, [COLOR_LOOKUP_TABLE + ecx]
			mov	dword esi, [FRAME]
			mov	byte  cl, al
			mov	byte  dh, dl
			shr	byte  al, 2
			and	byte  cl, 3
			cmp	dword [esi + 8 * ebp + 4], edx
			je	.skip1
			mov	dword [esi + 8 * ebp + 4], edx
			mov	dword [ebx + 8 * ebp + 4], edx
.skip1:			mov	byte  dl, [COLOR_LOOKUP_TABLE + ecx]
			mov	byte  cl, al
			mov	byte  dh, dl
			and	byte  cl, 3
			shl	dword edx, 16
			mov	byte  dl, [COLOR_LOOKUP_TABLE + ecx]
			mov	byte  dh, dl
			cmp	dword [esi + 8 * ebp], edx
			je	.skip2
			mov	dword [esi + 8 * ebp], edx
			mov	dword [ebx + 8 * ebp], edx
.skip2:			pop	dword esi
			inc	dword ebp
			mov	dword [LINE_CYCLE], ebp
			ret



;****************************************************************************
;* display_bitmap_mcm_ecm ***************************************************
;****************************************************************************
section .text
display_bitmap_mcm_ecm:
			xor	dword ecx, ecx
			xor	dword edx, edx
			jmp	display_bitmap_mcm.common
;********************************************* karsten.scheibler@bigfoot.de *

