;****************************************************************************
;****************************************************************************
;*
;* fb.nasm
;*
;****************************************************************************
;****************************************************************************


;TODO:
;
;* get Screen Geometry
;* check if 8 bpp




%include "sys_calls.inc"
%include "write.inc"

%define FB_NASM
%include "fb.inc"



;****************************************************************************
;* Data *********************************************************************
;****************************************************************************
section .data
d__x_resolution:	dd	0
d__x_padding:		dd	0
d__y_resolution:	dd	0
p__fb_address		dd	0

d__fb_handle:		dd	0



;****************************************************************************
;* fb_open ******************************************************************
;****************************************************************************
section .text
fb_open:
			;----
			;open
			;----

			SYS_OPEN  .cb__fb_device, O_RDWR
			test	dword eax, eax
			js	near .open_error
			mov	dword [d__fb_handle], eax
			mov	dword [.d__fb_handle], eax

			;--------------------
			;ioctl get screeninfo
			;--------------------

			SYS_IOCTL  eax, FBIOGET_VSCREENINFO, .ab__screeninfo
			test	dword eax, eax
			js	near .ioctl_error
			mov	dword eax, [.ab__screeninfo]
			mov	dword ebx, [.ab__screeninfo + 4]
			mov	dword ecx, [.ab__screeninfo + 8]
			mov	dword [d__x_resolution], eax
			sub	dword ecx, eax
			mov	dword [d__y_resolution], ebx
			mov	dword [d__x_padding], ecx

			;------------------------------
			;ioctl to change the colortable
			;------------------------------

			mov	dword eax, [d__fb_handle]
			SYS_IOCTL  eax, FBIOPUTCMAP, .ad__cmap_color_table
			test	dword eax, eax
			js	near .ioctl_error

			;----
			;mmap
			;----

			mov	dword eax, [d__x_resolution]
			add	dword eax, [d__x_padding]
			mul	dword [d__y_resolution]

			;!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
			;TODO: if carry set, then edx != 0 ==> error
			;!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

			mov	dword [.d__length], eax
			SYS_MMAP  .ad__mmap_data
			test	dword eax, eax
			js	near .mmap_error
			mov	dword [p__fb_address], eax
			ret

.cb__open_error:	db	"Error while opening fb device", 00ah
.open_error_length	equ	$ - .cb__open_error
.open_error:		push	dword eax
			WRITE_TEXT  .cb__open_error, .open_error_length
			jmp	.exit
.cb__ioctl_error:	db	"Error while ioctl on fb device", 00ah
.ioctl_error_length	equ	$ - .cb__ioctl_error
.ioctl_error:		push	dword eax
			WRITE_TEXT  .cb__ioctl_error, .ioctl_error_length
			jmp	.exit
.cb__mmap_error:	db	"Error while mmap on fb device", 00ah
.mmap_error_length	equ	$ - .cb__mmap_error
.mmap_error:		push	dword eax
			WRITE_TEXT  .cb__mmap_error, .mmap_error_length
.exit:			pop	dword eax
			WRITE_HEX  8
			WRITE_RETURN
			SYS_EXIT
			ret

section .bss
			align	4
.ab__screeninfo:	resd	0200h

section .data
			align	4
.ad__mmap_data:
.d__start:		dd	0
.d__length:		dd	0
.d__protection:		dd	(PROT_READ | PROT_WRITE)
.d__flags:		dd	MAP_SHARED
.d__fb_handle:		dd	0
.d__offset:		dd	0

.aw__red:		dw	00000h, 0e000h, 06000h, 00000h
			dw	06000h, 00000h, 00000h, 0e000h
			dw	06000h, 04000h, 0e000h, 02000h
			dw	04000h, 06000h, 06000h, 06000h
.aw__green:		dw	00000h, 0e000h, 02000h, 0e000h
			dw	02000h, 06000h, 00000h, 0e000h
			dw	04000h, 02000h, 06000h, 02000h
			dw	04000h, 0e000h, 06000h, 06000h
.aw__blue:		dw	00000h, 0e000h, 02000h, 0e000h
			dw	06000h, 04000h, 06000h, 06000h
			dw	02000h, 00000h, 06000h, 02000h
			dw	04000h, 06000h, 0e000h, 06000h
.aw__transparent:	times 16 dw 0000h

.ad__cmap_color_table:
			dd	16
			dd	16
			dd	.aw__red
			dd	.aw__green
			dd	.aw__blue
			dd	.aw__transparent

.cb__fb_device:		db	"/dev/fb0", 0




;****************************************************************************
;* fb_close *****************************************************************
;****************************************************************************
section .text
fb_close:
			SYS_CLOSE  [d__fb_handle]
			ret



;****************************************************************************
;* fb_clear *****************************************************************
;****************************************************************************
section .text
fb_clear:
			push	dword edi
			xor	dword eax, eax
			mov	dword edi, [p__fb_address]
			cld
			mov	dword ebx, [d__x_resolution]
			shr	dword ebx, 2
			mov	dword edx, [d__y_resolution]
.clear_line:		mov	dword ecx, ebx
			rep stosd
			add	dword edi, [d__x_padding]
			dec	dword edx
			jnz	.clear_line
			pop	dword edi
			ret

;****************************************************************************
;* fb_get_parameters ********************************************************
;****************************************************************************
section .text
fb_get_parameters:
			mov	dword eax, [d__x_resolution]
			mov	dword ebx, [d__x_padding]
			mov	dword ecx, [d__y_resolution]
			mov	dword edx, [p__fb_address]
			ret



;****************************************************************************
%ifdef FB_TEST

global _start


section .text
_start:
			call	fb_open
			call	fb_clear
			SYS_EXIT
%endif
;********************************************* karsten.scheibler@bigfoot.de *

