;****************************************************************************
;****************************************************************************
;*
;* sys_calls.inc
;*
;****************************************************************************
;****************************************************************************





%ifndef SYS_CALLS_INC
%define SYS_CALLS_INC

%include "tty_ioctl.inc"
%include "fb_ioctl.inc"
%include "audio_ioctl.inc"
%include "fcntl.inc"
%include "mmap.inc"
%include "time.inc"
%include "errors.inc"

;****************************************************************************
;* System Call Numbers ******************************************************
;****************************************************************************
%assign __NR_exit		1
%assign __NR_read		3
%assign __NR_write		4
%assign __NR_open		5
%assign __NR_close		6
%assign __NR_lseek		19
%assign __NR_ioctl		54
%assign __NR_fcntl		55
%assign __NR_gettimeofday	78
%assign __NR_mmap		90
%assign __NR_nanosleep		162

%macro SET_REGISTER 2
%ifnum %2
%if %2 = 0
			xor	dword %1, %1
%else
			mov	dword %1, %2
%endif
%else
			mov	dword %1, %2
%endif
%endmacro



;****************************************************************************
;****************************************************************************
;*
;* System Call Macros
;*
;****************************************************************************
;****************************************************************************





;****************************************************************************
;* SYS_EXIT *****************************************************************
;****************************************************************************
%macro SYS_EXIT 0-1
			xor	dword eax, eax
%if %0 > 0
			SET_REGISTER  ebx, %1
%else
			mov	dword ebx, eax
%endif
			inc	dword eax
			int	byte  080h
%endmacro



;****************************************************************************
;* SYS_READ *****************************************************************
;****************************************************************************
%macro SYS_READ 0-3
%if %0 > 0
			SET_REGISTER  ebx, %1		;file handle
%endif
%if %0 > 1
			SET_REGISTER  ecx, %2		;address of the data
%endif
%if %0 > 2
			SET_REGISTER  edx, %3		;length of the data
%endif
			mov	dword eax, __NR_read
			int	byte  080h
%endmacro



;****************************************************************************
;* SYS_WRITE ****************************************************************
;****************************************************************************
%macro SYS_WRITE 0-3
%if %0 > 0
			SET_REGISTER  ebx, %1		;file handle
%endif
%if %0 > 1
			SET_REGISTER  ecx, %2		;address of the data
%endif
%if %0 > 2
			SET_REGISTER  edx, %3		;length of the data
%endif
			mov	dword eax, __NR_write
			int	byte  080h
%endmacro



;****************************************************************************
;* SYS_OPEN *****************************************************************
;****************************************************************************
%macro SYS_OPEN 0-3
%if %0 > 0
			SET_REGISTER  ebx, %1		;address of nullterminated string
%endif
%if %0 > 1
			SET_REGISTER  ecx, %2		;acces mode
%else
			SET_REGISTER  ecx, O_RDONLY	;default: read only
%endif
%if %0 > 2
			SET_REGISTER  edx, %3		;permissions, only while creat necessary
%endif
			mov	dword eax, __NR_open
			int	byte  080h
%endmacro



;****************************************************************************
;* SYS_CLOSE ****************************************************************
;****************************************************************************
%macro SYS_CLOSE 0-1
%if %0 > 1
			SET_REGISTER  ebx, %1		;file handle
%endif
			mov	dword eax, __NR_close
			int	byte  080h
%endmacro



;****************************************************************************
;* SYS_LSEEK ****************************************************************
;****************************************************************************
%macro SYS_LSEEK 0-3
%if %0 > 0
			SET_REGISTER  ebx, %1		;file handle
%endif
%if %0 > 1
			SET_REGISTER  ecx, %2		;offset
%endif
%if %0 > 2
			SET_REGISTER  edx, %3		;origin
%endif
			mov	dword eax, __NR_lseek
			int	byte  080h
%endmacro



;****************************************************************************
;* SYS_IOCTL ****************************************************************
;****************************************************************************
%macro SYS_IOCTL 0-3
%if %0 > 0
			SET_REGISTER  ebx, %1		;file handle
%endif
%if %0 > 1
			SET_REGISTER  ecx, %2		;ioctl number
%endif
%if %0 > 2
			SET_REGISTER  edx, %3		;address of the data
%endif
			mov	dword eax, __NR_ioctl
			int	byte  080h
%endmacro



;****************************************************************************
;* SYS_FCNTL ****************************************************************
;****************************************************************************
%macro SYS_FCNTL 0-3
%if %0 > 0
			SET_REGISTER  ebx, %1		;file handle
%endif
%if %0 > 1
			SET_REGISTER  ecx, %2		;fcntl number
%endif
%if %0 > 2
			SET_REGISTER  edx, %3		;address of the data
%endif
			mov	dword eax, __NR_fcntl
			int	byte  080h
%endmacro



;****************************************************************************
;* SYS_FCNTL ****************************************************************
;****************************************************************************
%macro SYS_GETTIMEOFDAY 0-2
%if %0 > 0
			SET_REGISTER  ebx, %1		;address of struct timeval
%endif
%if %0 > 1
			SET_REGISTER  ecx, %2		;address of struct timezone
%endif
			mov	dword eax, __NR_gettimeofday
			int	byte  080h
%endmacro



;****************************************************************************
;* SYS_MMAP *****************************************************************
;****************************************************************************
%macro SYS_MMAP 0-1
%if %0 > 0
			SET_REGISTER  ebx, %1		;address of the mmap structure
%endif
			mov	dword eax, __NR_mmap
			int	byte  080h
%endmacro

;****************************************************************************
;* SYS_NANOSLEEP ************************************************************
;****************************************************************************
%macro SYS_NANOSLEEP 0-2
%if %0 > 0
			SET_REGISTER  ebx, %1		;address of timespec
%endif
%if %0 > 1
			SET_REGISTER  ecx, %2		;address of timespec
%endif
			mov	dword eax, __NR_nanosleep
			int	byte  080h
%endmacro

%endif
;********************************************* karsten.scheibler@bigfoot.de *

