;****************************************************************************
;****************************************************************************
;*
;* tty.nasm
;*
;****************************************************************************
;****************************************************************************




;%define TTY_DEBUG




%include "sys_calls.inc"
%include "defines.inc"
%include "main.inc"
%include "link.inc"

%ifdef TTY_DEBUG
%include "c64_memory.inc"
%endif

%define TTY_NASM
%include "tty.inc"



;****************************************************************************
;* tty_initialize ***********************************************************
;****************************************************************************
section .text
tty_initialize:
			SYS_WRITE  STDOUT, .ab__cursor_off, 12
			SYS_FCNTL  STDIN, F_SETFL, (O_RDONLY | O_NONBLOCK)
			SYS_IOCTL  STDIN, TC_GETS, termios__original
			SYS_IOCTL  STDIN, TC_GETS, termios__data
			and	dword [termios__data + d__termios_local_flags], (NOT32 - ICANON - ECHO)
			SYS_IOCTL  STDIN, TC_SETS, termios__data
			SYS_IOCTL  STDIN, VT_GETMODE, vtmode__original
			SYS_IOCTL  STDIN, VT_GETMODE, vtmode__data
			mov	byte  [vtmode__data + b__vtmode_mode], VT_PROCESS
			SYS_IOCTL  STDIN, VT_SETMODE, vtmode__data
			mov	byte  [flag__tty_initialized], 1
			ret

.ab__cursor_off:	db	01bh, "[?25l"
.ab__no_blanking:	db	01bh, "[9;0]"



;****************************************************************************
;* tty_initialize ***********************************************************
;****************************************************************************
section .text
tty_exit:
			test	byte  [flag__tty_initialized], NOT8
			jz	.skip
			SYS_IOCTL  STDIN, VT_SETMODE, vtmode__original
			SYS_IOCTL  STDIN, TC_SETS, termios__original
			SYS_WRITE  STDOUT, .ab__cursor_on, 12
.skip:			ret

.ab__cursor_on:		db	01bh, "[?25h"
.ab__blank_5:		db	01bh, "[9;5]"



;****************************************************************************
;* tty_clear_screen *********************************************************
;****************************************************************************
section .text
tty_clear_screen:
			SYS_WRITE  STDOUT, .ab__clear_screen, 10
			ret

.ab__clear_screen:	db	01bh, "[1;1H", 01bh, "[2J"



;****************************************************************************
;* tty_get_character ********************************************************
;****************************************************************************
section .text
tty_get_character:
			SYS_READ  0, b__character, 1
			test	dword eax, eax
			jz	.no_character
			js	.no_character
			xor	dword ebx, ebx
			mov	byte  bl, [b__character]
			and	byte  bl, 07fh
			cmp	byte  bl, 05bh
			jne	.no_escape
			SYS_READ  0, b__character, 1
			test	dword eax, eax
			jz	.no_character
			js	.no_character
			xor	dword ebx, ebx
			mov	byte  bl, [b__character]
			or	byte  bl, 080h
.no_escape:
%ifdef TTY_DEBUG
			mov	byte  [ab__c64_ram + 1024], bl
%endif
			mov	dword eax, [.ab__translate + 4 * ebx]
			call	link_keyboard
			ret
.no_character:		xor	dword eax, eax
			call	link_keyboard
			ret
.exit:			call	tty_clear_screen
			jmp	main_exit

			align 4
.ab__translate:
			;-----------
			;normal keys
			;-----------

			;-----------
			;000h - 00fh
			;-----------

			dd	0, 0, 0, 0
			dd	0, 0, 0, 0
			dd	0, 0, 025h, 0
			dd	0, 025h, 0, 0

			;-----------
			;010h - 01fh
			;-----------

			dd	0, 0, 0, 0
			dd	0, 0, 0, 0
			dd	0, 0, 0, 0
			dd	0, 0, 0, 0

			;-----------
			;020h - 02fh
			;-----------

			dd	020h, 01b31h, 01b32h, 01b33h
			dd	01b34h, 01b35h, 01b36h, 01b37h
			dd	01b38h, 01b39h, 02ah, 02bh
			dd	03ch, 02dh, 03eh, 03fh

			;-----------
			;030h - 03fh
			;-----------

			dd	030h, 031h, 032h, 033h
			dd	034h, 035h, 036h, 037h
			dd	038h, 039h, 03ah, 03bh
			dd	01b3ch, 03dh, 01b3eh, 01b3fh

			;-----------
			;040h - 04fh
			;-----------

			dd	040h, 01b01h, 01b02h, 01b03h
			dd	01b04h, 01b05h, 01b06h, 01b07h
			dd	01b08h, 01b09h, 01b0ah, 01b0bh
			dd	01b0ch, 01b0dh, 01b0eh, 01b0fh

			;-----------
			;050h - 05fh
			;-----------

			dd	01b10h, 01b11h, 01b12h, 01b13h
			dd	01b14h, 01b15h, 01b16h, 01b17h
			dd	01b18h, 01b19h, 01b1ah, 01b3ah
			dd	01ch, 01b3bh, 01eh, 01fh

			;-----------
			;060h - 06fh
			;-----------

			dd	040h, 001h, 002h, 003h
			dd	004h, 005h, 006h, 007h
			dd	008h, 009h, 00ah, 00bh
			dd	00ch, 00dh, 00eh, 00fh

			;-----------
			;070h - 07fh
			;-----------

			dd	010h, 011h, 012h, 013h
			dd	014h, 015h, 016h, 017h
			dd	018h, 019h, 01ah, 0
			dd	0, 0, 0, 022h

			;-------------------------
			;keys with escape sequence
			;-------------------------

			;-----------
			;080h - 08fh
			;-----------

			dd	0, 0, 0, 0
			dd	0, 0, 0, 0
			dd	0, 0, 0, 0
			dd	0, 0, 0, 0

			;-----------
			;090h - 09fh
			;-----------

			dd	0, 0, 0, 0
			dd	0, 0, 0, 0
			dd	0, 0, 0, 0
			dd	0, 0, 0, 0

			;-----------
			;0a0h - 0afh
			;-----------

			dd	0, 0, 0, 0
			dd	0, 0, 0, 0
			dd	0, 0, 0, 0
			dd	0, 0, 0, 0

			;-----------
			;0b0h - 0bfh
			;-----------

			dd	0, 021h, 0, 024h
			dd	01b21h, 0, 0, 0
			dd	0, 0, 0, 0
			dd	0, 0, 0, 0

			;-----------
			;0c0h - 0cfh
			;-----------

			dd	0, 01b27h, 027h, 028h
			dd	01b28h, 0, 0, 0
			dd	0, 0, 0, 0
			dd	0, 0, 0, 0

			;-----------
			;0d0h - 0dfh
			;-----------

			dd	0, 0, 0, 0
			dd	0, 0, 0, 0
			dd	0, 0, 0, 0
			dd	0, 0, 0, 0

			;-----------
			;0e0h - 0efh
			;-----------

			dd	0, 0, 0, 0
			dd	0, 0, 0, 0
			dd	0, 0, 0, 0
			dd	0, 0, 0, 0

			;-----------
			;0f0h - 0ffh
			;-----------

			dd	0, 0, 0, 0
			dd	0, 0, 0, 0
			dd	0, 0, 0, 0
			dd	0, 0, 0, 0




;****************************************************************************
;* tty_press_enter **********************************************************
;****************************************************************************
section .text
tty_press_enter:
			push	dword eax
			push	dword ebx
			push	dword ecx
			push	dword edx
			SYS_READ  1, b__character, 1
			pop	dword edx
			pop	dword ecx
			pop	dword ebx
			pop	dword eax
			ret



;****************************************************************************
;* data *********************************************************************
;****************************************************************************
section .data
flag__tty_initialized:	db	0


section .bss
			alignb 4
termios__data:		resb	termios_size

			alignb 4
termios__original:	resb	termios_size

			alignb 4
vtmode__data:		resb	vtmode_size

			alignb 4
vtmode__original:	resb	vtmode_size


			alignb 4
b__character:		resb	1
;********************************************* karsten.scheibler@bigfoot.de *

